#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django import forms
from django.contrib import admin
from django.core.files.base import ContentFile
from django.utils import six

from ..ui.forms import CodeMirrorPythonFileField, CodeMirrorRSTFileField, \
        CodeMirrorJSONFileField, CodeMirrorJSONCharField

from ..common.texts import Messages

from .models import Algorithm as AlgorithmModel
from .models import AlgorithmEndpoint as AlgorithmEndpointModel
from .models import validate_algorithm

from ..ui.forms import NameField

import simplejson as json


#----------------------------------------------------------


class AlgorithmModelForm(forms.ModelForm):

    name = NameField(
            widget=forms.TextInput(attrs=dict(size=80)),
            help_text=Messages['algo_name'],
            )

    declaration_file = CodeMirrorJSONFileField(
            label='Declaration',
            help_text=Messages['json'],
            )

    source_code_file = CodeMirrorPythonFileField(
            label='Source code',
            help_text=Messages['code'],
            )

    description_file = CodeMirrorRSTFileField(
            label='Description',
            required=False,
            allow_empty_file=True,
            help_text=Messages['description'],
            )

    parameters = CodeMirrorJSONCharField(
            readonly=True,
            required=False,
            )

    result_dataformat = CodeMirrorJSONCharField(
            readonly=True,
            required=False,
            )

    class Meta:
        model = AlgorithmModel
        exclude = []
        widgets = {
                'short_description': forms.TextInput(
                    attrs=dict(size=100),
                    ),
                }

    def clean_declaration_file(self):
        """Cleans-up the file data, make sure it is really new"""

        new_declaration = self.cleaned_data['declaration_file'].read()
        old_declaration = ''

        if self.instance and self.instance.declaration_file.name is not None:
            old_declaration = self.instance.declaration_string
            if new_declaration == old_declaration:
                self.changed_data.remove('declaration_file')
                content_file = ContentFile(old_declaration)
                content_file.name = self.instance.declaration_file.name
                return content_file

        # we validate the algorithm to present errors close to the field
        # on the form.
        try:
            validate_algorithm(json.loads(new_declaration))
        except Exception as e:
            raise forms.ValidationError(str(e))

        # if that works out, then we return the passed file
        self.cleaned_data['declaration_file'].seek(0) #reset ContentFile readout
        return self.cleaned_data['declaration_file']

    def clean(self):
        """Cleans-up the input data, make sure it overall validates"""

        # make sure we don't pass back a str field as 'file'
        if 'declaration_file' in self.data and \
                isinstance(self.data['declaration_file'], six.string_types):
            self.data['declaration_file'] = ContentFile(self.data['declaration_file'], name='unsaved')


#----------------------------------------------------------


class AlgorithmEndpointInline(admin.TabularInline):
    model           = AlgorithmEndpointModel
    can_delete      = False
    extra           = 0
    max_num         = 0
    readonly_fields = ('id', 'input', 'channel', 'name', 'dataformat')
    ordering        = ('input', 'channel', 'name')


#----------------------------------------------------------

def rehash_algorithm(modeladmin, request, queryset):
    """Recalculates the hash of an algorithm"""

    for q in queryset: q.save()

rehash_algorithm.short_description = 'Rehash selected algorithms'


class Algorithm(admin.ModelAdmin):

    list_display        = (
                           'id',
                           'author',
                           'name',
                           'version',
                           'splittable',
                           'short_description',
                           'creation_date',
                           'analysis',
                           'previous_version',
                           'fork_of',
                           'sharing',
                           'hash'
                           )

    search_fields       = [
            'author__username',
            'name',
            'short_description',
            'previous_version__author__username',
            'previous_version__name',
            'fork_of__author__username',
            'fork_of__name'
            ]
    list_filter         = ('sharing', )
    list_display_links  = ('id', 'name')
    readonly_fields     = ('hash', 'splittable', 'short_description')

    inlines = [
        AlgorithmEndpointInline,
    ]

    actions = [
        rehash_algorithm,
        ]

    form = AlgorithmModelForm

    filter_horizontal = [
        'shared_with',
        'shared_with_team',
        'usable_by',
        'usable_by_team',
        'referenced_libraries'
    ]

    fieldsets = (
        (None,
          dict(
            fields=('name', 'author',),
            ),
          ),
        ('Documentation',
          dict(
            classes=('collapse',),
            fields=('short_description', 'description_file',),
            ),
          ),
        ('Versioning',
          dict(
            classes=('collapse',),
            fields=('version', 'previous_version', 'fork_of'),
            ),
          ),
        ('Sharing',
          dict(
            classes=('collapse',),
            fields=('sharing', 'shared_with', 'shared_with_team', 'usable_by', 'usable_by_team'),
            ),
          ),
        ('Cached Information (read-only)',
          dict(
            classes=('collapse',),
            fields=('parameters', 'result_dataformat', 'referenced_libraries'),
            ),
          ),
        ('Definition',
          dict(
            fields=('hash', 'splittable', 'declaration_file', 'language', 'source_code_file'),
            ),
          ),
        )

admin.site.register(AlgorithmModel, Algorithm)
