#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.http import Http404
from django.http import HttpResponse
from django.http import HttpResponseForbidden
from django.http import HttpResponseBadRequest
from django.shortcuts import get_object_or_404
from django.conf import settings

import os

from .models import Algorithm
from .serializers import AlgorithmSerializer
from .serializers import FullAlgorithmSerializer
from .serializers import AlgorithmCreationSerializer

from ..code.api import ShareCodeView, RetrieveUpdateDestroyCodeView
from ..code.serializers import CodeDiffSerializer

from ..common.api import (CheckContributionNameView, ListContributionView,
    ListCreateContributionView)

from ..code.api import DiffView


#----------------------------------------------------------


class CheckAlgorithmNameView(CheckContributionNameView):
    """
    This view sanitizes an algorithm name and
    checks whether it is already used.
    """
    model = Algorithm


#----------------------------------------------------------


class ShareAlgorithmView(ShareCodeView):
    """
    This view allows to share an algorithm with
    other users and/or teams
    """
    model = Algorithm


#----------------------------------------------------------


class ListAlgorithmsView(ListContributionView):
    """
    List all available algorithms
    """
    model = Algorithm
    serializer_class = AlgorithmSerializer


#----------------------------------------------------------


class ListCreateAlgorithmsView(ListCreateContributionView):
    """
    Read/Write end point that list the algorithms available
    from a given author and allows the creation of new algorithms
    """
    model = Algorithm
    serializer_class = AlgorithmSerializer
    writing_serializer_class = AlgorithmCreationSerializer
    namespace = 'api_algorithms'


#----------------------------------------------------------


class RetrieveUpdateDestroyAlgorithmsView(RetrieveUpdateDestroyCodeView):
    """
    Read/Write/Delete endpoint for a given algorithm
    """

    model = Algorithm
    serializer_class = FullAlgorithmSerializer

    def do_update(self, request, author_name, object_name, version=None):
        modified, algorithm = super(RetrieveUpdateDestroyAlgorithmsView, self).do_update(request, author_name, object_name, version)

        if modified:
            # Delete existing experiments using the algorithm (code changed)
            experiments = list(set(map(lambda x: x.experiment,
                algorithm.blocks.iterator())))
            for experiment in experiments: experiment.delete()

        return modified, algorithm


#----------------------------------------------------------


class DiffAlgorithmView(DiffView):
    """
    This view shows the differences between two algorithms
    """
    model = Algorithm
    serializer_class = CodeDiffSerializer


#----------------------------------------------------------


def binary(request, author_name, object_name, version=None):
    """Returns the shared library of a binary algorithm
    """

    if request.method not in ['GET', 'POST']:
        return HttpResponseNotAllowed(['GET', 'POST'])

    # Retrieves the algorithm
    if version:
        algorithm = get_object_or_404(
                Algorithm,
                author__username__iexact=author_name,
                name__iexact=object_name,
                version=int(version),
                )
    else:
        algorithm = Algorithm.objects.filter(author__username__iexact=author_name,
                name__iexact=object_name).order_by('-version')
        if not algorithm:
            raise Http404()
        else:
            algorithm = algorithm[0]

    if not algorithm.is_binary():
        raise Http404()

    if request.method == 'GET':
        (has_access, _, accessibility) = algorithm.accessibility_for(request.user, without_usable=True)

        if not has_access:
            raise Http404()

        binary_data = algorithm.source_code

        response = HttpResponse(binary_data, content_type='application/octet-stream')

        response['Content-Length'] = len(binary_data)
        response['Content-Disposition'] = 'attachment; filename=%d.so' % algorithm.version

        return response

    else:
        if request.user.is_anonymous() or (request.user.username != author_name):
            return HttpResponseForbidden()

        if not request.FILES.has_key('binary'):
            return HttpResponseBadRequest()

        file = request.FILES['binary']

        binary_data = ''
        for chunk in file.chunks():
            binary_data += chunk

        algorithm.source_code = binary_data
        algorithm.save()

        return HttpResponse(status=204)
