#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################


from django import template
from django.conf import settings
from ...common.texts import Messages as Texts
from ...common.models import Shareable

register = template.Library()


@register.inclusion_tag('algorithms/panels/table.html', takes_context=True)
def algorithm_table(context, objects, owner, panel_id):
    '''Composes a algorithm list table

    This panel primarily exists for user's algorithm list page.

    Parameters:

        objects (iterable): An iterable containing Algorithm objects
        owner (bool): A flag indicating if the list is being created for the
          owner of his personal list page on the user-microsite or not.
        panel_id: The HTML id to set on the generated table. This is handy for
          the filter functionality normally available on list pages.

    '''

    return {
            'request': context['request'],
            'objects': objects,
            'owner': owner,
            'panel_id': panel_id,
            }


@register.inclusion_tag('algorithms/panels/actions.html', takes_context=True)
def algorithm_actions(context, obj, display_count):
    '''Composes the action buttons for a particular algorithm

    This panel primarily exists for showing action buttons for a given
    algorithm taking into consideration it is being displayed for a given user.

    Parameters:

        obj (Algorithm): The algorithm object concerned for which the
          buttons will be drawn.
        display_count (bool): If the set of buttons should include one with the
          number of experiments using this algorithm.

    '''
    request = context['request']
    return {
            'request': request,
            'object': obj,
            'display_count': display_count,
            'open_source': obj.open_source(request.user),
            }


@register.inclusion_tag('algorithms/panels/editor.html', takes_context=True)
def algorithm_editor(context, obj):
    request = context['request']
    return {
            'owner': request.user == obj.author,
            'object': obj,
            'texts': Texts,
            'open_source': obj.open_source(request.user),
            'downloadable': obj.is_binary() and ((request.user == obj.author) or (obj.sharing == Shareable.PUBLIC)),
            }


@register.inclusion_tag('algorithms/panels/sharing.html', takes_context=True)
def algorithm_sharing(context, obj):
    '''Composes the current sharing properties and a form to change them

    Parameters:

        obj (Algorithm): The algorithm object concerned for which the
          sharing panel will be drawn

    '''
    return {
            'request': context['request'],
            'object': obj,
            'owner': context['request'].user == obj.author,
            'users': context['users'],
            'teams': context['teams'],
            }


@register.assignment_tag(takes_context=True)
def visible_experiments(context, object):
    '''Calculates the visible experiments for a given algorithm and requestor'''

    return object.experiments.for_user(context['request'].user, True)
