#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

'''Scheduling functions and utilities'''

import logging
logger = logging.getLogger(__name__)

from django.db import transaction

from .models import Job, JobSplit, Queue, Worker


def _select_splits_for_queue(queue):
    '''Returns a list of job splits that can run now, at a certain queue

    Here is the work done:

    1. Find the queue availability. This is a bit tricky as queues are only
       allowed to consume a limited (configurable) number of slots in each
       worker, per user

    2. Calculate runnable job splits

    3. TODO: Calculates the list of job splits that can potentially run now
       (for which there is space in the current queue being analyzed), taking
       into consideration the relative use for every user and giving more
       priority to user with less load

    4. Return such a list clipping it so that the number of job splits returned
       does not exceed the queue availability

    The work is done inside the "queue" domain: it does not take into
    consideration other job splits that may also have slots shared on one of
    the machines this queue also has slots on.

    '''

    splits = JobSplit.objects.filter(job__block__queue=queue,
        worker__isnull=True, job__split_errors=0,
        status=Job.QUEUED).order_by('job__runnable_date')

    # lists of all users currently running
    users = set(splits.values_list('job__block__experiment__author', flat=True))

    # number of splits (== slots) running on this queue for each user
    user_slots = [JobSplit.objects.filter(job__block__experiment__author=k,
      job__block__queue=queue, status=Job.PROCESSING).count() for k in users]

    allowance = [queue.max_slots_per_user - k for k in user_slots]
    allowance = dict(zip(users, allowance))

    # limit runnable splits so we reach a maximum of allowed user slots
    splits_to_consider = []
    for s in splits:
        author = s.job.block.experiment.author.id
        if allowance[author] > 0:
            splits_to_consider.append(s)
            allowance[author] -= 1

    # TODO: Sort splits taking into consideration current user load

    # now we have a list of job splits, within each users' allowance, we could
    # submit, clip it by using the total number of available slots on the queue
    # the list is ordered having oldest job splits first
    return splits_to_consider[:queue.availability()]


def schedule():
    '''Schedules job splits that can run now, respecting user/queue usage

    The priorities are:

    1. job splits that require more cores to run (more resource intense)
    2. job splits that are older (runnable since a long time)

    Reasoning: Jobs which require more cores are prioritary over job splits
    that don't require that many resources. The reason for this is practical:
    nodes associated to queues with more cores maybe used by queues that
    requires less slots in case no job splits requiring more cores are present,
    which will happen most of the time if there are more slots available with
    less resources. This will not work if either priorities are not established
    for the use of each node or you only have 1 machine hosting all queues. In
    these cases, you may face a situation in which job splits requesting more
    cores block the execution of job splits requesting less cores because of
    the lack of farm avalability.

    For example, a good setup could be like this (note priorities in this case
    are not useful, so they are all set to 0):

    Queue A (1 core/4Gb RAM/60 minutes/1 job per user):
      - Host 1 (1 core/4Gb): 1 slot (priority 0)
      - Host 2 (2 cores/8Gb): 2 slots (priority 0)

    Queue B (2 cores/8Gb RAM/120 minutes/1 job per user):
      - Host 2 (2 cores/8Gb): 1 slot (priority 0)

    In this way: job splits for Queue B have priority of job splits for Queue
    A.

    If no job splits are scheduled requiring 2 cores, then we have 3 slots
    available to run simpler job splits. In case a job with 2 cores pops-up,
    then we must make sure 1-core job splits running at ``Host 2`` are freed so
    to run that job. To do that, we must first check on more resource intensive
    job splits and then order those by age, processing one after the other.

    You may further setup a queue for "special" users that can run more job
    splits in parallel, like this:

    Queue A (1 core/4Gb RAM/60 minutes/1 job per user):
      - Host 1 (1 core/4Gb): 1 slot (priority 0)
      - Host 2 (2 cores/8Gb): 2 slots (priority 0)

    Queue A* (1 core/4Gb RAM/60 minutes/1 job per user):
      - Host 1 (1 core/4Gb): 1 slot (priority 0)
      - Host 2 (2 cores/8Gb): 2 slots (priority 1)

    Queue B (2 cores/8Gb RAM/120 minutes/1 job per user):
      - Host 2 (2 cores/8Gb): 1 slot (priority 2)

    Queue A* may be setup so that some "special" users can run more job splits
    than others, having the same configuration as Queue A, except for the
    priority on Host 2, which is higher. Users that have access to Queue A or
    Queue A* can use Host 1 on a "first come first served" basis (as they have
    the same priority). Host 2 is primarily used for 2-core job splits. If job
    splits requiring a single core appears to run on Host 2 (from either Queue
    A or A*), then job splits from Queue A* have priority.

    Here is a summary of the work done:

    1. Use :py:func:`_select_splits_for_queue` to figure out which splits can
       actually run on a queue basis, respecting user limits

    2. For those splits, prioritize execution by: number of cores required per
       job split and then, by age.

    3. Remove splits, based on the order, which have the same output

    4. Attribute more resource intensive job splits to the slots that are free.
       If no slots are free, then virtually block free slots on hosts that can
       process job splits requiring more resources. This will guarantee the
       remaining cores will be freed in the future and this resource-intensive
       job splits will be able to run at a certain point.

    5. For each job split on the list, check the queue virtual availability
       (taking into consideration blocking as executed on step 3) and leave the
       job split on the list if there is a free slot, otherwise remove it.


    Returns:

      list: The list of splits assigned at this scheduling iteration

    '''

    # updates jobs with split errors, cancel experiments if problems occur
    for j in Job.objects.filter(split_errors__gt=0):
        j._split_indices()

    # get queues in a good order
    sorted_queues = Queue.objects.order_by('-cores_per_slot',
        'max_slots_per_user')

    splits_to_consider = \
        [_select_splits_for_queue(q) for q in sorted_queues]

    if not any(splits_to_consider): return []

    logger.debug('Considering splits: %s', splits_to_consider)

    # decides which split to run considering the 'on-the-fly' availability

    # workers that can run job splits
    whitelist = {}
    for worker in Worker.objects.filter(active=True):
        availability = worker.available_cores()
        if availability <= 0: continue
        whitelist[worker] = availability

    assigned_splits = []

    logger.debug('Worker availability: %s', whitelist)

    for batch in splits_to_consider:

        if not batch: continue #empty list
        candidates = batch[0].job.block.queue.worker_availability()

        for split in batch:

            assigned = False
            required_cores = split.job.block.queue.cores_per_slot

            for c in candidates:
                avail = whitelist.get(c, 0)
                if not avail: continue #should not use this worker
                if avail >= required_cores:
                    logger.debug("Assigning `%s' to worker `%s'", split, c)
                    split.schedule(c) #assign job split to worker
                    assigned_splits.append(split)
                    whitelist[c] -= required_cores
                    logger.debug("`%s' cores available: %d", c, whitelist[c])
                    assigned = True
                    break

            if not assigned and required_cores > 1:
                #blacklist the most promising worker, otherwise will never be
                #picked
                c0 = candidates[0]
                if c0 in whitelist:
                    logger.info("Could not assign `%s' so blacklisting `%s'",
                        split, c0)
                    del whitelist[c0]

    return assigned_splits
