#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import os
import sys
import time
import shutil
import tempfile
import collections
import time

from django.conf import settings
from django.core.urlresolvers import reverse
from django.core import management
from django.contrib.auth.models import User, Group
from django.test import TestCase, TransactionTestCase

from guardian.shortcuts import get_perms

from ..common.testutils import BaseTestCase as APITestCase, tearDownModule
from ..experiments.models import Experiment, Block
from ..algorithms.models import Algorithm
from ..utils.management.commands import install
from ..statistics.models import HourlyStatistics

from .models import Queue, Worker, Slot, Environment, Job, JobSplit, Result
from .utils import cleanup_cache, dump_backend, setup_backend
from .management.commands import qsetup
from .schedule import schedule


def _sleep(tries, condition):
  """For some reason, time.sleep is not reliable on this test unit. Use this"""

  seconds = 1.0 #between tries
  for i in range(tries):
      if condition(): return
      slept = 0
      while slept < seconds:
          start = time.time()
          time.sleep(seconds - slept)
          slept += time.time() - start


# Example configuration with 3 queues with an increasing amount of resources
# running on the same host
QUEUES_WITHOUT_PRIORITY = {
  "queues": collections.OrderedDict([
    ("q1", {
      "memory-limit": 4*1024,
      "time-limit": 180, #3 hours
      "cores-per-slot": 1,
      "max-slots-per-user": 4,
      "environments": ['environment (1)'],
      "groups": [
        "Default",
        ],
      "slots": {
        "node1": {
          "quantity": 4,
          "priority": 0
          }
        }
      }
    ),
    ("q2", {
      "memory-limit": 8*1024,
      "time-limit": 360, #6 hours
      "cores-per-slot": 2,
      "max-slots-per-user": 2,
      "environments": ['environment (1)'],
      "groups": [
        "Default",
        ],
      "slots": {
        "node1": {
          "quantity": 2,
          "priority": 0
          },
        }
      }
    ),
    ("q4", {
      "memory-limit": 16*1024,
      "time-limit": 720, #12 hours
      "cores-per-slot": 4,
      "max-slots-per-user": 1,
      "environments": ['environment (1)'],
      "groups": [
        "Default",
        ],
      "slots": {
        "node1": {
          "quantity": 1,
          "priority": 0
          },
        }
      }
    )
  ]),
  "workers": {
    "node1": {
      "cores": 4,
      "memory": 16*1024,
      }
    },
  "environments": {
    'environment (1)': {
      "name": 'environment',
      "version": '1',
      "short_description": "Test",
      "description": "Test environment",
      "languages": "python",
      },
    },
  }

# Example configuration with 3 queues sharing slots on 2 hosts
PRIORITY_QUEUES = {
    "queues": collections.OrderedDict([
      ("q1", {
        "memory-limit": 4*1024,
        "time-limit": 180, #3 hours
        "cores-per-slot": 1,
        "max-slots-per-user": 2,
        "environments": ['environment (1)'],
        "groups": [
          "Default",
          ],
        "slots": {
          "node1": {
            "quantity": 4,
            "priority": 5
            },
          "node2": {
            "quantity": 4,
            "priority": 0
            },
          }
        },
      ),
      ("q2", {
        "memory-limit": 8*1024,
        "time-limit": 360, #6 hours
        "cores-per-slot": 2,
        "max-slots-per-user": 1,
        "environments": ['environment (1)'],
        "groups": [
          "Default",
          ],
        "slots": {
          "node1": {
            "quantity": 2,
            "priority": 0
            },
          "node2": {
            "quantity": 2,
            "priority": 10
            }
          }
        },
      ),
      ("q1_special", {
        "memory-limit": 4*1024,
        "time-limit": 180, #3 hours
        "cores-per-slot": 1,
        "max-slots-per-user": 8,
        "environments": ['environment (1)'],
        "groups": [
          "Default",
          ],
        "slots": {
          "node1": {
            "quantity": 4,
            "priority": 0
            },
          "node2": {
            "quantity": 4,
            "priority": 5
            }
          }
        }
      ),
    ]),
    "workers": collections.OrderedDict([
      ("node1", {
        "cores": 4,
        "memory": 32*1024,
        }
      ),
      ("node2", {
        "cores": 4,
        "memory": 16*1024,
        }
      )
    ]),
    "environments": {
        'environment (1)': {
          "name": 'environment',
          "version": '1',
          "short_description": "Test",
          "description": "Test environment",
          "languages": "python",
          },
        },
  }


class CancelAllExperimentsAPI(APITestCase):

    def setUp(self):
        self.url = reverse('backend:cancel-experiments')


    def test_no_access_for_anonymous_user(self):
        response = self.client.get(self.url)
        self.checkResponse(response, 302) #redirects to login page


    def test_no_access_for_non_superuser(self):
        User.objects.create_user('johndoe', 'johndoe@test.org', '1234')
        self.client.login(username='johndoe', password='1234')
        response = self.client.get(self.url)
        self.checkResponse(response, 403)


class CacheCleanUp(TestCase):


    def setUp(self):
        self.cache = tempfile.mkdtemp(prefix='beat_')


    def tearDown(self):
        shutil.rmtree(self.cache)


    def touch(self, f, times=None):
        """Replicates the `touch' command-line utility"""
        with open(f, 'a'): os.utime(f, times)


    def J(self, *args):
        return os.path.join(*((self.cache,) + args))


    def prepare_cleanup_full(self):

        # creates a temporary directory structure
        os.makedirs(self.J('a', 'b', 'c'))
        os.makedirs(self.J('a', 'c', 'd'))
        os.makedirs(self.J('a', 'c', 'e'))
        self.touch(self.J('a', 'b', 'c', 'd.json'))
        self.touch(self.J('a', 'c', 'd', 'e.json'))


    def check_cleanup_full(self):

        assert not os.listdir(self.cache)


    def test_cache_cleanup_full(self):

        self.prepare_cleanup_full()
        cleanup_cache(self.cache, delete=True)
        self.check_cleanup_full()


    def test_cmd_cleanup_full(self):

        self.prepare_cleanup_full()
        management.call_command('cleanup_cache', path=self.cache,
            verbosity=0, delete=True)
        self.check_cleanup_full()


    def prepare_cleanup_aged(self):

        two_min_ago = time.time() - 60*2

        # creates a temporary directory structure
        os.makedirs(self.J('a', 'b', 'c'))
        os.makedirs(self.J('a', 'c', 'd'))
        os.makedirs(self.J('a', 'c', 'e'))
        self.touch(self.J('a', 'b', 'c', 'd.json'), (two_min_ago, two_min_ago))
        self.touch(self.J('a', 'c', 'd', 'e.json'))


    def check_cleanup_aged(self):

        assert os.path.exists(self.J('a', 'c', 'd', 'e.json'))
        assert not os.path.exists(self.J('a', 'b', 'c'))
        assert not os.path.exists(self.J('a', 'b', 'c', 'd.json'))
        assert not os.path.exists(self.J('a', 'b', 'e'))


    def test_cache_cleanup_aged(self):

        self.prepare_cleanup_aged()
        cleanup_cache(self.cache, age_in_minutes=2, delete=True)
        self.check_cleanup_aged()


    def test_cmd_cleanup_aged(self):

        self.prepare_cleanup_aged()
        management.call_command('cleanup_cache', path=self.cache,
            verbosity=0, olderthan=2, delete=True)
        self.check_cleanup_aged()


    def prepare_cleanup_lock(self):

        two_min_ago = time.time() - 60*2
        ten_min_ago = time.time() - 60*10

        # creates a temporary directory structure
        os.makedirs(self.J('a', 'b', 'c'))
        os.makedirs(self.J('a', 'c', 'd'))
        os.makedirs(self.J('a', 'c', 'e'))
        self.touch(self.J('a', 'b', 'c', 'd.json'), (two_min_ago, two_min_ago))
        self.touch(self.J('a', 'c', 'd', 'e.json'), (ten_min_ago, ten_min_ago))

        self.touch(self.J('a', 'c', 'd', 'e.lock')) #create a lock


    def check_cleanup_lock(self):

        assert os.path.exists(self.J('a', 'c', 'd', 'e.json'))
        assert not os.path.exists(self.J('a', 'b', 'c'))
        assert not os.path.exists(self.J('a', 'b', 'c', 'd.json'))
        assert not os.path.exists(self.J('a', 'b', 'e'))


    def test_cache_cleanup_lock(self):

        self.prepare_cleanup_lock()
        cleanup_cache(self.cache, delete=True)
        self.check_cleanup_lock()


    def test_cmd_cleanup_lock(self):

        self.prepare_cleanup_lock()
        management.call_command('cleanup_cache', path=self.cache,
            verbosity=0, delete=True)
        self.check_cleanup_lock()


class BaseBackendTestCase(TestCase):


    @classmethod
    def setUpTestData(cls):
        install.create_sites()
        system_user, plot_user, user = install.create_users('user', 'user')
        install.add_group('Default')

        setup_backend(qsetup.DEFAULT_CONFIGURATION)

        Worker.objects.update(active=True)
        env = Environment.objects.get(name='environment')
        queue = Queue.objects.first()

        template_data = dict(
            system_user = system_user,
            plot_user = plot_user,
            user = user,
            private = False,
            queue = queue.name,
            environment = dict(name=env.name, version=env.version),
            )
        prefix = os.path.join(
              os.path.dirname(os.path.dirname(os.path.realpath(sys.argv[0]))),
              'src',
              'beat.examples',
              )
        install.install_contributions(prefix, 'system', template_data)
        install.install_contributions(prefix, 'test', template_data)


    def check_single(self, xp):
        '''Checks user/user/single/1/single'''

        self.assertEqual(xp.blocks.count(), 2)

        b0 = xp.blocks.all()[0]

        self.assertEqual(b0.name, 'echo')
        self.assertEqual(b0.status, Block.NOT_CACHED)
        self.assertEqual(b0.algorithm,
            Algorithm.objects.get(name='integers_echo'))
        self.assertEqual(b0.dependencies.count(), 0)
        self.assertEqual(b0.dependents.count(), 1)
        self.assertEqual(b0.job.status, Job.QUEUED)
        self.assertEqual(b0.job.parent, None)
        self.assertEqual(b0.job.child_, None)
        self.assertEqual(b0.queue.name, 'queue')
        self.assertEqual(b0.environment.name, 'environment')
        self.assertEqual(b0.required_slots, 1)
        self.assertEqual(b0.inputs.count(), 1)
        self.assertEqual(b0.outputs.count(), 1)
        self.assertEqual(b0.job.splits.count(), 1)
        self.assertEqual(b0.job.splits.get().status, Job.QUEUED)
        assert b0.job.splits.get().worker is None
        assert not b0.done()

        b1 = xp.blocks.all()[1]

        self.assertEqual(b1.name, 'analysis')
        self.assertEqual(b1.status, Block.NOT_CACHED)
        self.assertEqual(b1.algorithm,
            Algorithm.objects.get(name='integers_echo_analyzer'))
        self.assertEqual(b1.dependencies.count(), 1)
        self.assertEqual(b1.dependents.count(), 0)
        self.assertEqual(b1.job.status, Job.QUEUED)
        self.assertEqual(b1.job.parent, None)
        self.assertEqual(b1.job.child_, None)
        self.assertEqual(b1.queue.name, 'queue')
        self.assertEqual(b1.environment.name, 'environment')
        self.assertEqual(b1.required_slots, 1)
        self.assertEqual(b1.inputs.count(), 1)
        self.assertEqual(b1.outputs.count(), 1)
        self.assertEqual(b1.job.splits.count(), 0) #not scheduled yet

        assert not b1.done()



class BackendSetup(BaseBackendTestCase):


    def check_default_config(self):

        # checks all is there
        self.assertEqual(dump_backend(), qsetup.DEFAULT_CONFIGURATION)

        worker = Worker.objects.get()
        queue = Queue.objects.get()
        Worker.objects.update(active=True)

        self.assertEqual(worker.available_cores(), qsetup.CORES)
        self.assertEqual(list(worker.slots.values_list('id', flat=True)),
            list(queue.slots.values_list('id', flat=True)))

        # worker has no job splits assigned to it
        self.assertEqual(worker.splits.count(), 0)

        self.assertEqual(queue.availability(), qsetup.CORES)
        self.assertEqual(queue.number_of_slots(), qsetup.CORES)
        self.assertEqual(queue.worker_availability(), [worker])

        # checks the single slot and priority
        slot = queue.slots.get()
        self.assertEqual(slot.quantity, qsetup.CORES)
        self.assertEqual(slot.priority, 0)
        self.assertEqual(slot.worker, worker)

        # checks no orphan slots exist
        self.assertEqual(Slot.objects.filter(queue=None).count(), 0)
        self.assertEqual(Slot.objects.filter(worker=None).count(), 0)


    def test_setup(self):

        self.check_default_config()


    def test_cmd_reset(self):

        # installs the default configuration command
        management.call_command('qsetup', verbosity=0, reset=True)
        self.check_default_config()


    def check_noprior_config(self):

        qs = Queue.objects.all()

        self.assertEqual(qs.count(), 3)

        q1, q2, q3 = qs.order_by('name')

        self.assertEqual(q1.name, 'q1')
        self.assertEqual(q2.name, 'q2')
        self.assertEqual(q3.name, 'q4')
        self.assertEqual(q1.splits().count(), 0)
        self.assertEqual(q2.splits().count(), 0)
        self.assertEqual(q3.splits().count(), 0)

        self.assertEqual(q1.number_of_slots(), 4)
        self.assertEqual(q2.number_of_slots(), 2)
        self.assertEqual(q3.number_of_slots(), 1)
        self.assertEqual(q1.availability(), 4)
        self.assertEqual(q2.availability(), 2)
        self.assertEqual(q3.availability(), 1)
        self.assertEqual(q1.environments.count(), 1)
        self.assertEqual(q2.environments.count(), 1)
        self.assertEqual(q3.environments.count(), 1)

        self.assertEqual(q1.environments.first(), q2.environments.first())
        self.assertEqual(q2.environments.first(), q3.environments.first())

        env = q1.environments.first()

        self.assertEqual(env.name, 'environment')
        self.assertEqual(env.version, '1')

        self.assertEqual(q1.slots.count(), 1)
        self.assertEqual(q2.slots.count(), 1)
        self.assertEqual(q3.slots.count(), 1)

        slot1 = q1.slots.first()
        slot2 = q2.slots.first()
        slot3 = q3.slots.first()

        self.assertEqual(slot1.quantity, 4)
        self.assertEqual(slot1.priority, 0)
        self.assertEqual(slot1.queue, q1)
        self.assertEqual(slot2.quantity, 2)
        self.assertEqual(slot2.priority, 0)
        self.assertEqual(slot2.queue, q2)
        self.assertEqual(slot3.quantity, 1)
        self.assertEqual(slot3.priority, 0)
        self.assertEqual(slot3.queue, q3)

        worker1 = slot1.worker
        worker2 = slot2.worker
        worker3 = slot3.worker

        self.assertEqual(worker1, worker2)
        self.assertEqual(worker2, worker3)

        self.assertEqual(worker1.name, 'node1')
        self.assertEqual(list(worker1.splits.all()), [])
        self.assertEqual(worker1.memory, 16*1024)
        self.assertEqual(worker1.cores, 4)
        self.assertEqual(worker1.available_cores(), 4)

        self.assertEqual(worker1.slots.count(), 3)

        self.assertEqual(set(worker1.slots.all()),
            set(list(q1.slots.all()) + list(q2.slots.all()) + \
                list(q3.slots.all())))

        avail1 = q1.worker_availability()
        self.assertEqual(avail1, [worker1])

        avail2 = q2.worker_availability()
        self.assertEqual(avail2, [worker1])

        avail3 = q3.worker_availability()
        self.assertEqual(avail2, [worker1])

        # checks no orphan slots exist
        self.assertEqual(Slot.objects.filter(queue=None).count(), 0)
        self.assertEqual(Slot.objects.filter(worker=None).count(), 0)


    def test_reconfigure_noprior(self):

        setup_backend(QUEUES_WITHOUT_PRIORITY)
        Worker.objects.update(active=True)
        self.check_noprior_config()


    def test_reconfigure_fail_qenv_used(self):

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        # schedules the experiment and check it
        xp.schedule()
        self.check_single(xp)

        try:
            setup_backend(QUEUES_WITHOUT_PRIORITY)
        except RuntimeError as e:
            assert str(e).find('on the following queue/environment combinations') != -1
        else:
            assert False, 'Queue re-configuration worked with q/env in use'


    def check_prior_config(self):

        qs = Queue.objects.all()

        self.assertEqual(qs.count(), 3)

        q1, q1_special, q2 = qs.order_by('name')

        self.assertEqual(q1.name, 'q1')
        self.assertEqual(q2.name, 'q2')
        self.assertEqual(q1_special.name, 'q1_special')
        self.assertEqual(q1.splits().count(), 0)
        self.assertEqual(q2.splits().count(), 0)
        self.assertEqual(q1_special.splits().count(), 0)

        self.assertEqual(q1.number_of_slots(), 8)
        self.assertEqual(q2.number_of_slots(), 4)
        self.assertEqual(q1_special.number_of_slots(), 8)
        self.assertEqual(q1.availability(), 8)
        self.assertEqual(q2.availability(), 4)
        self.assertEqual(q1_special.availability(), 8)
        self.assertEqual(q1.environments.count(), 1)
        self.assertEqual(q2.environments.count(), 1)
        self.assertEqual(q1_special.environments.count(), 1)

        self.assertEqual(q1.environments.first(), q2.environments.first())
        self.assertEqual(q2.environments.first(),
            q1_special.environments.first())

        env = q1.environments.first()

        self.assertEqual(env.name, 'environment')
        self.assertEqual(env.version, '1')

        self.assertEqual(q1.slots.count(), 2)
        self.assertEqual(q1_special.slots.count(), 2)
        self.assertEqual(q2.slots.count(), 2)

        slot11, slot12 = q1.slots.all()
        slot1_special1, slot1_special2 = q1_special.slots.all()
        slot21, slot22 = q2.slots.all()

        self.assertEqual(slot11.quantity, 4)
        self.assertEqual(slot11.priority, 5)
        self.assertEqual(slot12.quantity, 4)
        self.assertEqual(slot12.priority, 0)
        self.assertEqual(slot11.queue, q1)
        self.assertEqual(slot12.queue, q1)

        self.assertEqual(slot21.quantity, 2)
        self.assertEqual(slot21.priority, 0)
        self.assertEqual(slot22.quantity, 2)
        self.assertEqual(slot22.priority, 10)
        self.assertEqual(slot21.queue, q2)
        self.assertEqual(slot22.queue, q2)

        self.assertEqual(slot1_special1.quantity, 4)
        self.assertEqual(slot1_special1.priority, 0)
        self.assertEqual(slot1_special2.quantity, 4)
        self.assertEqual(slot1_special2.priority, 5)
        self.assertEqual(slot1_special1.queue, q1_special)
        self.assertEqual(slot1_special2.queue, q1_special)

        worker1 = slot11.worker
        worker2 = slot12.worker
        worker21 = slot21.worker
        worker22 = slot22.worker
        worker1_special1 = slot1_special1.worker
        worker1_special2 = slot1_special2.worker

        self.assertEqual(worker1, worker21)
        self.assertEqual(worker1, worker1_special1)
        self.assertEqual(worker2, worker22)
        self.assertEqual(worker2, worker1_special2)

        self.assertEqual(worker1.name, 'node1')
        self.assertEqual(worker1.splits.count(), 0)
        self.assertEqual(worker1.memory, 32*1024)
        self.assertEqual(worker1.cores, 4)
        self.assertEqual(worker1.available_cores(), 4)

        self.assertEqual(worker2.name, 'node2')
        self.assertEqual(worker2.splits.count(), 0)
        self.assertEqual(worker2.memory, 16*1024)
        self.assertEqual(worker2.cores, 4)
        self.assertEqual(worker2.available_cores(), 4)

        self.assertEqual(worker1.slots.count(), 3)
        self.assertEqual(worker2.slots.count(), 3)

        avail1 = q1.worker_availability()
        self.assertEqual(avail1, [worker1, worker2])

        avail2 = q2.worker_availability()
        self.assertEqual(avail2, [worker2, worker1])

        avail1_special = q1_special.worker_availability()
        self.assertEqual(avail1_special, [worker2, worker1])

        # checks no orphan slots exist
        self.assertEqual(Slot.objects.filter(queue=None).count(), 0)
        self.assertEqual(Slot.objects.filter(worker=None).count(), 0)


    def test_reconfigure_priors(self):

        setup_backend(PRIORITY_QUEUES)
        Worker.objects.update(active=True)
        self.check_prior_config()




class Scheduling(BaseBackendTestCase):


    def check_stats_success(self, split):

        assert abs(split.job.block.speed_up_real() - 1.0) < 0.1
        assert abs(split.job.block.speed_up_maximal() - 1.0) < 0.1
        assert split.job.block.linear_execution_time() > 0.0
        assert split.job.block.queuing_time() > 0.0
        assert split.job.block.stdout() is None
        assert split.job.block.stderr() is None
        assert split.job.block.error_report() is None


    def test_success(self):

        # tests a simple successful experiment scheduling and execution

        current_stats = HourlyStatistics.objects.count()

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        # schedules the experiment and check it
        xp.schedule()
        self.check_single(xp)
        assigned_splits = schedule()

        # schedules the first runnable block
        assert xp.blocks.first().job.runnable_date is not None
        assert xp.blocks.last().job.runnable_date is None

        worker = Worker.objects.get()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'echo')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # simulate job start on worker
        split.start()
        self.assertEqual(split.job.status, Job.PROCESSING)
        self.assertEqual(split.job.block.status, Block.PROCESSING)
        self.assertEqual(split.job.block.experiment.status, Experiment.RUNNING)

        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # no job can be run right now
        assigned_splits = schedule()
        self.assertEqual(len(assigned_splits), 0)

        # simulate end job signal
        split.end(Result(status=0))
        self.assertEqual(split.job.status, Job.COMPLETED)
        self.assertEqual(split.job.block.status, Block.CACHED)
        self.assertEqual(split.job.block.experiment.status, Experiment.RUNNING)

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        self.check_stats_success(split)

        # assert we have no database traces after the block is done
        self.assertEqual(Job.objects.filter(block=split.job.block).count(), 0)
        self.assertEqual(JobSplit.objects.filter(job=split.job).count(), 0)
        self.assertEqual(Result.objects.filter(job__isnull=True).count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)

        # since this job was successful, the next one should be ready to run

        # schedules the last block of the experiment
        assert xp.blocks.last().job.runnable_date is not None
        assigned_splits = schedule()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'analysis')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # simulate job start on worker
        split.start()
        self.assertEqual(split.job.status, Job.PROCESSING)
        self.assertEqual(split.job.block.status, Block.PROCESSING)
        self.assertEqual(split.job.block.experiment.status, Experiment.RUNNING)

        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # no job can be run right now
        assigned_splits = schedule()
        self.assertEqual(len(assigned_splits), 0)

        # simulate end job signal
        split.end(Result(status=0))

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        self.assertEqual(split.job.status, Job.COMPLETED)
        self.assertEqual(split.job.block.status, Block.CACHED)
        self.assertEqual(split.job.block.experiment.status, Experiment.DONE)

        self.check_stats_success(split)

        # assert we have no database traces after the last block is done
        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)
        self.assertEqual(Result.objects.count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)


    def test_does_not_reassign(self):

        # tests if the scheduling routine never re-assigns splits which are
        # already assigned.

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        # schedules the experiment and check it
        xp.schedule()
        self.check_single(xp)
        assigned_splits = schedule()

        self.assertEqual(len(assigned_splits), 1)

        assigned_splits = schedule()
        self.assertEqual(len(assigned_splits), 0)


    def test_worker_activation(self):

        # tests that scheduling depends on worker activation

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        # de-activates worker
        Worker.objects.update(active=False)

        # schedules the experiment and check it
        xp.schedule()
        self.check_single(xp)

        # no job can be run right now
        assigned_splits = schedule()
        self.assertEqual(len(assigned_splits), 0)

        # re-activate the worker, show it now schedules fine
        Worker.objects.update(active=True)
        assigned_splits = schedule()
        self.assertEqual(len(assigned_splits), 1)

        # the rest would continue like with test_success


    def test_fails_on_first_block(self):

        # tests that, if we fail on the first block, experiment fails, all
        # stops as foreseen

        current_stats = HourlyStatistics.objects.count()

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        # schedules the experiment and check it
        xp.schedule()
        self.check_single(xp)

        # schedules the first runnable block
        assert xp.blocks.first().job.runnable_date is not None
        assert xp.blocks.last().job.runnable_date is None
        assigned_splits = schedule()

        worker = Worker.objects.get()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'echo')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # simulate job start on worker
        split.start()
        self.assertEqual(split.job.status, Job.PROCESSING)
        self.assertEqual(split.job.block.status, Block.PROCESSING)
        self.assertEqual(split.job.block.experiment.status, Experiment.RUNNING)

        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # no job can be run right now
        assigned_splits = schedule()
        self.assertEqual(len(assigned_splits), 0)

        # simulate end job signal, faiulre
        split.end(Result(status=1))
        self.assertEqual(split.job.status, Job.FAILED)
        self.assertEqual(split.job.block.status, Block.FAILED)
        split.job.block.experiment.refresh_from_db()
        self.assertEqual(split.job.block.experiment.status, Experiment.FAILED)

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        # assert we have no database traces after the last block is done
        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)
        self.assertEqual(Result.objects.count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)


    def test_fails_on_last_block(self):

        # tests a simple successful experiment scheduling and execution

        current_stats = HourlyStatistics.objects.count()

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        # schedules the experiment and check it
        xp.schedule()
        self.check_single(xp)

        # schedules the first runnable block
        assert xp.blocks.first().job.runnable_date is not None
        assert xp.blocks.last().job.runnable_date is None
        assigned_splits = schedule()

        worker = Worker.objects.get()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'echo')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # simulate job start on worker
        split.start()
        self.assertEqual(split.job.status, Job.PROCESSING)
        self.assertEqual(split.job.block.status, Block.PROCESSING)
        self.assertEqual(split.job.block.experiment.status, Experiment.RUNNING)

        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # no job can be run right now
        assigned_splits = schedule()
        self.assertEqual(len(assigned_splits), 0)

        # simulate end job signal
        split.end(Result(status=0))
        self.assertEqual(split.job.status, Job.COMPLETED)
        self.assertEqual(split.job.block.status, Block.CACHED)
        self.assertEqual(split.job.block.experiment.status, Experiment.RUNNING)

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        self.check_stats_success(split)

        # assert we have no database traces after the last block is done
        self.assertEqual(Job.objects.filter(block=split.job.block).count(), 0)
        self.assertEqual(JobSplit.objects.filter(job=split.job).count(), 0)
        self.assertEqual(Result.objects.filter(job__isnull=True).count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)

        # since this job was successful, the next one should be ready to run

        # schedules the last block of the experiment
        assert xp.blocks.last().job.runnable_date is not None
        assigned_splits = schedule()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'analysis')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # simulate job start on worker
        split.start()
        self.assertEqual(split.job.status, Job.PROCESSING)
        self.assertEqual(split.job.block.status, Block.PROCESSING)
        self.assertEqual(split.job.block.experiment.status, Experiment.RUNNING)

        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # no job can be run right now
        assigned_splits = schedule()
        self.assertEqual(len(assigned_splits), 0)

        # simulate end job signal
        split.end(Result(status=1))

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        self.assertEqual(split.job.status, Job.FAILED)
        self.assertEqual(split.job.block.status, Block.FAILED)
        self.assertEqual(split.job.block.experiment.status, Experiment.FAILED)

        # assert we have no database traces after the last block is done
        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)
        self.assertEqual(Result.objects.count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)


    def test_cancel_before_starting(self):

        # tests experiment cancellation before the experiment is started

        current_stats = HourlyStatistics.objects.count()

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        # schedules the experiment and check it
        xp.schedule()
        self.check_single(xp)

        # schedules the first runnable block
        assert xp.blocks.first().job.runnable_date is not None
        assert xp.blocks.last().job.runnable_date is None

        xp.cancel()

        self.assertEqual(
            [str(k) for k in xp.blocks.values_list('status', flat=True)],
            [Block.CANCELLED, Block.CANCELLED]
            )
        xp.refresh_from_db()
        self.assertEqual(xp.status, Experiment.FAILED)

        # assert we have no database traces after the last block is done
        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)
        self.assertEqual(Result.objects.count(), 0)

        worker = Worker.objects.get()
        self.assertEqual(worker.available_cores(), qsetup.CORES)


    def test_cancel_after_success(self):

        # tests experiment cancellation while the experiment is running

        current_stats = HourlyStatistics.objects.count()

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        # schedules the experiment and check it
        xp.schedule()
        self.check_single(xp)

        # schedules the first runnable block
        assert xp.blocks.first().job.runnable_date is not None
        assert xp.blocks.last().job.runnable_date is None
        assigned_splits = schedule()

        worker = Worker.objects.get()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'echo')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # simulate job start on worker
        split.start()
        self.assertEqual(split.job.status, Job.PROCESSING)
        self.assertEqual(split.job.block.status, Block.PROCESSING)
        self.assertEqual(split.job.block.experiment.status, Experiment.RUNNING)

        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # no job can be run right now
        assigned_splits = schedule()
        self.assertEqual(len(assigned_splits), 0)

        # simulate end job signal
        split.end(Result(status=0))
        self.assertEqual(split.job.status, Job.COMPLETED)
        self.assertEqual(split.job.block.status, Block.CACHED)
        self.assertEqual(split.job.block.experiment.status, Experiment.RUNNING)

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        self.check_stats_success(split)

        # assert we have no database traces after the last block is done
        self.assertEqual(Job.objects.filter(block=split.job.block).count(), 0)
        self.assertEqual(JobSplit.objects.filter(job=split.job).count(), 0)
        self.assertEqual(Result.objects.filter(job__isnull=True).count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)

        # since this job was successful, the next one should be ready to run

        # schedules the last block of the experiment
        assert xp.blocks.last().job.runnable_date is not None
        xp.cancel()

        self.assertEqual(
            [str(k) for k in xp.blocks.order_by('id').values_list('status', flat=True)],
            [Block.CACHED, Block.CANCELLED]
            )
        self.assertEqual(xp.status, Experiment.FAILED)

        # assert we have no database traces after the last block is done
        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)
        self.assertEqual(Result.objects.count(), 0)

        worker = Worker.objects.get()
        self.assertEqual(worker.available_cores(), qsetup.CORES)


    def test_cancel_while_running(self):

        # tests experiment cancellation while a block is running

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        # schedules the experiment and check it
        xp.schedule()
        self.check_single(xp)

        # schedules the first runnable block
        assert xp.blocks.first().job.runnable_date is not None
        assert xp.blocks.last().job.runnable_date is None
        assigned_splits = schedule()

        worker = Worker.objects.get()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'echo')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # simulate job start on worker
        split.start()
        self.assertEqual(split.job.status, Job.PROCESSING)
        self.assertEqual(split.job.block.status, Block.PROCESSING)
        self.assertEqual(split.job.block.experiment.status, Experiment.RUNNING)

        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # no job can be run right now
        assigned_splits = schedule()
        self.assertEqual(len(assigned_splits), 0)

        xp.cancel()

        # simulate worker cancelling
        split.refresh_from_db()
        self.assertEqual(split.status, Job.CANCEL)
        split.end(None, Job.CANCELLED)

        xp.refresh_from_db()
        self.assertEqual(
            [str(k) for k in xp.blocks.order_by('id').values_list('status', flat=True)],
            [Block.CANCELLED, Block.CANCELLED]
            )
        self.assertEqual(xp.status, Experiment.FAILED)

        # assert we have no database traces after the last block is done
        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)
        self.assertEqual(Result.objects.count(), 0)

        worker = Worker.objects.get()
        self.assertEqual(worker.available_cores(), qsetup.CORES)


    def test_cancel_after_failure(self):

        # tests that, if we fail on the first block, experiment fails and a
        # cancellation that comes after that is a NOOP

        current_stats = HourlyStatistics.objects.count()

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        # schedules the experiment and check it
        xp.schedule()
        self.check_single(xp)

        # schedules the first runnable block
        assert xp.blocks.first().job.runnable_date is not None
        assert xp.blocks.last().job.runnable_date is None
        assigned_splits = schedule()

        worker = Worker.objects.get()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'echo')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # simulate job start on worker
        split.start()
        self.assertEqual(split.job.status, Job.PROCESSING)
        self.assertEqual(split.job.block.status, Block.PROCESSING)
        self.assertEqual(split.job.block.experiment.status, Experiment.RUNNING)

        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # no job can be run right now
        assigned_splits = schedule()
        self.assertEqual(len(assigned_splits), 0)

        # simulate end job signal, faiulre
        split.end(Result(status=1))
        self.assertEqual(split.job.status, Job.FAILED)
        self.assertEqual(split.job.block.status, Block.FAILED)
        split.job.block.experiment.refresh_from_db()
        self.assertEqual(split.job.block.experiment.status, Experiment.FAILED)

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        # assert we have no database traces after the last block is done
        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)
        self.assertEqual(Result.objects.count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)

        xp.cancel()
        self.assertEqual(split.job.block.experiment.status, Experiment.FAILED)


    def test_blocking_success(self):

        # tests two experiments that are similar can be scheduled at the same
        # time and we'll optimise correctly and only run one of them. The other
        # is updated as the blocking experiment is executed.

        current_stats = HourlyStatistics.objects.count()

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        xpc = xp.fork(name='single_copy')

        # schedules the experiment and check it
        xp.schedule()
        xpc.schedule()

        # schedules the first runnable block
        assert xp.blocks.first().job.runnable_date is not None
        assert xp.blocks.last().job.runnable_date is None
        assert xpc.blocks.first().job.runnable_date is None
        assert xpc.blocks.last().job.runnable_date is None

        assigned_splits = schedule()

        worker = Worker.objects.get()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'echo')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # checks the jobs are connected one to the other across experiments
        self.assertEqual(xp.blocks.first().job.child.block.experiment, xpc)
        self.assertEqual(xp.blocks.last().job.child.block.experiment, xpc)

        # simulate job start on worker
        split.start()
        self.assertEqual(split.job.status, Job.PROCESSING)
        self.assertEqual(split.job.block.status, Block.PROCESSING)
        self.assertEqual(split.job.block.experiment.status, Experiment.RUNNING)
        self.assertEqual(split.job.child.status, Job.PROCESSING)
        self.assertEqual(split.job.child.block.status, Block.PROCESSING)
        self.assertEqual(split.job.child.block.experiment.status, Experiment.RUNNING)

        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # no job can be run right now
        assigned_splits = schedule()
        self.assertEqual(len(assigned_splits), 0)

        # simulate end job signal
        split.end(Result(status=0))
        self.assertEqual(split.job.status, Job.COMPLETED)
        self.assertEqual(split.job.block.status, Block.CACHED)
        self.assertEqual(split.job.block.experiment.status, Experiment.RUNNING)
        self.assertEqual(split.job.child.status, Job.COMPLETED)
        self.assertEqual(split.job.child.block.status, Block.CACHED)
        self.assertEqual(split.job.child.block.experiment.status,
            Experiment.RUNNING)

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        self.check_stats_success(split)

        # assert we have no database traces after the block is done
        self.assertEqual(Job.objects.filter(block=split.job.block).count(), 0)
        self.assertEqual(
            Job.objects.filter(block=split.job.child.block).count(), 0)
        self.assertEqual(JobSplit.objects.filter(job=split.job.child).count(),
            0)
        self.assertEqual(Result.objects.filter(job__isnull=True).count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)

        # since this job was successful, the next one should be ready to run

        # schedules the last block of the experiment
        assert xp.blocks.last().job.runnable_date is not None
        assigned_splits = schedule()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'analysis')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # simulate job start on worker
        split.start()
        self.assertEqual(split.job.status, Job.PROCESSING)
        self.assertEqual(split.job.block.status, Block.PROCESSING)
        self.assertEqual(split.job.block.experiment.status, Experiment.RUNNING)
        self.assertEqual(split.job.child.status, Job.PROCESSING)
        self.assertEqual(split.job.child.block.status, Block.PROCESSING)
        self.assertEqual(split.job.child.block.experiment.status,
            Experiment.RUNNING)

        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # no job can be run right now
        assigned_splits = schedule()
        self.assertEqual(len(assigned_splits), 0)

        # simulate end job signal
        split.end(Result(status=0))

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        self.assertEqual(split.job.status, Job.COMPLETED)
        self.assertEqual(split.job.block.status, Block.CACHED)
        self.assertEqual(split.job.block.experiment.status, Experiment.DONE)
        self.assertEqual(split.job.child.status, Job.COMPLETED)
        self.assertEqual(split.job.child.block.status, Block.CACHED)
        self.assertEqual(split.job.child.block.experiment.status,
            Experiment.DONE)

        self.check_stats_success(split)

        # assert we have no database traces after the last block is done
        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)
        self.assertEqual(Result.objects.count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)


    def test_blocking_failure(self):

        # tests two experiments that are similar can be scheduled at the same
        # time and we'll optimise correctly and only run one of them. If the
        # blocking experiment fails, so does the blocked one too.

        current_stats = HourlyStatistics.objects.count()

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        xpc = xp.fork(name='single_copy')

        # schedules the experiment and check it
        xp.schedule()
        xpc.schedule()

        # schedules the first runnable block
        assert xp.blocks.first().job.runnable_date is not None
        assert xp.blocks.last().job.runnable_date is None
        assert xpc.blocks.first().job.runnable_date is None
        assert xpc.blocks.last().job.runnable_date is None

        assigned_splits = schedule()

        worker = Worker.objects.get()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'echo')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # checks the jobs are connected one to the other across experiments
        self.assertEqual(xp.blocks.first().job.child.block.experiment, xpc)
        self.assertEqual(xp.blocks.last().job.child.block.experiment, xpc)

        # simulate job start on worker
        split.start()
        self.assertEqual(split.job.status, Job.PROCESSING)
        self.assertEqual(split.job.block.status, Block.PROCESSING)
        self.assertEqual(split.job.block.experiment.status, Experiment.RUNNING)
        self.assertEqual(split.job.child.status, Job.PROCESSING)
        self.assertEqual(split.job.child.block.status, Block.PROCESSING)
        self.assertEqual(split.job.child.block.experiment.status, Experiment.RUNNING)

        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # no job can be run right now
        assigned_splits = schedule()
        self.assertEqual(len(assigned_splits), 0)

        # simulate end job signal
        split.end(Result(status=1))
        self.assertEqual(split.job.status, Job.FAILED)
        self.assertEqual(split.job.block.status, Block.FAILED)
        split.job.block.experiment.refresh_from_db()
        self.assertEqual(split.job.block.experiment.status, Experiment.FAILED)
        self.assertEqual(split.job.child.status, Job.FAILED)
        self.assertEqual(split.job.child.block.status, Block.FAILED)
        split.job.child.block.experiment.refresh_from_db()
        self.assertEqual(split.job.child.block.experiment.status,
            Experiment.FAILED)

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        # assert we have no database traces after the last block is done
        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)
        self.assertEqual(Result.objects.count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)


    def test_blocking_cancel_after_success(self):

        # tests two experiments that are similar can be scheduled at the same
        # time and we'll optimise correctly and only run one of them. If the
        # first experiment is cancelled, then the second one proceeds normally.

        current_stats = HourlyStatistics.objects.count()

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        xpc = xp.fork(name='single_copy')

        # schedules the experiment and check it
        xp.schedule()
        xpc.schedule()

        # schedules the first runnable block
        assert xp.blocks.first().job.runnable_date is not None
        assert xp.blocks.last().job.runnable_date is None
        assert xpc.blocks.first().job.runnable_date is None
        assert xpc.blocks.last().job.runnable_date is None

        assigned_splits = schedule()

        worker = Worker.objects.get()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'echo')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # checks the jobs are connected one to the other across experiments
        self.assertEqual(xp.blocks.first().job.child.block.experiment, xpc)
        self.assertEqual(xp.blocks.last().job.child.block.experiment, xpc)

        # simulate job start on worker
        split.start()
        self.assertEqual(split.job.status, Job.PROCESSING)
        self.assertEqual(split.job.block.status, Block.PROCESSING)
        self.assertEqual(split.job.block.experiment.status, Experiment.RUNNING)
        self.assertEqual(split.job.child.status, Job.PROCESSING)
        self.assertEqual(split.job.child.block.status, Block.PROCESSING)
        self.assertEqual(split.job.child.block.experiment.status, Experiment.RUNNING)

        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # no job can be run right now
        assigned_splits = schedule()
        self.assertEqual(len(assigned_splits), 0)

        # simulate end job signal
        split.end(Result(status=0))
        self.assertEqual(split.job.status, Job.COMPLETED)
        self.assertEqual(split.job.block.status, Block.CACHED)
        self.assertEqual(split.job.block.experiment.status, Experiment.RUNNING)
        self.assertEqual(split.job.child.status, Job.COMPLETED)
        self.assertEqual(split.job.child.block.status, Block.CACHED)
        self.assertEqual(split.job.child.block.experiment.status,
            Experiment.RUNNING)

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        self.check_stats_success(split)

        # assert we have no database traces after the block is done
        self.assertEqual(Job.objects.filter(block=split.job.block).count(), 0)
        self.assertEqual(
            Job.objects.filter(block=split.job.child.block).count(), 0)
        self.assertEqual(JobSplit.objects.filter(job=split.job.child).count(),
            0)
        self.assertEqual(Result.objects.filter(job__isnull=True).count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)

        # cancels the blocking experiment - the blocked one must continue
        xp.cancel()
        self.assertEqual(
            [str(k) for k in xp.blocks.order_by('id').values_list('status', flat=True)],
            [Block.CACHED, Block.CANCELLED]
            )
        self.assertEqual(xp.status, Experiment.FAILED)

        # assert we have no database traces after the last block is done
        self.assertEqual(Job.objects.filter(block__in=xp.blocks.all()).count(), 0)
        self.assertEqual(JobSplit.objects.filter(job__block__in=xp.blocks.all()).count(), 0)
        self.assertEqual(Result.objects.count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)

        # since the first job was successful, the second block of the
        # previously blocked experiment must be ready to run

        # schedules the last block of the experiment
        assert xpc.blocks.last().job.runnable_date is not None
        assigned_splits = schedule()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xpc)
        self.assertEqual(split.job.block.name, 'analysis')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # the rest would continue normally


    def test_blocking_cancel_while_running(self):

        # tests two experiments that are similar can be scheduled at the same
        # time and we'll optimise correctly and only run one of them. If the
        # first experiment is cancelled while one of the blocks is running,
        # then the second one proceeds normally.

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        xpc = xp.fork(name='single_copy')

        # schedules the experiment and check it
        xp.schedule()
        xpc.schedule()

        # schedules the first runnable block
        assert xp.blocks.first().job.runnable_date is not None
        assert xp.blocks.last().job.runnable_date is None
        assert xpc.blocks.first().job.runnable_date is None
        assert xpc.blocks.last().job.runnable_date is None

        assigned_splits = schedule()

        worker = Worker.objects.get()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'echo')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # checks the jobs are connected one to the other across experiments
        self.assertEqual(xp.blocks.first().job.child.block.experiment, xpc)
        self.assertEqual(xp.blocks.last().job.child.block.experiment, xpc)

        # simulate job start on worker
        split.start()
        self.assertEqual(split.job.status, Job.PROCESSING)
        self.assertEqual(split.job.block.status, Block.PROCESSING)
        self.assertEqual(split.job.block.experiment.status, Experiment.RUNNING)
        self.assertEqual(split.job.child.status, Job.PROCESSING)
        self.assertEqual(split.job.child.block.status, Block.PROCESSING)
        self.assertEqual(split.job.child.block.experiment.status, Experiment.RUNNING)

        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # no job can be run right now
        assigned_splits = schedule()
        self.assertEqual(len(assigned_splits), 0)

        # cancels the blocking experiment - the blocked one must continue
        xp.cancel()

        # simulate worker cancelling
        split.refresh_from_db()
        self.assertEqual(split.status, Job.CANCEL)
        split.end(None, Job.CANCELLED)

        xp.refresh_from_db()
        self.assertEqual(
            [str(k) for k in xp.blocks.order_by('id').values_list('status', flat=True)],
            [Block.CANCELLED, Block.CANCELLED]
            )
        self.assertEqual(xp.status, Experiment.FAILED)

        # assert we have no database traces after the last block is cancel
        self.assertEqual(Job.objects.filter(block__in=xp.blocks.all()).count(), 0)
        self.assertEqual(JobSplit.objects.filter(job__block__in=xp.blocks.all()).count(), 0)
        self.assertEqual(Result.objects.count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)

        # since the first job was successful, the second block of the
        # previously blocked experiment must be ready to run

        # schedules the last block of the experiment
        assert xpc.blocks.first().job.runnable_date is not None
        assigned_splits = schedule()

        assigned_splits = JobSplit.objects.filter(worker__isnull=False)
        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits.first()
        self.assertEqual(split.job.block.experiment, xpc)
        self.assertEqual(split.job.block.name, 'echo')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # the rest would continue normally


    def test_blocking_cancel_blocked(self):

        # tests two experiments that are similar can be scheduled at the same
        # time and we'll optimise correctly and only run one of them. If the
        # blocked experiment is cancelled, this does not affect the running
        # experiment.

        current_stats = HourlyStatistics.objects.count()

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        xpc = xp.fork(name='single_copy')

        # schedules the experiment and check it
        xp.schedule()
        xpc.schedule()

        # schedules the first runnable block
        assert xp.blocks.first().job.runnable_date is not None
        assert xp.blocks.last().job.runnable_date is None
        assert xpc.blocks.first().job.runnable_date is None
        assert xpc.blocks.last().job.runnable_date is None

        assigned_splits = schedule()

        worker = Worker.objects.get()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'echo')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # checks the jobs are connected one to the other across experiments
        self.assertEqual(xp.blocks.first().job.child.block.experiment, xpc)
        self.assertEqual(xp.blocks.last().job.child.block.experiment, xpc)

        # simulate job start on worker
        split.start()
        self.assertEqual(split.job.status, Job.PROCESSING)
        self.assertEqual(split.job.block.status, Block.PROCESSING)
        self.assertEqual(split.job.block.experiment.status, Experiment.RUNNING)
        self.assertEqual(split.job.child.status, Job.PROCESSING)
        self.assertEqual(split.job.child.block.status, Block.PROCESSING)
        self.assertEqual(split.job.child.block.experiment.status, Experiment.RUNNING)

        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # no job can be run right now
        assigned_splits = schedule()
        self.assertEqual(len(assigned_splits), 0)

        # simulate end job signal
        split.end(Result(status=0))
        self.assertEqual(split.job.status, Job.COMPLETED)
        self.assertEqual(split.job.block.status, Block.CACHED)
        self.assertEqual(split.job.block.experiment.status, Experiment.RUNNING)
        self.assertEqual(split.job.child.status, Job.COMPLETED)
        self.assertEqual(split.job.child.block.status, Block.CACHED)
        self.assertEqual(split.job.child.block.experiment.status,
            Experiment.RUNNING)

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        self.check_stats_success(split)

        # cancels the blocked experiment - the blocking one must continue
        xpc.cancel()
        self.assertEqual(
            [str(k) for k in xpc.blocks.order_by('id').values_list('status',
              flat=True)],
            [Block.CACHED, Block.CANCELLED]
            )
        self.assertEqual(xpc.status, Experiment.FAILED)

        # assert we have no database traces after the last cancel
        self.assertEqual(Job.objects.filter(block__in=xpc.blocks.all()).count(), 0)
        self.assertEqual(JobSplit.objects.filter(job__block__in=xpc.blocks.all()).count(), 0)
        self.assertEqual(Result.objects.count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)

        # since the first job was successful, the second block of the
        # running experiment must be ready to run

        # schedules the last block of the experiment
        assert xp.blocks.last().job.runnable_date is not None
        assigned_splits = schedule()

        assigned_splits = JobSplit.objects.filter(worker__isnull=False)
        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits.first()
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'analysis')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # the rest would continue normally


    def test_schedule_without_queue(self):

        # tests that an experiment with a queue that disappeared is correctly
        # aborted
        setup_backend(QUEUES_WITHOUT_PRIORITY)

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        self.assertRaises(RuntimeError, xp.schedule)


    def test_split_no_index(self):

        # tests a simple experiment with splitting and show it can fail
        # gracefully

        current_stats = HourlyStatistics.objects.count()

        fullname = 'user/user/single/1/single_large'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        worker = Worker.objects.get()

        # schedules the experiment and check it
        xp.schedule()
        xp.refresh_from_db()
        self.assertEqual(xp.status, Experiment.FAILED)

        self.assertEqual(xp.blocks.first().status, Block.CANCELLED)
        assert xp.blocks.first().error_report().find(settings.DEFAULT_USER_ERROR) == 0
        self.assertEqual(xp.blocks.last().status, Block.CANCELLED)

        # assert we have no database traces after the last block is done
        self.assertEqual(Job.objects.filter(block__in=xp.blocks.all()).count(), 0)
        self.assertEqual(JobSplit.objects.filter(job__block__in=xp.blocks.all()).count(), 0)
        self.assertEqual(Result.objects.count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)


    def test_schedules_two_jobs(self):

        # tests a simple scheduling activity in which two jobs of the same
        # experiment must be scheduled concurrently, provided there is enough
        # space

        current_stats = HourlyStatistics.objects.count()

        fullname = 'user/user/triangle/1/triangle'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        worker = Worker.objects.get()

        # schedules the experiment and check it
        xp.schedule()
        xp.refresh_from_db()
        self.assertEqual(xp.status, Experiment.SCHEDULED)

        assigned_splits = schedule()

        self.assertEqual(len(assigned_splits), 2)
        self.assertEqual(assigned_splits[0].job.block.experiment, xp)
        self.assertEqual(assigned_splits[1].job.block.experiment, xp)
        self.assertNotEqual(assigned_splits[0], assigned_splits[1])


    def test_cancel_concurrent_job(self):

        # tests a simple scheduling activity in which two jobs of the same
        # experiment must be scheduled concurrently, provided there is enough
        # space. Then, fails one of them, waits the experiment to fail
        # completely. Processing jobs must be cancelled.

        current_stats = HourlyStatistics.objects.count()

        fullname = 'user/user/triangle/1/triangle'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        worker = Worker.objects.get()

        # schedules the experiment and check it
        xp.schedule()
        xp.refresh_from_db()
        self.assertEqual(xp.status, Experiment.SCHEDULED)

        assigned_splits = schedule()

        self.assertEqual(len(assigned_splits), 2)
        self.assertEqual(assigned_splits[0].job.block.experiment, xp)
        self.assertEqual(assigned_splits[1].job.block.experiment, xp)
        self.assertNotEqual(assigned_splits[0], assigned_splits[1])


        # simulate job start on worker
        assigned_splits[0].start()
        assigned_splits[1].start()

        # now fail one of the jobs, the end result is the experiment fails
        assigned_splits[1].end(Result(status=15)) #simulated sigterm sent
        self.assertEqual(assigned_splits[1].job.status, Job.FAILED)

        # cancels the job which is marked for cancelling, checks the final
        # experiment state is as expected (this is the worker job)
        self.assertEqual(assigned_splits[0].job.splits.first().status,
            Job.CANCEL)
        assigned_splits[0].job.splits.first().end(None, Job.CANCELLED)

        xp.refresh_from_db()
        self.assertEqual(
            [str(k) for k in xp.blocks.order_by('id').values_list('status',
              flat=True)],
            [Block.CANCELLED, Block.FAILED, Block.CANCELLED, Block.CANCELLED]
            )
        self.assertEqual(xp.status, Experiment.FAILED)



class SchedulingPriority(BaseBackendTestCase):


    def set_globals(self, xp, queue, environment):
        '''Sets the global queue of the experiment'''

        decl = xp.declaration
        decl['globals']['queue'] = queue.name
        decl['globals']['environment']['name'] = environment.name
        decl['globals']['environment']['version'] = environment.version
        xp.declaration = decl
        xp.save() #reloads all blocks


    def reset_slots(self, xp):
        '''Only use one slot in all blocks'''

        decl = xp.declaration
        for b in decl['blocks']:
            if 'nb_slots' in decl['blocks'][b]:
                del decl['blocks'][b]['nb_slots']
        xp.declaration = decl
        xp.save() #reloads all blocks


    def test_priority_multicore(self):

        # tests that in an heterogeneous backend setup, priority is given to
        # jobs that require more cores correctly.

        setup_backend(QUEUES_WITHOUT_PRIORITY)
        Worker.objects.update(active=True)

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])
        fullname = 'user/user/single/1/single_add'
        xp_add = Experiment.objects.get(name=fullname.split(os.sep)[-1])
        fullname = 'user/user/single/1/single_add2'
        xp_add2 = Experiment.objects.get(name=fullname.split(os.sep)[-1])
        fullname = 'user/user/single/1/single_large'
        xp_large = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        q1 = Queue.objects.get(name='q1')
        q2 = Queue.objects.get(name='q2')
        env = Environment.objects.get(name='environment')

        # reset queue and environment to new backend configuration
        self.set_globals(xp, q1, env)
        self.set_globals(xp_add, q1, env)
        self.set_globals(xp_add2, q1, env)
        self.set_globals(xp_large, q2, env) #notice different queue
        self.reset_slots(xp_large) #one slot per block only

        xp.schedule()
        xp_add.schedule()
        xp_add2.schedule()
        xp_large.schedule()

        assigned_splits = schedule()

        self.assertEqual(len(assigned_splits), 3)

        self.assertEqual(assigned_splits[0].job.block.experiment, xp_large)
        # then, the scheduling order is respected
        self.assertEqual(assigned_splits[1].job.block.experiment, xp)
        self.assertEqual(assigned_splits[2].job.block.experiment, xp_add)
        # notice that the last experiment is not assigned


    def test_priority_multicore_delayed(self):

        # tests that in an heterogeneous backend setup, priority is given to
        # jobs that require more cores correctly. In this test, specifically,
        # we verify that, if the farm is taken, new jobs that require more
        # resources will block other possible jobs to run if they cannot run,
        # even if free cores are available.

        setup_backend(QUEUES_WITHOUT_PRIORITY)
        Worker.objects.update(active=True)

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])
        fullname = 'user/user/single/1/single_add'
        xp_add = Experiment.objects.get(name=fullname.split(os.sep)[-1])
        fullname = 'user/user/single/1/single_add2'
        xp_add2 = Experiment.objects.get(name=fullname.split(os.sep)[-1])
        fullname = 'user/user/single/1/single_large'
        xp_large = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        q1 = Queue.objects.get(name='q1')
        q4 = Queue.objects.get(name='q4')
        env = Environment.objects.get(name='environment')

        # reset queue and environment to new backend configuration
        self.set_globals(xp, q1, env)
        self.set_globals(xp_add, q1, env)
        self.set_globals(xp_large, q4, env) #notice different queue
        self.reset_slots(xp_large) #one slot per block only

        xp.schedule()
        assigned_splits = schedule()

        self.assertEqual(len(assigned_splits), 1)
        self.assertEqual(assigned_splits[0].job.block.experiment, xp)
        split = assigned_splits[0]

        xp_large.schedule() #will now block anything else from running
        xp_add.schedule()

        assigned_splits = schedule()
        self.assertEqual(len(assigned_splits), 0)

        # start/end the xp block and schedule again
        split.start()
        split.end(Result(0))

        # now, the job with more cores should be scheduled first
        assigned_splits = schedule()
        self.assertEqual(len(assigned_splits), 1)
        self.assertEqual(assigned_splits[0].job.block.experiment, xp_large)


    def test_priorities(self):

        # tests that in an heterogeneous backend setup, priority is given to
        # different computers based on their priority settings

        setup_backend(PRIORITY_QUEUES)
        Worker.objects.update(active=True)

        q1 = Queue.objects.get(name='q1')
        env = Environment.objects.get(name='environment')

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])
        self.set_globals(xp, q1, env)

        q1_special = Queue.objects.get(name='q1_special')
        fullname = 'user/user/single/1/single_add'
        xp_add = Experiment.objects.get(name=fullname.split(os.sep)[-1])
        self.set_globals(xp_add, q1_special, env)

        q2 = Queue.objects.get(name='q2')
        fullname = 'user/user/single/1/single_large'
        xp_large = Experiment.objects.get(name=fullname.split(os.sep)[-1])
        self.set_globals(xp_large, q2, env) #notice different queue
        self.reset_slots(xp_large) #one slot per block only

        node1 = Worker.objects.get(name='node1')
        node2 = Worker.objects.get(name='node2')

        # verify that xp_large has priority, other jobs corresponding to
        # q1/_special
        xp.schedule()
        xp_add.schedule()
        xp_large.schedule()

        assigned_splits = schedule()
        self.assertTrue(len(assigned_splits), 3)

        self.assertEqual(assigned_splits[0].job.block.experiment, xp_large)
        self.assertEqual(assigned_splits[0].job.block.name, 'echo')
        self.assertEqual(assigned_splits[0].job.splits.first().worker, node2)

        self.assertEqual(assigned_splits[1].job.block.experiment, xp)
        self.assertEqual(assigned_splits[1].job.block.name, 'echo')
        self.assertEqual(assigned_splits[1].job.splits.first().worker, node1)

        self.assertEqual(assigned_splits[2].job.block.experiment, xp_add)
        self.assertEqual(assigned_splits[2].job.block.name, 'echo')
        self.assertEqual(assigned_splits[2].job.splits.first().worker, node2)


class Working(BaseBackendTestCase):


    def setUp(self):

        from . import utils
        self.process = utils.resolve_process_path()
        self.environments = utils.find_environments(None)

        if not os.path.exists(settings.CACHE_ROOT):
            os.makedirs(settings.CACHE_ROOT)


    def tearDown(self):
        if os.path.exists(settings.CACHE_ROOT):
            shutil.rmtree(settings.CACHE_ROOT)


    def check_stats_success(self, block):

        assert abs(block.speed_up_real() - 1.0) < 0.1
        assert abs(block.speed_up_maximal() - 1.0) < 0.1
        assert block.linear_execution_time() > 0.0
        assert block.queuing_time() > 0.0
        assert block.stdout() == ''
        assert block.stderr() == ''
        assert block.error_report() == ''


    def test_success(self):

        # tests an experiment can actually be run

        current_stats = HourlyStatistics.objects.count()

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        # schedules the experiment and check it
        xp.schedule()

        # schedules the first runnable block
        assert xp.blocks.first().job.runnable_date is not None
        assert xp.blocks.last().job.runnable_date is None

        assigned_splits = schedule()

        worker = Worker.objects.get()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'echo')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # actually runs the job (blocking)
        split.process()

        # at this point, job should have been successful
        xp.refresh_from_db()
        block = xp.blocks.first()
        self.assertEqual(block.status, Block.CACHED)
        self.assertEqual(xp.status, Experiment.RUNNING)

        # all caches must be have been generated
        assert all([k.index_checksums() for k in block.outputs.all()])

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        self.check_stats_success(block)

        # assert we have no database traces after the block is done
        self.assertEqual(Job.objects.filter(block=split.job.block).count(), 0)
        self.assertEqual(JobSplit.objects.filter(job=split.job).count(), 0)
        self.assertEqual(Result.objects.filter(job__isnull=True).count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)

        # since this job was successful, the next one should be ready to run

        # schedules the last block of the experiment
        assert xp.blocks.last().job.runnable_date is not None
        assigned_splits = schedule()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'analysis')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # actually runs the job (blocking)
        split.process()

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        xp.refresh_from_db()
        block = xp.blocks.last()
        self.assertEqual(block.status, Block.CACHED)
        self.assertEqual(xp.status, Experiment.DONE)

        # all caches must be have been generated
        assert all([k.index_checksums() for k in block.outputs.all()])

        self.check_stats_success(block)

        # assert we have no database traces after the last block is done
        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)
        self.assertEqual(Result.objects.count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)


    def test_failure(self):

        # tests an experiment can fail and we can handle it fine

        current_stats = HourlyStatistics.objects.count()

        fullname = 'user/user/single/1/single_error'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        # schedules the experiment and check it
        xp.schedule()

        # schedules the first runnable block
        assert xp.blocks.first().job.runnable_date is not None
        assert xp.blocks.last().job.runnable_date is None

        assigned_splits = schedule()

        worker = Worker.objects.get()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'echo')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # actually runs the job (blocking)
        split.process()

        # at this point, job should have failed
        xp.refresh_from_db()
        block = xp.blocks.first()
        self.assertEqual(block.status, Block.FAILED)
        self.assertEqual(block.experiment.status, Experiment.FAILED)

        # all caches have not been generated
        assert all([not k.exists() for k in block.outputs.all()])

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        assert abs(block.speed_up_real() - 1.0) < 0.1
        assert abs(block.speed_up_maximal() - 1.0) < 0.1
        assert block.linear_execution_time() > 0.0
        assert block.queuing_time() > 0.0
        assert block.stdout() == ''
        assert block.error_report().find('Error') != -1

        # assert we have no database traces after the block is done
        self.assertEqual(Job.objects.filter(block=split.job.block).count(), 0)
        self.assertEqual(JobSplit.objects.filter(job=split.job).count(), 0)
        self.assertEqual(Result.objects.filter(job__isnull=True).count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)


    def test_skip(self):

        # tests an experiment can actually be completely skipped if all files
        # are already cached

        current_stats = HourlyStatistics.objects.count()

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        # schedules the experiment and check it
        xp.schedule()

        # schedules the first runnable block
        assert xp.blocks.first().job.runnable_date is not None
        assert xp.blocks.last().job.runnable_date is None

        assigned_splits = schedule()

        worker = Worker.objects.get()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'echo')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # actually runs the job (blocking)
        split.process()

        # at this point, job should have been successful
        xp.refresh_from_db()
        block = xp.blocks.first()
        self.assertEqual(block.status, Block.CACHED)
        self.assertEqual(block.experiment.status, Experiment.RUNNING)

        # all caches must be have been generated
        assert all([k.index_checksums() for k in block.outputs.all()])

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        self.check_stats_success(block)

        # assert we have no database traces after the block is done
        self.assertEqual(Job.objects.filter(block=split.job.block).count(), 0)
        self.assertEqual(JobSplit.objects.filter(job=split.job).count(), 0)
        self.assertEqual(Result.objects.filter(job__isnull=True).count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)

        # since this job was successful, the next one should be ready to run

        # schedules the last block of the experiment
        assert xp.blocks.last().job.runnable_date is not None
        assigned_splits = schedule()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'analysis')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # actually runs the job (blocking)
        split.process()

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        xp.refresh_from_db()
        block = xp.blocks.last()
        self.assertEqual(block.status, Block.CACHED)
        self.assertEqual(block.experiment.status, Experiment.DONE)

        # all caches must be have been generated
        assert all([k.index_checksums() for k in block.outputs.all()])

        self.check_stats_success(block)

        # assert we have no database traces after the last block is done
        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)
        self.assertEqual(Result.objects.count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)

        # now we fork and re-run the same experiment
        xpc = xp.fork(name='single_copy')

        # schedules the experiment (it should immediately load from the db)
        xpc.schedule()

        self.assertEqual(xpc.status, Experiment.DONE)


    def test_does_not_skip(self):

        # tests an experiment can actually be partially skipped if some files
        # are ready

        current_stats = HourlyStatistics.objects.count()

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        # schedules the experiment and check it
        xp.schedule()

        # schedules the first runnable block
        assert xp.blocks.first().job.runnable_date is not None
        assert xp.blocks.last().job.runnable_date is None

        assigned_splits = schedule()

        worker = Worker.objects.get()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'echo')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # actually runs the job (blocking)
        split.process()

        # at this point, job should have been successful
        xp.refresh_from_db()
        block = xp.blocks.first()
        self.assertEqual(block.status, Block.CACHED)
        self.assertEqual(block.experiment.status, Experiment.RUNNING)

        # all caches must be have been generated
        assert all([k.index_checksums() for k in block.outputs.all()])

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        self.check_stats_success(block)

        # assert we have no database traces after the block is done
        self.assertEqual(Job.objects.filter(block=split.job.block).count(), 0)
        self.assertEqual(JobSplit.objects.filter(job=split.job).count(), 0)
        self.assertEqual(Result.objects.filter(job__isnull=True).count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)

        # now we cancel the experiment
        xp.cancel()

        # we fork it and re-run it - only the last block will run again
        xpc = xp.fork(name='single_copy')
        xpc.schedule()

        # schedules the first runnable block
        assert not hasattr(xpc.blocks.first(), 'job')
        assert xpc.blocks.first().status == Block.CACHED

        # since this job was successful, the next one should be ready to run

        # schedules the last block of the experiment
        assert xpc.blocks.last().job.runnable_date is not None
        assigned_splits = schedule()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xpc)
        self.assertEqual(split.job.block.name, 'analysis')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # actually runs the job (blocking)
        split.process()

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        xpc.refresh_from_db()
        block = xpc.blocks.last()
        self.assertEqual(block.status, Block.CACHED)
        self.assertEqual(block.experiment.status, Experiment.DONE)

        # all caches must be have been generated
        assert all([k.index_checksums() for k in block.outputs.all()])

        self.check_stats_success(block)

        # assert we have no database traces after the last block is done
        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)
        self.assertEqual(Result.objects.count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)

        # asserts the old experiment is still on a failed state
        self.assertEqual(
            [str(k) for k in xp.blocks.order_by('id').values_list('status',
              flat=True)],
            [Block.CACHED, Block.CANCELLED]
            )
        self.assertEqual(xp.status, Experiment.FAILED)


    def test_partially_blocking(self):

        # tests an experiment can actually be partially skipped if some files
        # are ready on the cache - blocking occurs as foreseeable on blocks to
        # run

        current_stats = HourlyStatistics.objects.count()

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        # schedules the experiment and check it
        xp.schedule()

        # schedules the first runnable block
        assert xp.blocks.first().job.runnable_date is not None
        assert xp.blocks.last().job.runnable_date is None

        assigned_splits = schedule()

        worker = Worker.objects.get()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'echo')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # actually runs the job (blocking)
        split.process()

        # at this point, job should have been successful
        xp.refresh_from_db()
        block = xp.blocks.first()
        self.assertEqual(block.status, Block.CACHED)
        self.assertEqual(block.experiment.status, Experiment.RUNNING)

        # all caches must be have been generated
        assert all([k.index_checksums() for k in block.outputs.all()])

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        self.check_stats_success(block)

        # assert we have no database traces after the block is done
        self.assertEqual(Job.objects.filter(block=split.job.block).count(), 0)
        self.assertEqual(JobSplit.objects.filter(job=split.job).count(), 0)
        self.assertEqual(Result.objects.filter(job__isnull=True).count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)

        # now we fork and wait the second experiment to hook-into the last
        # block
        xpc = xp.fork(name='single_copy')
        xpc.schedule()

        self.assertEqual([k.status for k in xpc.blocks.all()],
            [Block.CACHED, Block.NOT_CACHED])
        assert xpc.blocks.last().job.parent == xp.blocks.last().job


class WorkingExternally(TransactionTestCase):


    def setUp(self):

        from . import utils
        self.process = utils.resolve_process_path()
        self.environments = utils.find_environments(None)

        if not os.path.exists(settings.CACHE_ROOT):
            os.makedirs(settings.CACHE_ROOT)

        install.create_sites()
        system_user, plot_user, user = install.create_users('user', 'user')
        install.add_group('Default')

        setup_backend(qsetup.DEFAULT_CONFIGURATION)

        Worker.objects.update(active=True)
        env = Environment.objects.get(name='environment')
        queue = Queue.objects.first()

        template_data = dict(
            system_user = system_user,
            plot_user = plot_user,
            user = user,
            private = False,
            queue = queue.name,
            environment = dict(name=env.name, version=env.version),
            )
        prefix = os.path.join(
              os.path.dirname(os.path.dirname(os.path.realpath(sys.argv[0]))),
              'src',
              'beat.examples',
              )
        install.install_contributions(prefix, 'system', template_data)
        install.install_contributions(prefix, 'test', template_data)


    def tearDown(self):
        if os.path.exists(settings.CACHE_ROOT):
            shutil.rmtree(settings.CACHE_ROOT)
        if os.path.exists(settings.PREFIX):
            shutil.rmtree(settings.PREFIX)


    def test_success(self):

        # tests an experiment can actually be run

        current_stats = HourlyStatistics.objects.count()

        fullname = 'user/user/single/1/single'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        # schedules the experiment and check it
        xp.schedule()

        # schedules the first runnable block
        assert xp.blocks.first().job.runnable_date is not None
        assert xp.blocks.last().job.runnable_date is None

        assigned_splits = schedule()

        worker = Worker.objects.get()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'echo')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # actually runs the job (non-blocking)
        worker.work(self.environments, self.process)

        def condition():
            xp.refresh_from_db()
            block = xp.blocks.first()
            return block.status == Block.CACHED
        _sleep(20, condition)

        # at this point, split should have been successful which shall
        # trigger job deletion and block update
        xp.refresh_from_db()
        block = xp.blocks.first()

        self.assertEqual(block.status, Block.CACHED)
        self.assertEqual(xp.status, Experiment.RUNNING)

        # all caches must be have been generated
        assert all([k.index_checksums() for k in split.job.block.outputs.all()])

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        # assert we have no database traces after the block is done
        self.assertEqual(Job.objects.filter(block=split.job.block).count(), 0)
        self.assertEqual(JobSplit.objects.filter(job=split.job).count(), 0)
        self.assertEqual(Result.objects.filter(job__isnull=True).count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)

        # since this job was successful, the next one should be ready to run

        # schedules the last block of the experiment
        assert xp.blocks.last().job.runnable_date is not None
        assigned_splits = schedule()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'analysis')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # actually runs the job (non-blocking)
        worker.work(self.environments, self.process)

        def condition():
            xp.refresh_from_db()
            return xp.status == Experiment.DONE
        _sleep(20, condition) #wait job completion

        # checks the number of statistics objects has increased by 1
        self.assertEqual(HourlyStatistics.objects.count(), current_stats + 1)

        # at this point, split should have been successful which shall
        # trigger job deletion and block update
        xp.refresh_from_db()
        block = xp.blocks.last()

        self.assertEqual(block.status, Block.CACHED)
        self.assertEqual(xp.status, Experiment.DONE)

        # all caches must be have been generated
        assert all([k.index_checksums() for k in split.job.block.outputs.all()])

        # assert we have no database traces after the last block is done
        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)
        self.assertEqual(Result.objects.count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)


    def test_cancel_running(self):

        # tests an experiment can be cancelled while running

        current_stats = HourlyStatistics.objects.count()

        fullname = 'user/user/single/1/single_sleep'
        xp = Experiment.objects.get(name=fullname.split(os.sep)[-1])

        # schedules the experiment and check it
        xp.schedule()

        # schedules the first runnable block
        assert xp.blocks.first().job.runnable_date is not None
        assert xp.blocks.last().job.runnable_date is None

        assigned_splits = schedule()

        worker = Worker.objects.get()

        self.assertEqual(len(assigned_splits), 1)
        split = assigned_splits[0]
        self.assertEqual(split.job.block.experiment, xp)
        self.assertEqual(split.job.block.name, 'echo')
        self.assertEqual(split.worker, worker)
        self.assertEqual(worker.name, qsetup.HOSTNAME)
        self.assertEqual(worker.available_cores(), qsetup.CORES-1)

        # actually runs the job (non-blocking)
        worker.work(self.environments, self.process)

        def condition():
            xp.refresh_from_db()
            return xp.status == Experiment.RUNNING

        _sleep(20, condition)

        # Just to be sure that the docker container really started
        time.sleep(3)

        # cancels the experiment
        xp.cancel()
        split.refresh_from_db()
        self.assertEqual(split.status, Job.CANCEL)

        # launch another working cycle to kill the process
        worker.work(self.environments, self.process)

        def condition():
            xp.refresh_from_db()
            return xp.status == Experiment.FAILED and Job.objects.count() == 0
        _sleep(20, condition)
        xp.refresh_from_db()

        # assert we have no database traces after the last block is done
        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)
        self.assertEqual(Result.objects.count(), 0)

        self.assertEqual(worker.available_cores(), qsetup.CORES)

        # asserts the old experiment is still on a failed state
        self.assertEqual(
            [str(k) for k in xp.blocks.order_by('id').values_list('status',
              flat=True)],
            [Block.CANCELLED, Block.CANCELLED]
            )
        self.assertEqual(xp.status, Experiment.FAILED)
