#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.db import models
from django.db.models import Q
from django.contrib.auth.models import User

from ..common.models import Contribution
from ..common.models import StoredContribution
from ..common.models import StoredContributionManager
from ..common.exceptions import ShareError
from ..common.signals import shared
from ..common import storage

from ..team.models import Team

from beat.core.utils import extension_for_language

import beat.core.hash

import simplejson


#----------------------------------------------------------


class CodeManager(StoredContributionManager):

    def create_object(self, author, name, short_description='', description='',
                      declaration=None, code=None, version=1, previous_version=None,
                      fork_of=None):

        create = getattr(self, 'create_{}'.format(self.model.__name__.lower()))
        return create(author=author, name=name, short_description=short_description,
                      description=description, declaration=declaration, code=code,
                      version=version, previous_version=previous_version,
                      fork_of=fork_of)

    def for_user(self, user, add_public=False):
        if user.is_anonymous():
            query = Q(sharing=Code.PUBLIC) |\
                    (Q(sharing=Code.USABLE) & (Q(shared_with=None) & Q(shared_with_team=None)))
            return self.filter(query).distinct()

        teams = Team.objects.filter(members=user)
        query = Q(author=user) |\
                Q(sharing=Code.USABLE) |\
                Q(usable_by=user) |\
                Q(usable_by_team__in=teams) |\
                Q(shared_with=user) |\
                Q(shared_with_team__in=teams)

        if add_public:
            query |= Q(sharing=Code.PUBLIC)

        return self.filter(query).distinct()

    def for_team(self, team):
        return self.filter(Q(usable_by_team=team) |
                           Q(shared_with_team=team))

    def from_author(self, user, author_name, add_public=False):
        if user.is_anonymous():
            objects_for_user = self.public().filter(author__username=author_name)
        elif user.username == author_name:
            # Retrieve all the objects of the specified user
            asked_user = User.objects.get(username=author_name)
            objects_for_user = self.for_user(asked_user, add_public)
        else:
            teams = Team.objects.filter(members=user)
            objects_for_user = self.filter(Q(author__username=author_name) &
                                                (Q(sharing=Code.PUBLIC) |
                                                 Q(sharing=Code.USABLE) |
                                                 Q(usable_by=user) |
                                                 Q(usable_by_team__in=teams) |
                                                 Q(shared_with=user) |
                                                 Q(shared_with_team__in=teams))).distinct()

        return objects_for_user.order_by('author__username', 'name', '-version').select_related()

    def create_code(self, author, name, default, short_description='',
                    description='', declaration=None, code=None, version=1,
                    previous_version=None, fork_of=None):

        # Create the database representation
        code_db = self.model(
            author = author,
            name = self.model.sanitize_name(name),
            version = version,
            sharing = Code.PRIVATE,
            previous_version = previous_version,
            fork_of = fork_of,
        )

        # Check the provided declaration
        if declaration is None:
            if previous_version is not None:
                declaration = previous_version.declaration
            elif fork_of is not None:
                declaration = fork_of.declaration
            else:
                declaration = default.data
        elif not(isinstance(declaration, dict)):
            declaration = simplejson.loads(declaration)

        if len(short_description) > 0:
            declaration['description'] = short_description

        code_db.declaration = declaration

        # Figure out the language
        language = declaration.get('language', 'unknown')
        code_db.language = getattr(Code, language.upper())

        # Check the provided source code
        if code is None:
            if previous_version is not None:
                code_db.source_code = previous_version.source_code
            elif fork_of is not None:
                code_db.source_code = fork_of.source_code
            elif code_db.language != Code.CXX:
                code_db.source_code = default.code
        else:
            code_db.source_code = code

        # Check the provided description
        if not(description):
            if previous_version is not None:
                code_db.description = previous_version.description
            elif fork_of is not None:
                code_db.description = fork_of.description
        else:
            code_db.description = description

        try:
            code_db.save()
        except Exception:
            import traceback
            return (None, traceback.format_exc())

        return (code_db, None)


#----------------------------------------------------------


def get_contribution_source_code_filename(obj, path):
    return obj.source_code_filename()


#----------------------------------------------------------


# Use those function to add a 'source_code' property to a model, by doing:
# source_code = property(beat.web.code.models.get_source_code,
#                        beat.web.code.models.set_source_code)

def set_source_code(instance, value):
    storage.set_file_content(instance, 'source_code_file', instance.source_code_filename(), value)


def get_source_code(instance):
    return storage.get_file_content(instance, 'source_code_file')


#----------------------------------------------------------


class Code(StoredContribution):

    #_____ Constants __________

    # All possible values should be in sync with beat.core.utils
    UNKNOWN = 'U'
    CXX     = 'C'
    MATLAB  = 'M'
    PYTHON  = 'P'
    R       = 'R'

    CODE_LANGUAGE = (
        (UNKNOWN, 'Unknown'),
        (CXX,     'Cxx'),
        (MATLAB,  'Matlab'),
        (PYTHON,  'Python'),
        (R,       'R'),
    )

    CODE_NAMES = {
        CXX: 'C++',
    }


    #_____ Fields __________

    # For technical reason, it is not possible to declare the required field here. It
    # must be declared in each subclass of Code, like this:
    #
    # source_code_file = models.FileField(
    #                             storage=AppropriateStorageClass(),
    #                             upload_to=get_contribution_source_code_filename,
    #                             blank=True, null=True,
    #                             max_length=200,
    #                             db_column='source_code'
    #                            )

    usable_by = models.ManyToManyField(User, related_name='usable_%(class)ss',
                                       blank=True)

    usable_by_team = models.ManyToManyField(Team,
        related_name='usable_%(class)ss', blank=True)

    language = models.CharField(max_length=1, choices=CODE_LANGUAGE,
                                default=PYTHON)

    objects = CodeManager()


    #_____ Meta parameters __________

    class Meta(StoredContribution.Meta):
        abstract = True


    #_____ Methods to implement __________

    def validate(self, declaration):
        raise NotImplementedError()


    #_____ Methods __________

    def source_code_filename(self):
        return self.hashed_path(extension_for_language(self.get_language_display().lower()))


    def share(self, public, users=None, teams=None):
        self._share_libraries(public=public, users=users, teams=teams)

        sharing = self.sharing

        if public:
            # Check that the request make sense
            if self.sharing != Contribution.PUBLIC:
                # Accessible to some users
                if users is not None:
                    # Filter out the users that already have this sharing access or better
                    users_with_better_access = [self.author.username]
                    users_with_better_access.extend(map(lambda x: x.username, self.shared_with.iterator()))

                    users = filter(lambda x: x not in users_with_better_access, users)

                    # Remove the sharing preferences of users that already have a lower sharing access
                    db_users = self.usable_by.filter(username__in=users)
                    for db_user in db_users:
                        self.usable_by.remove(db_user)

                    # Update the algorithm entry in the database
                    db_users = User.objects.filter(username__in=users)
                    for db_user in db_users:
                        self.shared_with.add(db_user)

                    if self.sharing == Contribution.PRIVATE:
                        sharing = Contribution.SHARED

                # Accessible to some teams
                if (teams is not None) and (len(teams) > 0):
                    # 'teams' might be a list of names or a list of DB objects
                    if isinstance(teams[0], Team):
                        # Filter out the teams that already have this sharing access or better
                        teams_with_better_access = self.shared_with_team.all()

                        db_teams = filter(lambda x: x not in teams_with_better_access, teams)

                        # Update the code entry in the database
                        for db_team in db_teams:
                            self.usable_by_team.remove(db_team)
                            self.shared_with_team.add(db_team)
                    else:
                        # Filter out the teams that already have this sharing access or better
                        teams_with_better_access = map(lambda x: x.fullname(), self.shared_with_team.iterator())

                        teams = filter(lambda x: x not in teams_with_better_access, teams)

                        # Update the algorithm entry in the database
                        for team_name in teams:
                            parts = team_name.split('/')

                            if len(parts) == 1:
                                parts = [self.author.username, team_name]

                            db_team = Team.objects.filter(owner__username=parts[0], name=parts[1])
                            if len(db_team) == 1:
                                self.usable_by_team.remove(db_team[0])
                                self.shared_with_team.add(db_team[0])

                    if self.sharing == Contribution.PRIVATE:
                        sharing = Contribution.SHARED

                # Accessible to everybody
                if users is None and teams is None:
                    # Update the code entry in the database
                    sharing = Contribution.PUBLIC

        else:
            # Check that the request make sense
            if not self.sharing in [Contribution.PUBLIC, Contribution.USABLE]:
                # Usable by some users
                sharing = None
                if users is not None:
                    # Filter out the users that already have this sharing access or better
                    users_with_better_access = [self.author.username]
                    users_with_better_access.extend(map(lambda x: x.username, self.shared_with.iterator()))
                    users_with_better_access.extend(map(lambda x: x.username, self.usable_by.iterator()))

                    users = filter(lambda x: x not in users_with_better_access, users)

                    # Update the algorithm entry in the database
                    db_users = User.objects.filter(username__in=users)
                    for db_user in db_users:
                        self.usable_by.add(db_user)

                    sharing = Contribution.SHARED

                if (teams is not None) and (len(teams) > 0):
                    # 'teams' might be a list of names or a list of DB objects
                    if isinstance(teams[0], Team):
                        # Filter out the teams that already have this sharing access or better
                        teams_with_better_access = list(self.shared_with_team.all())
                        teams_with_better_access.extend(list(self.usable_by_team.all()))

                        db_teams = filter(lambda x: x not in teams_with_better_access, teams)

                        # Update the code entry in the database
                        for db_team in db_teams:
                            self.usable_by_team.add(db_team)
                    else:
                        # Filter out the teams that already have this sharing access or better
                        teams_with_better_access = map(lambda x: x.fullname(), self.shared_with_team.iterator())
                        teams_with_better_access.extend(map(lambda x: x.fullname(), self.usable_by_team.iterator()))

                        new_teams = filter(lambda x: x not in teams_with_better_access, teams)

                        # Update the algorithm entry in the database
                        for team_name in new_teams:
                            parts = team_name.split('/')

                            if len(parts) == 1:
                                parts = [self.author.username, team_name]

                            db_team = Team.objects.filter(owner__username=parts[0], name=parts[1])
                            if len(db_team) == 1:
                                self.usable_by_team.add(db_team[0])

                    sharing = Contribution.SHARED

                # Usable for everybody
                if users is None and teams is None:
                    sharing = Contribution.USABLE

        self._update_sharing(sharing, users, teams)


    def is_accessible(self, public, users=None, teams=None):
        errors = []

        if public:
            if self.sharing == Contribution.PUBLIC:
                return errors

            if (users is None and teams is None):
                errors.append("The {} '{}' isn't accessible to all".format(self.model_name(), self.fullname()))

            # Public some teams
            if users is not None:
                # Retrieve the users that don't have enough access to the algorithm
                users_with_better_access = [self.author.username]
                users_with_better_access.extend(map(lambda x: x.username, self.shared_with.iterator()))

                missing_users = filter(lambda x: x not in users_with_better_access, users)

                if len(missing_users) > 0:
                    errors.append("The {} '{}' isn't accessible to the following users: {}".format(self.model_name(), self.fullname(), ', '.join(missing_users)))

            # Accessible to some teams
            if teams is not None:
                # Retrieve the teams that don't have enough access to the algorithm
                teams_with_better_access = map(lambda x: x.fullname(), self.shared_with_team.iterator())

                missing_teams = filter(lambda x: x not in teams_with_better_access, teams)

                if len(missing_teams) > 0:
                    errors.append("The {} '{}' isn't accessible to the following teams: {}".format(self.model_name(), self.fullname(), ', '.join(missing_teams)))

        else:
            if self.sharing in [Contribution.PUBLIC, Contribution.USABLE]:
                return errors

            if (users is None and teams is None):
                errors.append("The {} '{}' isn't usable by all".format(self.model_name(), self.fullname()))

            # Usable by some users
            if users is not None:
                # Retrieve the users that don't have enough access to the algorithm
                users_with_better_access = [self.author.username]
                users_with_better_access.extend(map(lambda x: x.username, self.shared_with.iterator()))
                users_with_better_access.extend(map(lambda x: x.username, self.usable_by.iterator()))

                missing_users = filter(lambda x: x not in users_with_better_access, users)

                if len(missing_users) > 0:
                    errors.append("The {} '{}' isn't usable by the following users: {}".format(self.model_name(), self.fullname(), ', '.join(missing_users)))

            # Usable by some teams
            if teams is not None:
                # Retrieve the teams that don't have enough access to the algorithm
                teams_with_better_access = map(lambda x: x.fullname(), self.shared_with_team.iterator())
                teams_with_better_access.extend(map(lambda x: x.fullname(), self.usable_by_team.iterator()))

                missing_teams = filter(lambda x: x not in teams_with_better_access, teams)

                if len(missing_teams) > 0:
                    errors.append("The {} '{}' isn't usable by the following teams: {}".format(self.model_name(), self.fullname(), ', '.join(missing_teams)))

        return errors


    def sharing_preferences(self):
        result = {
            'status': self.get_sharing_display().lower()
        }

        if self.shared_with.count() > 0:
            result['shared_with'] = map(lambda x: x.username, self.shared_with.iterator())

        if self.shared_with_team.count() > 0:
            result['shared_with_team'] = map(lambda x: x.fullname(), self.shared_with_team.iterator())

        if self.usable_by.count() > 0:
            result['usable_by'] = map(lambda x: x.username, self.usable_by.iterator())

        if self.usable_by_team.count() > 0:
            result['usable_by_team'] = map(lambda x: x.fullname(), self.usable_by_team.iterator())

        return result


    def open_source(self, user):
        '''Tells if a given user can view the code of this contribution'''

        if self.author == user: return True
        accessible, open_source, level = self.accessibility_for(user)
        return open_source


    def is_binary(self):
        return self.language in [Code.CXX]


    def language_fullname(self):
        if Code.CODE_NAMES.has_key(self.language):
            return Code.CODE_NAMES[self.language]
        return filter(lambda x: x[0] == self.language, Code.CODE_LANGUAGE)[0][1]


    def json_language(self):
        return Code.language_identifier(self.language)


    @staticmethod
    def language_identifier(db_language):
        return filter(lambda x: x[0] == db_language, iter(Code.CODE_LANGUAGE))[0][1].lower()

    @staticmethod
    def language_db(language_identifier):
        return filter(lambda x: x[1].lower() == language_identifier, iter(Code.CODE_LANGUAGE))[0][0]


    #_____ Overrides __________

    def save(self, *args, **kwargs):

        # Invoke the base implementation
        super(Code, self).save(*args, **kwargs)

        # Ensures that the sharing informations are consistent
        if self.sharing == Code.PUBLIC:
            self.shared_with.clear()
            self.usable_by.clear()
        elif self.sharing == Code.USABLE:
            self.usable_by.clear()


    #_____ Properties __________

    source_code = property(get_source_code, set_source_code)


    #_____ Protected methods __________

    def _save_preprocessing(self):
        # Make sure the declaration is valid, raises SyntaxError if a problem occurs
        wrapper = self.validate(self.declaration)

        # Update the DB entry using the validated declaration
        declaration_hash = beat.core.hash.hashJSON(self.declaration, 'description')
        code_hash = beat.core.hash.hash(self.source_code)

        self.hash = beat.core.hash.hash(dict(
            declaration=declaration_hash,
            code=code_hash,
        ))

        self.short_description = wrapper.description if wrapper.description is not None else ''

        return wrapper


    def _accessibility_for_user(self, user, without_usable=False):
        """Returns a tuple (<has_access>, <open_source>, <accessibility>), with <accessibility> being
           either 'public', 'private', 'confidential' """

        if self.author == user:
            if self.sharing == Contribution.PRIVATE:
                return (True, False, 'private')
            elif self.sharing == Contribution.PUBLIC:
                return (True, True, 'public')
            else:
                return (True, False, 'confidential')

        else:
            if self.sharing == Contribution.PRIVATE:
                return (False, False, None)
            elif self.sharing == Contribution.PUBLIC:
                return (True, True, 'public')
            elif not without_usable and (self.sharing == Contribution.USABLE):
                return (True, False, 'confidential')
            elif not user.is_anonymous():
                if self.shared_with.filter(id=user.id).exists() or (self.shared_with_team.filter(members=user).count() > 0):
                    return (True, True, 'confidential')
                elif not without_usable and (self.usable_by.filter(id=user.id).exists() or (self.usable_by_team.filter(members=user).count() > 0)):
                    return (True, False, 'confidential')

        return (False, False, None)


    def _accessibility_for_team(self, team, without_usable=False):
        """Team specific accessibility check """

        if self.sharing == Contribution.PRIVATE:
            return (False, False, None)
        elif self.sharing == Contribution.PUBLIC:
            return (True, True, 'public')
        elif not without_usable and (self.sharing == Contribution.USABLE):
            return (True, False, 'confidential')
        elif self.shared_with_team.filter(id=team.id).exists():
            return (True, True, 'confidential')
        elif not without_usable and self.usable_by_team.filter(id=team.id).exists():
            return (True, False, 'confidential')

        return (False, False, None)

    #_____ Protected methods __________

    def _share_libraries(self, public, users, teams):
        # Retrieve and process the list of referenced libraries
        needed_libraries = self.referenced_libraries.all()
        own_needed_libraries = needed_libraries.filter(author=self.author)
        other_needed_libraries = needed_libraries.exclude(author=self.author)

        # Ensure that all needed libraries from other users/teams have the necessary sharing
        # preferences
        errors = []
        for needed_library in other_needed_libraries:
            errors.extend(needed_library.is_accessible(public=public, users=users, teams=teams))

        if len(errors) > 0:
            raise ShareError(errors)

        for library in own_needed_libraries:
            library.share(public=public, users=users, teams=teams)
