#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.conf import settings

from django.core.exceptions import ValidationError
from django.contrib.auth.models import User

from rest_framework import serializers

from ..team.models import Team
from ..ui.templatetags.gravatar import gravatar_url
from ..common.utils import validate_restructuredtext, ensure_html

from .models import Shareable, Versionable, Contribution
from .exceptions import ContributionCreationError
from .fields import JSONSerializerField, StringListField

import simplejson as json
import difflib


#----------------------------------------------------------


class DiffSerializer(serializers.Serializer):
    diff = serializers.SerializerMethodField()

    def get_diff(self, obj):
        source1 = json.dumps(obj['object1'].declaration, indent=4)
        source2 = json.dumps(obj['object2'].declaration, indent=4)
        diff = difflib.ndiff(source1.splitlines(), source2.splitlines())

        return '\n'.join(filter(lambda x: x[0] != '?', list(diff)))


#----------------------------------------------------------


class CheckNameSerializer(serializers.Serializer):
    name = serializers.CharField()
    used = serializers.SerializerMethodField()

    def validate_name(self, name):
        return Contribution.sanitize_name(name)

    def get_used(self, obj):
        name = obj.get('name')
        model = self.context.get('model')
        user  = self.context.get('user')
        return model.objects.filter(author=user, name=name).exists()

    def create(self, validated_data):
        return validated_data


#----------------------------------------------------------


class SharingSerializer(serializers.Serializer):
    users = StringListField(required=False)
    teams = StringListField(required=False)

    def validate_users(self, users):
        user_accounts = User.objects.filter(username__in=users)
        if len(users) != user_accounts.count():
            unknown_users = filter(lambda x: x not in map(lambda x: x.username, user_accounts), users)
            if len(unknown_users) > 1:
                raise serializers.ValidationError(['Unknown usernames: ' + ', '.join(unknown_users)])
            else:
                raise serializers.ValidationError(['Unknown username: ' + unknown_users[0]])
        return users

    def validate_teams(self, teams):
        unknown_teams = []

        user = self.context.get('user')

        for team_name in teams:
            parts = team_name.split('/')

            if len(parts) > 2:
                unknown_teams.append(team_name)
                continue

            elif len(parts) == 1:
                parts = [user.username, team_name]

            db_team = Team.objects.filter(owner__username=parts[0], name=parts[1])
            if len(db_team) == 0:
                unknown_teams.append(team_name)

        if len(unknown_teams) > 1:
            raise serializers.ValidationError('Unknown teams: ' + ', '.join(unknown_teams))
        elif len(unknown_teams) == 1:
            raise serializers.ValidationError('Unknown team: ' + unknown_teams[0])

        return teams


#----------------------------------------------------------


class DynamicFieldsSerializer(serializers.ModelSerializer):
    class Meta:
        default_fields = []

    def __init__(self, *args, **kwargs):
        # Don't pass the 'fields' arg up to the superclass
        fields = kwargs.pop('fields', self.Meta.default_fields)

        # Instantiate the superclass normally
        super(DynamicFieldsSerializer, self).__init__(*args, **kwargs)

        # Drop any fields that are not specified in the `fields` argument.
        allowed = set(fields)
        existing = set(self.fields.keys())
        for field_name in existing - allowed:
            self.fields.pop(field_name)


#----------------------------------------------------------


class ShareableSerializer(DynamicFieldsSerializer):
    accessibility = serializers.SerializerMethodField()
    sharing = serializers.SerializerMethodField()
    modifiable = serializers.BooleanField()
    deletable = serializers.BooleanField()
    is_owner = serializers.SerializerMethodField()

    class Meta(DynamicFieldsSerializer.Meta):
        model = Shareable
        default_fields = DynamicFieldsSerializer.Meta.default_fields + \
                         ['is_owner', 'modifiable', 'deletable', 'sharing']

    def get_accessibility(self, obj):
        if obj.sharing == Versionable.PUBLIC:
            return 'public'
        elif obj.sharing == Versionable.SHARED or obj.sharing == Versionable.USABLE:
            return 'confidential'
        else:
            return 'private'


    def get_sharing(self, obj):
        user = self.context.get('user')
        sharing = None
        if hasattr(obj, 'author') and user == obj.author:
            sharing = {'status': obj.get_sharing_display().lower()}

            if obj.shared_with.count() > 0:
                sharing['shared_with'] = map(lambda x: x.username,
                                                       obj.shared_with.iterator())

            if obj.shared_with_team.count() > 0:
                sharing['shared_with_team'] = map(lambda x: x.fullname(),
                                                         obj.shared_with_team.iterator())

        return sharing

    def get_is_owner(self, obj):
        if hasattr(obj, 'author'):
            return obj.author == self.context.get('user')
        return False


#----------------------------------------------------------


class VersionableSerializer(ShareableSerializer):
    name = serializers.CharField(source='fullname')
    fork_of = serializers.SerializerMethodField()
    last_version = serializers.SerializerMethodField()
    previous_version = serializers.SerializerMethodField()
    history = serializers.SerializerMethodField()
    short_description = serializers.CharField(max_length=100)

    code = serializers.SerializerMethodField()

    class Meta(ShareableSerializer.Meta):
        model = Versionable
        default_fields = ShareableSerializer.Meta.default_fields + ['name', 'version', 'last_version',
                                                                    'short_description',
                                                                    'fork_of', 'previous_version',
                                                                    'accessibility',
                                                                    'hash',
                                                                    'creation_date']

    def get_fork_of(self, obj):
        if not(obj.fork_of):
            return None

        accessibility_infos = obj.fork_of.accessibility_for(self.context.get('user'))
        return obj.fork_of.fullname() if accessibility_infos[0] else None

    def get_last_version(self, obj):
        return self.Meta.model.objects.is_last_version(obj)

    def get_previous_version(self, obj):
        if not(obj.previous_version):
            return None

        user = self.context.get('user')

        previous_version = obj.previous_version
        while previous_version is not None:
            accessibility_infos = previous_version.accessibility_for(user)
            if accessibility_infos[0]:
                return obj.previous_version.fullname()

            previous_version = previous_version.previous_version

        return None

    def get_history(self, obj):

        return obj.api_history(self.context.get('user'))


#----------------------------------------------------------


class ContributionSerializer(VersionableSerializer):
    description = serializers.SerializerMethodField()
    declaration = serializers.SerializerMethodField()
    html_description = serializers.SerializerMethodField()

    class Meta(VersionableSerializer.Meta):
        model = Contribution
        extra_fields = ['description', 'declaration']
        exclude = ['description_file', 'declaration_file']

    def get_description(self, obj):
        return obj.description.decode('utf-8')

    def get_declaration(self, obj):
        object_format = self.context.get('object_format')
        if object_format == 'string':
            return json.dumps(obj.declaration, indent=4)
        else:
            return obj.declaration

    def get_html_description(self, obj):
        description = obj.description
        if len(description) > 0:
            return ensure_html(description)
        return ''


#----------------------------------------------------------


class ContributionCreationSerializer(serializers.ModelSerializer):
    declaration = JSONSerializerField(required=False)
    description = serializers.CharField(required=False, allow_blank=True)
    fork_of = serializers.CharField(required=False)
    previous_version = serializers.CharField(required=False)

    class Meta:
        fields = ['name', 'short_description', 'description',
                  'declaration', 'previous_version',
                  'fork_of']
        beat_core_class = None

    def validate_description(self, description):
        if description.find('\\') >= 0: #was escaped, unescape
            description = description.decode('string_escape')
        return description

    def validate(self, data):
        user = self.context.get('user')
        name = self.Meta.model.sanitize_name(data['name'])
        data['name'] = name

        if data.has_key('previous_version'):
            previous_version_id = self.Meta.beat_core_class.Storage(settings.PREFIX,
                data['previous_version'])
            if previous_version_id.username is None:
                previous_version_id.username = user.username
        else:
            previous_version_id = None

        if data.has_key('fork_of'):
            fork_of_id = self.Meta.beat_core_class.Storage(settings.PREFIX,
                data['fork_of'])
            if fork_of_id.username is None:
                fork_of_id.username = user.username
        else:
            fork_of_id = None

        # Retrieve the previous version (if applicable)
        if previous_version_id is not None:
            try:
                previous_version = self.Meta.model.objects.get(
                        author__username__iexact=previous_version_id.username,
                        name=previous_version_id.name,
                        version=previous_version_id.version)
            except:
                raise serializers.ValidationError("Data format '%s' not found" % \
                        previous_version_id.fullname)

            is_accessible = previous_version.accessibility_for(user)
            if not is_accessible[0]:
                raise serializers.ValidationError('No access allowed')
            data['previous_version'] = previous_version

        # Retrieve the forked algorithm (if applicable)
        if fork_of_id is not None:
            try:
                fork_of = self.Meta.model.objects.get(author__username__iexact=fork_of_id.username,
                        name=fork_of_id.name,
                        version=fork_of_id.version)
            except:
                raise serializers.ValidationError("{} '{}' not found".format(self.Meta.model.__name__, fork_of_id.fullname))

            is_accessible = fork_of.accessibility_for(user)
            if not is_accessible[0]:
                raise serializers.ValidationError('No access allowed')
            data['fork_of'] = fork_of

        # Determine the version number
        last_version = None

        if previous_version_id is not None:
            if (previous_version_id.username == user.username) and \
               (previous_version_id.name == name):
                last_version = self.Meta.model.objects.filter(author=user, name=name).order_by('-version')[0]

        if last_version is None:
            if self.Meta.model.objects.filter(author=user, name=name).count() > 0:
                raise serializers.ValidationError('This {} name already exists on this account'.format(self.Meta.model.__name__.lower()))

        data['version'] = (last_version.version + 1 if last_version is not None else 1)

        return data

    def create(self, validated_data):
        (db_object, errors) = self.Meta.model.objects.create_object(**validated_data)
        if errors:
            raise ContributionCreationError(errors)
        return db_object
