#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import os

import simplejson

from django.db import models
from django.conf import settings
from django.core.urlresolvers import reverse

import beat.core.database

from ..dataformats.models import DataFormat

from ..common.models import Shareable
from ..common.models import Versionable
from ..common.models import VersionableManager
from ..common.storage import OverwriteStorage
from ..common import storage

from ..common.models import get_contribution_declaration_filename
from ..common.models import get_contribution_description_filename
from ..code.models import get_contribution_source_code_filename

from ..common.models import get_description
from ..common.models import set_description
from ..common.models import get_declaration
from ..common.models import set_declaration
from ..common.models import get_declaration_string
from ..code.models import get_source_code
from ..code.models import set_source_code


#----------------------------------------------------------


def validate_database(declaration):
    """Validates the declaration of a database JSON string"""

    database = beat.core.database.Database(settings.PREFIX, declaration)

    if not database.valid:
        errors = 'The database declaration is **invalid**. Errors:\n  * ' + \
                 '\n  * '.join(database.errors)
        raise SyntaxError(errors)

    # TODO: for each protocol, if a template already exists, that the inserted
    # protocol respects that template fully (i.e., has the same number of sets,
    # each set with the same outputs and dataformats.

    return database


#----------------------------------------------------------


class DatabaseStorage(OverwriteStorage):

    def __init__(self, *args, **kwargs):
        super(DatabaseStorage, self).__init__(*args, location=settings.DATABASES_ROOT, **kwargs)


#----------------------------------------------------------


class DatabaseManager(VersionableManager):


    def get_by_natural_key(self, name, version):

        return self.get(name=name, version=version)


    def create_database(self, name, short_description='', description='',
                        declaration=None, code=None, version=1,
                        previous_version=None):
        """Convenience function to create a new database from its parts"""

        # Create the database representation
        database = self.model(
            name = name,
            version = version,
            sharing = self.model.PRIVATE,
            previous_version = previous_version,
        )

        # Makes sure we get a declaration in string format
        if declaration is None:
            if previous_version is not None:
                declaration = previous_version.declaration
            else:
                default = beat.core.database.Database(settings.PREFIX, data=None)
                declaration = default.data
        elif not(isinstance(declaration, dict)):
            declaration = simplejson.loads(declaration)

        if len(short_description) > 0:
            declaration['description'] = short_description

        database.declaration = declaration

        # Check the provided code
        if code is None:
            if previous_version is not None:
                code = previous_version.source_code
            else:
                code = ''

        database.source_code = code

        # Check the provided description
        if not description:
            if previous_version is not None:
                description = previous_version.description
            else:
                description = ''

        database.description = description

        # Save the database (will run the validation)
        try:
            database.save()
        except Exception:
            import traceback
            return (None, traceback.format_exc())

        return (database, None)


#----------------------------------------------------------


class Database(Versionable):

    #_____ Fields __________

    declaration_file = models.FileField(
                                    storage=DatabaseStorage(),
                                    upload_to=get_contribution_declaration_filename,
                                    blank=True, null=True,
                                    max_length=200,
                                    db_column='declaration'
                                   )

    description_file = models.FileField(
                                    storage=DatabaseStorage(),
                                    upload_to=get_contribution_description_filename,
                                    blank=True, null=True,
                                    max_length=200,
                                    db_column='description'
                                   )

    source_code_file = models.FileField(
                                    storage=DatabaseStorage(),
                                    upload_to=get_contribution_source_code_filename,
                                    blank=True, null=True,
                                    max_length=200,
                                    db_column='source_code'
                                   )


    objects = DatabaseManager()


    #_____ Meta parameters __________

    class Meta(Versionable.Meta):
        unique_together = ('name', 'version')


    #_____ Utilities __________

    def get_absolute_url(self):

        return reverse(
                'databases:view',
                args=(self.name, self.version,),
                )


    def natural_key(self):

        return (self.name, self.version)


    #_____ Overrides __________

    def save(self, *args, **kwargs):

        # load database descriptor declaration
        declaration = self.declaration

        # tries a simple validation
        wrapper = validate_database(declaration)

        # reset the description
        self.short_description = wrapper.description if (wrapper is not None) and (wrapper.description is not None) else ''

        # Save the changed files (if necessary)
        storage.save_files(self)

        super(Database, self).save(*args, **kwargs)

        # if the filename has changed, move the declaration
        if self.declaration_filename() != self.declaration_file.name:
            storage.rename_file(self, 'declaration_file', self.declaration_filename())
            storage.rename_file(self, 'description_file', self.description_filename())
            storage.rename_file(self, 'source_code_file', self.source_code_filename())


    #_____ Methods __________

    def fullname(self):
        return '%s/%d' % (self.name, self.version)

    def fullpath(self, extension):
        return os.path.join(self.name, str(self.version) + extension)

    def declaration_filename(self):
        return self.fullpath('.json')

    def description_filename(self):
        return self.fullpath('.rst')

    def source_code_filename(self):
        return self.fullpath('.py')

    def all_referenced_dataformats(self):
        result = []
        for database_protocol in self.protocols.all():
            result.extend(database_protocol.all_referenced_dataformats())
        return list(set(result))

    def all_needed_dataformats(self):
        result = []
        for database_protocol in self.protocols.all():
            result.extend(database_protocol.all_needed_dataformats())
        return list(set(result))

    #_____ Properties __________

    description = property(get_description, set_description)
    declaration = property(get_declaration, set_declaration)
    declaration_string = property(get_declaration_string)
    source_code = property(get_source_code, set_source_code)


#----------------------------------------------------------


class DatabaseProtocolManager(models.Manager):

    def get_by_natural_key(self, database_name, database_version, name):
        return self.get(
                database__name=database_name,
                database__version=database_version,
                name=name,
                )


class DatabaseProtocol(models.Model):

    objects = DatabaseProtocolManager()

    database = models.ForeignKey(Database, related_name='protocols',
        on_delete=models.CASCADE)
    name = models.CharField(max_length=200, blank=True)

    class Meta:
        unique_together = ('database', 'name')
        ordering = ['name']

    def __str__(self):
        return self.fullname()

    def natural_key(self):
        return self.database.natural_key() + (self.name,)

    def fullname(self):
        if self.name != '':
            return self.database.fullname() + '@' + self.name
        else:
            return self.database.fullname()

    def all_referenced_dataformats(self):
        result = []
        for database_set in self.sets.all():
            result.extend(database_set.all_referenced_dataformats())
        return list(set(result))

    def all_needed_dataformats(self):
        result = []
        for database_set in self.sets.all():
            result.extend(database_set.all_needed_dataformats())
        return list(set(result))

    def set_template_basename(self):
        if not self.sets.count(): return "unknown"
        dbset = self.sets.all()[0]
        return dbset.template.name.rsplit('__')[0]


#----------------------------------------------------------


class DatabaseSetTemplateManager(models.Manager):

    def get_by_natural_key(self, name):
        return self.get(name=name)


class DatabaseSetTemplate(models.Model):

    objects = DatabaseSetTemplateManager()

    name = models.CharField(max_length=200, unique=True)

    def __str__(self):
        return self.name

    def natural_key(self):
        return (self.name,)


#----------------------------------------------------------


class DatabaseSetManager(models.Manager):

    def get_by_natural_key(self, database_name, database_version, protocol_name, name, template_name):
        return self.get(
                protocol__database__name=database_name,
                protocol__database__version=database_version,
                protocol__name=protocol_name,
                name=name,
                template__name=template_name,
                )


class DatabaseSet(models.Model):

    objects     = DatabaseSetManager()

    protocol    = models.ForeignKey(DatabaseProtocol, related_name='sets',
        on_delete=models.CASCADE)
    name        = models.CharField(max_length=200, blank=True)
    template    = models.ForeignKey(DatabaseSetTemplate, related_name='sets',
        on_delete=models.CASCADE)

    class Meta:
        unique_together = ('protocol', 'name', 'template')

    def __str__(self):
        return self.fullname()

    def natural_key(self):
        return self.protocol.natural_key() + (self.name,) + self.template.natural_key()

    def fullname(self):
        if self.name != '':
            return self.protocol.fullname() + '.' + self.name
        else:
            return self.protocol.fullname()

    def all_referenced_dataformats(self):
        return list(set(map(lambda o: o.dataformat, self.template.outputs.all())))

    def all_needed_dataformats(self):
        result = self.all_referenced_dataformats()

        for dataformat in list(result):
            result.extend(dataformat.all_needed_dataformats())

        return list(set(result))


#----------------------------------------------------------


class DatabaseSetTemplateOutput(models.Model):
    template        = models.ForeignKey(DatabaseSetTemplate,
        related_name='outputs', on_delete=models.CASCADE)
    name            = models.CharField(max_length=200)
    dataformat      = models.ForeignKey(DataFormat,
        related_name='database_outputs', on_delete=models.CASCADE)

    class Meta:
        unique_together = ('template', 'name')

    def __str__(self):
        return self.fullname()

    def fullname(self):
        return self.template.name + '.' + self.name


#----------------------------------------------------------


class DatabaseSetOutputManager(models.Manager):

    def get_by_natural_key(self, hash):
        return self.get(hash=hash)


class DatabaseSetOutput(models.Model):
    template = models.ForeignKey(DatabaseSetTemplateOutput,
        related_name='instances', on_delete=models.CASCADE)
    set = models.ForeignKey(DatabaseSet, related_name='outputs',
        on_delete=models.CASCADE)
    hash = models.CharField(max_length=64, unique=True)

    objects = DatabaseSetOutputManager()

    def __str__(self):
        return self.fullname()

    def fullname(self):
        return '%s.%s.%s.%s' % (
            self.set.protocol.database.fullname(),
            self.set.protocol.name,
            self.set.name,
            self.template.name,
            )

    def all_referenced_dataformats(self):
        return self.template.all_referenced_dataformats()

    def all_needed_dataformats(self):
        return self.template.all_needed_dataformats()

    def natural_key(self):
        return (self.hash,)
