#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################


from django import forms

from ..common.texts import Messages

from ..ui.forms import NameField
from ..ui.forms import CodeMirrorRSTCharField
from ..ui.forms import CodeMirrorJSONCharField

from .models import DataFormat
from .models import validate_format


#----------------------------------------------------------


class CreationForm(forms.Form):

    version = forms.IntegerField(
            required=True,
            min_value=1,
            widget=forms.HiddenInput(), #make it hidden
            )

    name = NameField(
            label='Name',
            max_length=100,
            widget=forms.TextInput(attrs=dict(size="100%")),
            help_text=Messages['format_name'],
            required=True,
            )

    file = CodeMirrorJSONCharField(
            label='Format',
            help_text=Messages['format'],
            required=True,
            )

    short_description = forms.CharField(
            label='Short description',
            max_length=100,
            widget=forms.TextInput(attrs=dict(size="100%")),
            help_text=Messages['short_description'],
            required=False,
            )

    description = CodeMirrorRSTCharField(
            label='Description',
            help_text=Messages['description'],
            required=False,
            )

    def clean_file(self):
        """Cleans-up the file data, make sure it is really new"""

        try:
            validate_format(self.cleaned_data['file'])
        except Exception as e:
            raise forms.ValidationError(str(e))

        # if that works out, then we return the passed file
        return self.cleaned_data['file']

    def clean_name(self):
        """Makes sure the user, name, version is not repeated"""

        name = self.cleaned_data['name']
        version = self.cleaned_data['version']

        obj = DataFormat.objects.filter(
                author = self.__user__,
                name = name,
                version = version,
                )
        if obj:
            raise forms.ValidationError("Choose another name for your data format. The version %d of data format %s already exists" % (version, name))

        return name

    def __init__(self, *args, **kwargs):
        self.__user__ = kwargs.pop('request_user', None)
        self.__name_is_set__ = kwargs.pop('name', None)
        super(CreationForm, self).__init__(*args, **kwargs)
        if self.__name_is_set__:
            self.fields['name'].widget.attrs['readonly'] = 'True'


#----------------------------------------------------------


class UpdateForm(forms.Form):

    file = CodeMirrorJSONCharField(
            label='Declaration',
            help_text=Messages['format'],
            )

    short_description = forms.CharField(
            label='Short description',
            widget=forms.TextInput(attrs=dict(size="100%")),
            help_text=Messages['short_description'],
            required=False,
            )

    description = CodeMirrorRSTCharField(
            label='Description',
            help_text=Messages['description'],
            required=False,
            )


    def clean_file(self):
        """Cleans-up the file data, make sure it is valid"""

        try:
            validate_format(self.cleaned_data['file'])
        except Exception as e:
            raise forms.ValidationError(str(e))

        # if that works out, then we return the passed file
        return self.cleaned_data['file']


#----------------------------------------------------------


class DisplayForm(forms.Form):

    file = CodeMirrorJSONCharField(
            label='Source code',
            help_text=Messages['format'],
            readonly=True, #cannot edit code on display
            )
