#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################


import os
import shutil

from django.test import TestCase
from django.conf import settings
from django.contrib.auth.models import User

from ...common.testutils import tearDownModule

from beat.core.dataformat import Storage, DataFormat as CoreDataFormat

from ..models import DataFormat


class DataFormatsCreationFunction(TestCase):

    def setUp(self):
        if os.path.exists(settings.DATAFORMATS_ROOT):
            shutil.rmtree(settings.DATAFORMATS_ROOT)


    def tearDown(self):
        if os.path.exists(settings.DATAFORMATS_ROOT):
            shutil.rmtree(settings.DATAFORMATS_ROOT)

    def validate(self, dataformat):
        storage = Storage(settings.PREFIX, dataformat.fullname())
        assert storage.exists

        serializer = CoreDataFormat(settings.PREFIX, storage.fullname)
        assert serializer.valid, '\n  * %s' % '\n  * '.join(serializer.errors)


    def test_without_declaration(self):
        user = User.objects.create_user('johndoe', 'johndoe@test.org', '1234')
        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user,
            name='format1',
            short_description='some description',
            )
        assert dataformat, errors

        db_dataformat = DataFormat.objects.get(author=user, name='format1')
        self.assertEqual(db_dataformat.short_description, 'some description')
        self.assertEqual(db_dataformat.short_description, 'some description')
        self.assertTrue(db_dataformat.extend is None)
        self.assertEqual(db_dataformat.referenced_formats.count(), 0)
        self.assertTrue(db_dataformat.modifiable())

        self.validate(db_dataformat)


    def test_with_declaration(self):
        user = User.objects.create_user('johndoe', 'johndoe@test.org', '1234')
        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user,
            name='format1',
            short_description='some description',
            declaration={
                'value': 'int32'
            },
            )
        assert dataformat, errors

        db_dataformat = DataFormat.objects.get(author=user, name='format1')
        self.assertEqual(db_dataformat.short_description, 'some description')
        self.assertEqual(db_dataformat.short_description, 'some description')
        self.assertTrue(db_dataformat.extend is None)
        self.assertEqual(db_dataformat.referenced_formats.count(), 0)
        self.assertTrue(db_dataformat.modifiable())

        self.validate(db_dataformat)


    def test_extended_format(self):
        user = User.objects.create_user('johndoe', 'johndoe@test.org', '1234')
        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user,
            name='format1',
            short_description='some description',
            declaration={
                'value': 'int32'
            },
            )
        assert dataformat, errors

        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user,
            name='format2',
            short_description='some description',
            declaration={
                '#extends': 'johndoe/format1/1',
                'value2': 'int32'
            },
            )
        assert dataformat, errors

        db_dataformat = DataFormat.objects.get(author=user, name='format2')
        self.assertEqual(db_dataformat.short_description, 'some description')
        self.assertEqual(db_dataformat.short_description, 'some description')
        self.assertTrue(db_dataformat.extend is not None)
        self.assertEqual(db_dataformat.extend.name, 'format1')
        self.assertEqual(db_dataformat.referenced_formats.count(), 0)
        self.assertTrue(db_dataformat.modifiable())

        self.validate(db_dataformat)


    def test_composed_format(self):
        user = User.objects.create_user('johndoe', 'johndoe@test.org', '1234')
        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user,
            name='format1',
            short_description='some description',
            declaration={
                'value': 'int32'
            },
        )
        assert dataformat, errors

        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user,
            name='format2',
            short_description='some description',
            declaration={
                'value2': 'int32'
            },
        )
        assert dataformat, errors

        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user,
            name='format3',
            short_description='some description',
            declaration={
                'field1': 'johndoe/format1/1',
                'field2': 'johndoe/format2/1'
            },
        )
        assert dataformat, errors

        self.assertTrue(dataformat is not None)

        db_dataformat = DataFormat.objects.get(author=user, name='format3')
        self.assertEqual(db_dataformat.short_description, 'some description')
        self.assertEqual(db_dataformat.short_description, 'some description')
        self.assertTrue(db_dataformat.extend is None)
        self.assertEqual(db_dataformat.referenced_formats.count(), 2)
        self.assertEqual(db_dataformat.referenced_formats.filter(name='format1').count(), 1)
        self.assertEqual(db_dataformat.referenced_formats.filter(name='format2').count(), 1)
        self.assertTrue(db_dataformat.modifiable())

        self.validate(db_dataformat)


    def test_extended_composed_format(self):
        user = User.objects.create_user('johndoe', 'johndoe@test.org', '1234')
        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user,
            name='format1',
            short_description='some description',
            declaration={
                'value': 'int32'
            },
            )
        assert dataformat, errors

        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user,
            name='format2',
            short_description='some description',
            declaration={
                'value2': 'int32'
            },
            )
        assert dataformat, errors

        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user,
            name='format3',
            short_description='some description',
            declaration={
                'value3': 'int32'
            },
            )
        assert dataformat, errors

        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user,
            name='format4',
            short_description='some description',
            declaration={
                '#extends': 'johndoe/format3/1',
                'field1': 'johndoe/format1/1',
                'field2': 'johndoe/format2/1'
            },
            )
        assert dataformat, errors

        db_dataformat = DataFormat.objects.get(author=user, name='format4')
        self.assertEqual(db_dataformat.short_description, 'some description')
        self.assertEqual(db_dataformat.short_description, 'some description')
        self.assertTrue(db_dataformat.extend is not None)
        self.assertEqual(db_dataformat.extend.name, 'format3')
        self.assertEqual(db_dataformat.referenced_formats.count(), 2)
        self.assertEqual(db_dataformat.referenced_formats.filter(name='format1').count(), 1)
        self.assertEqual(db_dataformat.referenced_formats.filter(name='format2').count(), 1)
        self.assertTrue(db_dataformat.modifiable())

        self.validate(db_dataformat)
