#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.contrib.auth.models import User

from beat.web.team.models import Team

from ..models import DataFormat

from .core import DataFormatsAccessibilityFunctionsBase

class NotSharedDataFormat_CheckTeamAccessibilityFunction(DataFormatsAccessibilityFunctionsBase):

    def test_team_has_no_access(self):
        dataformat = DataFormat.objects.get(name='personal')

        errors = dataformat.is_accessible(teams=['jackdoe/teamdoe'])
        self.assertEqual(len(errors), 1)


class PublicForOneTeamDataFormat_CheckAccessibilityFunction(DataFormatsAccessibilityFunctionsBase):

    def test_authorized_team_has_full_access(self):
        dataformat = DataFormat.objects.get(name='public_for_one_team')
        name = 'jackdoe/teamdoe'

        errors = dataformat.is_accessible(teams=[name])
        self.assertEqual(len(errors), 0)


    def test_unauthorized_team_has_no_access(self):
        dataformat = DataFormat.objects.get(name='public_for_one_team')
        name = 'jackdoe/teamdoe2'

        errors = dataformat.is_accessible(teams=[name])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The dataformat 'jackdoe/public_for_one_team/1' isn't accessible to the following teams: {}".format(name))


    def test_mixed_case(self):
        dataformat = DataFormat.objects.get(name='public_for_one_team')
        teams = ['jackdoe/teamdoe', 'jackdoe/teamdoe2']

        errors = dataformat.is_accessible(teams=teams)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The dataformat 'jackdoe/public_for_one_team/1' isn't accessible to the following teams: {}".format(teams[1]))


class PublicForTeamsDataFormat_CheckAccessibilityFunction(DataFormatsAccessibilityFunctionsBase):

    def test_authorized_team_has_full_access(self):
        dataformat = DataFormat.objects.get(name='public_for_teams')
        teams = ['jackdoe/teamdoe']

        errors = dataformat.is_accessible(teams=teams)
        self.assertEqual(len(errors), 0)


    def test_authorized_teams_has_full_access(self):
        dataformat = DataFormat.objects.get(name='public_for_teams')
        teams = ['jackdoe/teamdoe', 'jackdoe/teamdoe2']

        errors = dataformat.is_accessible(teams=teams)
        self.assertEqual(len(errors), 0)


    def test_unauthorized_team_has_no_access(self):
        dataformat = DataFormat.objects.get(name='public_for_teams')
        name = 'jackdoe/teamdoe3'

        errors = dataformat.is_accessible(teams=[name])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The dataformat 'jackdoe/public_for_teams/1' isn't accessible to the following teams: {}".format(name))


    def test_mixed_case(self):
        dataformat = DataFormat.objects.get(name='public_for_teams')
        teams = ['jackdoe/teamdoe', 'jackdoe/teamdoe2', 'jackdoe/teamdoe3']

        errors = dataformat.is_accessible(teams=teams)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The dataformat 'jackdoe/public_for_teams/1' isn't accessible to the following teams: {}".format(teams[2]))


class PublicForAllUsersDataFormat_CheckTeamAccessibilityFunction(DataFormatsAccessibilityFunctionsBase):

    def test_teams_have_full_access(self):
        dataformat = DataFormat.objects.get(name='public_for_all')
        teams = ['jackdoe/teamdoe', 'jackdoe/teamdoe2']

        errors = dataformat.is_accessible(teams=teams)
        self.assertEqual(len(errors), 0)


class NotSharedDataFormat_TeamShareFunction(DataFormatsAccessibilityFunctionsBase):

    def test_sharing_as_public_for_one_team(self):
        dataformat = DataFormat.objects.get(name='personal')
        team = ['jackdoe/teamdoe']
        dataformat.share(teams=team)

        reference = {
            'status': 'shared',
            'shared_with_team': team
        }

        self.checkObjectDBSharingPreferences(dataformat, reference)


    def test_sharing_as_public_for_multiple_teams(self):
        dataformat = DataFormat.objects.get(name='personal')
        teams = ['jackdoe/teamdoe', 'jackdoe/teamdoe2']
        dataformat.share(teams=teams)

        reference = {
            'status': 'shared',
            'shared_with_team': teams
        }

        self.checkObjectDBSharingPreferences(dataformat, reference)


class PublicForOneTeamDataFormat_ShareFunction(DataFormatsAccessibilityFunctionsBase):

    def test_sharing_as_public_for_another_team(self):
        dataformat = DataFormat.objects.get(name='public_for_one_team')

        dataformat.share(teams=['jackdoe/teamdoe2'])

        reference = {
            'status': 'shared',
            'shared_with_team': ['jackdoe/teamdoe', 'jackdoe/teamdoe2']
        }

        self.checkObjectDBSharingPreferences(dataformat, reference)


    def test_sharing_as_public_for_multiple_teams(self):
        dataformat = DataFormat.objects.get(name='public_for_one_team')

        dataformat.share(teams=['jackdoe/teamdoe2', 'jackdoe/teamdoe3'])

        reference = {
            'status': 'shared',
            'shared_with_team': ['jackdoe/teamdoe', 'jackdoe/teamdoe2', 'jackdoe/teamdoe3']
        }

        self.checkObjectDBSharingPreferences(dataformat, reference)

class TeamDeletionPropagation(DataFormatsAccessibilityFunctionsBase):

    def test_propagation(self):

        jackdoe = User.objects.get(username='jackdoe')
        johndoe = User.objects.get(username='johndoe')
        janedoe = User.objects.get(username='janedoe')

        team_to_delete = Team.objects.create(name='team_to_delete', owner=jackdoe)
        team_to_delete.members.add(janedoe, johndoe)

        self.assertEqual(DataFormat.objects.for_user(janedoe).count(), 1)
        self.assertEqual(DataFormat.objects.for_user(johndoe).count(), 3)

        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=jackdoe,
            name='format1',
            short_description='some description',
            declaration={
                'value': 'int32'
            },
        )
        assert dataformat, errors

        dataformat.share(teams=[team_to_delete])

        self.assertEqual(DataFormat.objects.for_user(janedoe).count(), 2)
        self.assertEqual(DataFormat.objects.for_user(johndoe).count(), 4)


        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=jackdoe,
            name='format2',
            short_description='some description',
            declaration={
                'value2': 'int32'
            },
        )
        assert dataformat, errors

        dataformat.share(teams=[team_to_delete])

        self.assertEqual(DataFormat.objects.for_user(janedoe).count(), 3)
        self.assertEqual(DataFormat.objects.for_user(johndoe).count(), 5)

        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=janedoe,
            name='format3',
            short_description='some description',
            declaration={
                'field1': 'jackdoe/format1/1',
                'field2': 'jackdoe/format2/1'
            },
        )
        assert dataformat, errors

        self.assertEqual(DataFormat.objects.for_user(janedoe).count(), 4)
        self.assertEqual(DataFormat.objects.for_user(johndoe).count(), 5)


        jackdoe_db_dataformat1 = DataFormat.objects.get(author=jackdoe, name='format1')
        self.assertEqual(jackdoe_db_dataformat1.shared_with_team.all().count(), 1)
        jackdoe_db_dataformat2 = DataFormat.objects.get(author=jackdoe, name='format2')
        self.assertEqual(jackdoe_db_dataformat2.shared_with_team.all().count(), 1)

        janedoe_db_dataformat = DataFormat.objects.get(author=janedoe, name='format3')
        self.assertEqual(janedoe_db_dataformat.referenced_formats.count(), 2)
        self.assertEqual(janedoe_db_dataformat.shared_with_team.all().count(), 0)
        self.assertEqual(DataFormat.objects.for_user(janedoe).count(), 4)
        self.assertEqual(DataFormat.objects.for_user(johndoe).count(), 5)

        team_to_delete.delete()

        self.assertEqual(jackdoe_db_dataformat1.shared_with_team.all().count(), 0)
        self.assertEqual(jackdoe_db_dataformat2.shared_with_team.all().count(), 0)

        self.assertEqual(DataFormat.objects.for_user(janedoe).count(), 4)
        self.assertEqual(DataFormat.objects.for_user(johndoe).count(), 3)
