#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.db import models
from django.contrib.auth.models import User
from django.core.urlresolvers import reverse

from ..algorithms.models import Algorithm
from ..experiments.models import Experiment
from ..plotters.models import Plotter
from ..plotters.models import PlotterParameter
from ..common.texts import Messages

import random
from datetime import datetime
import simplejson as json


#----------------------------------------------------------


class ReportManager(models.Manager):

    def create_object(self, author, name, short_description='', description='', content={},
                      experiments=[]):

        # Creation of the report
        report                   = Report()
        report.name              = name
        report.author            = author
        report.short_description = short_description
        report.description       = description
        if isinstance(content, dict):
            report.content       = json.dumps(content, indent=4)
        else:
            report.content       = content
        report.creation_date     = datetime.now()
        report.publication_date  = None
        report.expiration_date   = None
        report.status            = self.model.EDITABLE

        report.save()

        # Add the experiments
        result = report.add_experiments(experiments)
        if not(result['success']):
            report.delete()
            report = None

        return (report, result)

    def editable(self):
        return self.filter(status=Report.EDITABLE)

    def locked(self):
        return self.filter(status=Report.LOCKED)

    def published(self):
        return self.filter(status=Report.PUBLISHED)

#----------------------------------------------------------


class Report(models.Model):

    #_____ Constants __________

    EDITABLE  = 'E'
    LOCKED    = 'L'
    PUBLISHED = 'P'

    REPORT_STATUS = (
        (EDITABLE,  'Editable'),
        (LOCKED,    'Locked'),
        (PUBLISHED, 'Published'),
    )


    #_____ Fields __________

    status            = models.CharField(max_length=1, choices=REPORT_STATUS, default=EDITABLE)
    name              = models.CharField(max_length=200, help_text=Messages['name'], blank=False)
    number            = models.IntegerField(blank=True)
    author            = models.ForeignKey(User, related_name='%(class)ss')
    experiments       = models.ManyToManyField(Experiment, related_name='reports', blank=True)
    creation_date     = models.DateTimeField()
    expiration_date   = models.DateTimeField(null=True, blank=True)
    publication_date  = models.DateTimeField(null=True, blank=True)
    short_description = models.CharField(max_length=100, default='', blank=True, help_text=Messages['short_description'])
    description       = models.TextField(default='', blank=True)
    content           = models.TextField(default='{}', blank=True)
    analyzer          = models.ForeignKey(Algorithm, related_name='reports', null=True, blank=True)

    # read-only parameters that are updated at every save(), if required
    referenced_plotters   = models.ManyToManyField(Plotter, related_name='reports', blank=True)
    referenced_plotterparameters = models.ManyToManyField(PlotterParameter, related_name='reports', blank=True)

    objects = ReportManager()


    def __str__(self):
        return "Report %s (#%d)" % (self.fullname(), self.number)


    def fullname(self):
        return '%s/%s' % (self.author.username, self.name)


    def get_absolute_url(self):
        return reverse(
                'reports:view',
                args=(
                    self.number,
                    ),
                )


    def get_author_absolute_url(self):
        return reverse(
                'reports:author-view',
                args=(
                    self.author.username,
                    self.name
                    ),
                )


    def get_api_update_url(self):
        return reverse(
                'api_reports:object',
                args=(
                    self.author.username,
                    self.name
                    ),
                )


    def get_api_add_url(self):
        return reverse(
                'api_reports:add_experiments',
                args=(
                    self.author.username,
                    self.name
                    ),
                )


    def get_api_remove_url(self):
        return reverse(
                'api_reports:remove_experiments',
                args=(
                    self.author.username,
                    self.name
                    ),
                )


    def save(self, *args, **kwargs):
        if self.number is None:
            # Generate a unique report number
            used_numbers = map(lambda x: x.number, Report.objects.all())
            number = 0
            while (number == 0) or number in used_numbers:
                number = random.randint(100000, 2**31)

            self.number = number

        report_content = json.loads(self.content)
        report_content_charts = dict(filter(lambda item: item[0].startswith("chart"),report_content.iteritems()))

        super(Report, self).save(*args, **kwargs)

        self.referenced_plotters.clear()
        self.referenced_plotterparameters.clear()
        for key, value in report_content_charts.iteritems():
            plotter_parts = value['data']['plotter'].split("/")
            plotterparameters_parts = value['selected_template'].split("/")
            plotter = Plotter.objects.get(author__username = plotter_parts[0], name = plotter_parts[1], version = plotter_parts[2])
            plotterparameter = PlotterParameter.objects.get(author__username = plotterparameters_parts[0], name = plotterparameters_parts[1], version = plotterparameters_parts[2])
            self.referenced_plotters.add(plotter)
            self.referenced_plotterparameters.add(plotterparameter)


    def add_experiments(self, experiment_fullname_list):
        # Check that the report is modifiable
        if self.status != Report.EDITABLE:
            return {
                'success': False,
                'error': "Report not modifiable",
            }

        # Process the list of experiments
        common_analyzers       = None
        accessible_experiments = []
        inaccessible_experiments = []

        experiments, unknown_experiments = self._get_experiments_from_fullname_list(experiment_fullname_list)

        for experiment in experiments:
            # Check experiments accessibility
            (is_accessible, accessibility) = experiment.accessibility_for(self.author)
            if not(is_accessible) or (experiment.status != Experiment.DONE):
                inaccessible_experiments.append(experiment.fullname())
                continue

            accessible_experiments.append(experiment)

            if self.analyzer is None:
                if common_analyzers is None:
                    common_analyzers = map(lambda x: x.algorithm, experiment.blocks.filter(analyzer=True))
                else:
                    analyzers = map(lambda x: x.algorithm, experiment.blocks.filter(analyzer=True))
                    common_analyzers = filter(lambda x: x in analyzers, common_analyzers)
                    if len(common_analyzers) == 0:
                        return {
                            'success': False,
                            'error': "No common analyzer",
                        }


        # Check that we have common analyzers (if necessary)
        if (self.analyzer is None) and (common_analyzers is not None):
            if len(common_analyzers) == 1:
                self.analyzer = common_analyzers[0]
                self.save()

            elif len(common_analyzers) > 1:
                return {
                    'success': False,
                    'common_analyzers': map(lambda x: x.fullname(), common_analyzers),
                }


        # Add the experiments to the report
        incompatible_experiments = []
        for experiment in accessible_experiments:
            if len(experiment.blocks.filter(analyzer=True, algorithm=self.analyzer)) >= 1:
                self.experiments.add(experiment)
            else:
                incompatible_experiments.append(experiment.fullname())

        return {
            'success':                  True,
            'unknown_experiments':      unknown_experiments,
            'inaccessible_experiments': inaccessible_experiments,
            'incompatible_experiments': incompatible_experiments,
        }


    def remove_experiments(self, experiment_fullname_list):

        # Check that the report is modifiable
        if self.status != Report.EDITABLE:
            return

        experiments, unknown_experiments = self._get_experiments_from_fullname_list(experiment_fullname_list)

        for experiment in experiments:
            self.experiments.remove(experiment)

        if self.experiments.count() == 0:
            self.analyzer = None
            self.save()


    def _get_experiments_from_fullname_list(self, experiment_fullname_list):
        experiments = []
        unknown_experiments = []

        for experiment_name in experiment_fullname_list:
            parts = experiment_name.split('/')

            if len(parts) != 5:
                raise ValidationError({'experiment': 'Invalid experiment full name'})

            try:
                experiment = Experiment.objects.get(author__username=parts[0],
                                                    toolchain__author__username=parts[1],
                                                    toolchain__name=parts[2],
                                                    toolchain__version=int(parts[3]),
                                                    name=parts[4])
            except Experiment.DoesNotExist:
                unknown_experiments.append(experiment_name)
            else:
                experiments.append(experiment)

        return experiments, unknown_experiments


    def _get_experiments_and_alias(self, alias_filter):

        experiments_list = []
        alias_list = []

        report_content = json.loads(self.content)

        if not report_content.has_key("alias_experiments"):
            report_content["alias_experiments"] = {}

        for experiment in self.experiments.iterator():
            if not report_content["alias_experiments"].has_key(experiment.fullname()):
                report_content["alias_experiments"][experiment.fullname()] = experiment.name
            try:
                index = alias_filter.index(report_content["alias_experiments"][experiment.fullname()])
                experiments_list.append(experiment.fullname())
            except:
                pass


        alias_list = map(lambda x: report_content["alias_experiments"][x], experiments_list)

        return experiments_list, alias_list
