#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.contrib.auth.models import User, AnonymousUser

from rest_framework import serializers

from .models import Report
from ..common.models import Contribution
from ..common.fields import JSONSerializerField
from ..experiments.models import Experiment
from ..ui.templatetags.markup import restructuredtext
from ..common.utils import validate_restructuredtext

import simplejson as json


#----------------------------------------------------------


class BasicReportSerializer(serializers.ModelSerializer):
    name = serializers.SerializerMethodField()
    is_owner = serializers.SerializerMethodField()
    status = serializers.SerializerMethodField()
    members = serializers.SerializerMethodField()
    content = serializers.SerializerMethodField()
    author = serializers.SerializerMethodField()
    experiments = serializers.SerializerMethodField()
    analyzer = serializers.SerializerMethodField()
    html_description = serializers.SerializerMethodField()
    add_url = serializers.SerializerMethodField()

    class Meta:
        model = Report
        fields = ['short_description', 'status', 'is_owner', 'accessibility']

    def get_name(self, obj):
        return '{}/{}'.format(obj.author.username, obj.name)

    def get_author(self, obj):
        return obj.author.username

    def get_is_owner(self, obj):
        user = self.context["request"].user
        if not user.is_anonymous():
            return obj.author == user
        return False

    def get_content(self, obj):
        return json.loads(obj.content)

    def get_status(self, obj):
        if obj.status == Report.EDITABLE:
            return 'editable'
        elif obj.status == Report.LOCKED:
            return 'locked'
        elif obj.status == Report.PUBLISHED:
            return 'published'
        else:
            return 'editable'

    def get_experiments(self, obj):
        return map(lambda x: x.fullname(), obj.experiments.iterator())

    def get_analyzer(self, obj):
        if obj.analyzer is not None:
            return obj.analyzer.fullname()
        else:
            return None

    def get_html_description(self, obj):
        d = obj.description
        if len(d) > 0:
            return restructuredtext(d)
        return ''

    def get_add_url(self, obj):
        return obj.get_api_add_url()


#----------------------------------------------------------


class SimpleReportSerializer(BasicReportSerializer):

    class Meta(BasicReportSerializer.Meta):
        fields = ['name', 'number', 'short_description', 'is_owner', 'author','status', 'description', 'creation_date', 'html_description', 'add_url']


#----------------------------------------------------------


class FullReportSerializer(BasicReportSerializer):


    class Meta(BasicReportSerializer.Meta):
        fields = ['name', 'number', 'short_description', 'description', 'is_owner', 'author','status', 'creation_date', 'publication_date', 'experiments', 'analyzer', 'content', 'html_description']


#----------------------------------------------------------


class CreatedReportSerializer(BasicReportSerializer):

    class Meta(BasicReportSerializer.Meta):
        fields = ['name',  'short_description', 'experiments']


#----------------------------------------------------------


class UpdatedReportSerializer(BasicReportSerializer):

    class Meta(BasicReportSerializer.Meta):
        fields = ['short_description', 'description', 'experiments', 'content']


#----------------------------------------------------------


class ReportCreationFailedException(Exception):
    pass


#----------------------------------------------------------


class ReportCreationSerializer(serializers.ModelSerializer):
    content = JSONSerializerField()
    experiments = serializers.ListField(child=serializers.CharField())

    class Meta:
        model = Report
        fields = ['name', 'short_description', 'description', 'content', 'experiments']

    def create(self, validated_data):
        report = None
        if not validated_data.has_key('name'):
            raise serializers.ValidationError('No name provided')
        try:
            report = Report.objects.get(author=self.context['request'].user, name=validated_data['name'])
        except:
            pass

        if report is not None:
            raise serializers.ValidationError('A report with this name already exists')

        (report, self.details) = self.Meta.model.objects.create_object(author=self.context['request'].user, **validated_data)
        if report is None:
            raise ReportCreationFailedException()

        return report

    def update(self, instance, validated_data):
        instance.short_description = validated_data.get('short_description', instance.short_description)
        instance.description       = validated_data.get('description', instance.description)
        instance.content           = validated_data.get('content', instance.content)
        instance.publication_date  = validated_data.get('publication_date', instance.publication_date)

        instance.save()

        if validated_data.has_key('experiments'):
            experiments         = validated_data.get('experiments', [])
            current_experiments = map(lambda x: x.fullname(), instance.experiments.iterator())

            experiments_to_remove = filter(lambda x: x not in experiments, current_experiments)
            experiments_to_add    = filter(lambda x: x not in current_experiments, experiments)

            instance.remove_experiments(experiments_to_remove)

            self.details = instance.add_experiments(experiments_to_add)
        else:
            self.details = None

        return instance

    def validate_name(self, value):
        return Contribution.sanitize_name(value)

    def validate_content(self, value):
        return json.dumps(value)

    def validate_experiments(self, value):
        return value

    def validate_description(self, value):
        validate_restructuredtext(value)
        return value

    def to_representation(self, obj):
        serializer = CreatedReportSerializer(obj, context=self.context)
        return serializer.data


#----------------------------------------------------------


class ReportUpdateSerializer(ReportCreationSerializer):

    class Meta(ReportCreationSerializer.Meta):
        fields = ['short_description', 'description', 'content', 'experiments']

    def to_representation(self, obj):
        serializer = UpdatedReportSerializer(obj)
        return serializer.data
