/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 *
 * This file is part of the beat.web module of the BEAT platform.
 *
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 *
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
 */

/*
 * ReportService
 * Desc:
 * 	The datastore for the reports app, including data fetched from the
 * 	server and cross-component app state
 */
angular.module('reportApp').factory('ReportService', ['$http', function($http){
	let reportServiceInstance = {};
	// experiments of reports are in arbitrary groups,
	// in a many-to-many relationship
	let groupData = [];

	// represents a Group in the report
	// has a name and a list of experiments that belong to it
	class Group {
		constructor (name) {
			this._name = name;
			this.experimentNames = new Set();
		}

		// get the experiment names in this group
		get experiments () {
			return Array.from(this.experimentNames);
		}

		// get the group name
		get name () {
			return this._name;
		}

		// check if an exp is in this group
		hasExperiment (expName) {
			return this.experimentNames.has(expName);
		}

		// add an exp to this group
		addExperiment (expName) {
			return this.experimentNames.add(expName);
		}

		// rm an exp from this group
		removeExperiment (expName) {
			return this.experimentNames.delete(expName);
		}
	};

	// gets groups
	reportServiceInstance.groups = groupData;

	// serializes groups as an object with form:
	// {
	// 	<group name 1>: <array of exp names for group 1>,
	// 	...
	// }
	reportServiceInstance.serializeGroups = () => {
		return groupData
		.map(g => { return { [g.name]: g.experiments }; })
		.reduce((o, g) => Object.assign(o, g), {});
	};

	// create a new group for the report
	// returns false if it already exists
	// returns true if it adds a new group
	reportServiceInstance.createGroup = (name) => {
		if(typeof name !== 'string'){
			throw new Error(`new group name is not a string: ${JSON.stringify(name)}`);
		}

		if(groupData.find(g => g.name === name)){
			return false;
		}

		groupData.push(new Group(name));
		return true;
	};

	// delete a group
	reportServiceInstance.deleteGroup = (name) => {
		let idx = groupData.indexOf(groupData.find(g => g.name === name));
		if (idx > -1) {
			groupData.splice(idx, 1);
		}
	};

	// load group info from the serialized format:
	// {
	// 	<group name 1>: <array of exp names for group 1>,
	// 	...
	// }
	reportServiceInstance.loadGroups = (data) => {
		// wipe data and load groups
		groupData.splice(0, groupData.length);
		Object.entries(data)
		.forEach(([groupName, expNames]) => {
			let g = new Group(groupName);
			expNames.forEach(n => g.addExperiment(n));
			groupData.push(g);
		});
	};

	// save group info via sending to server
	reportServiceInstance.saveGroups = () => {
		console.warn(`not implemented: reportServiceInstance.saveGroups`);
	};

	// add experiment to group
	reportServiceInstance.addExperimentToGroup = (expName, groupName) => {
		checkForGroup(groupName);

		let group = groupData.find(g => g.name === groupName);
		return group.addExperiment(expName);
	};

	// rm experiment from group
	reportServiceInstance.removeExperimentFromGroup = (expName, groupName) => {
		checkForGroup(groupName);
		let group = groupData.find(g => g.name === groupName);
		return group.removeExperiment(expName);
	};

	// gets experiments from a group
	reportServiceInstance.getGroupExperiments = (groupName) => {
		checkForGroup(groupName);
		let group = groupData.find(g => g.name === groupName);
		return group.experiments;
	};

	// gets groups for an experiment
	reportServiceInstance.getExperimentGroups = (expName) => {
		return groupData.filter(g => g.hasExperiment(expName)).map(g => g.name);
	};

	// helper to assert that a group exists
	function checkForGroup (groupName) {
		if(!groupData.find(g => g.name === groupName)){
			throw new Error(`Could not find group "${JSON.stringify(groupName)}"`);
		}
	}

	return reportServiceInstance;
}]);
