#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################


"""\
Starts the scheduling process.

Usage:
  %(prog)s [-v ... | --verbose ...] [--settings=<file>] [--period=<seconds>]
  %(prog)s (-h | --help)
  %(prog)s (-V | --version)


Options:
  -h, --help                        Show this help message
  -V, --version                     Show program's version number
  -v, --verbose                     Increases the output verbosity level
  -S <file>, --settings=<file>      The module name to the Django settings
                                    file [default: beat.web.settings.settings]
  -p <seconds, --period=<seconds>   The time, in seconds, in which this
                                    scheduler will try to allocate job splits
                                    to existing workers. If not set, use the
                                    value available on the Django settings
                                    file, at the variable
                                    `SCHEDULING_INTERVAL`.


Examples:

  To start the scheduling process do the following:

    $ %(prog)s

  You can pass the ``-v`` flag to start the scheduler with the logging level
  set to ``INFO`` or ``-vv`` to set it to ``DEBUG``. By default, the logging
  level is set to ``WARNING`` if no ``-v`` flag is passed.

"""

import os
import sys
import time
import signal
import docopt
import logging

stop = False

def main(user_input=None):

    arguments = docopt.docopt(
      __doc__ % dict(
        prog=os.path.basename(sys.argv[0]),
        ),
      )

    os.environ.setdefault('DJANGO_SETTINGS_MODULE', arguments['--settings'])
    from django.conf import settings
    from django import setup
    setup()

    logger = logging.getLogger('beat.web')
    if arguments['--verbose'] == 1: logger.setLevel(logging.INFO)
    elif arguments['--verbose'] >= 2: logger.setLevel(logging.DEBUG)

    # installs SIGTERM handler
    def handler(signum, frame):
        #ignore further signals
        signal.signal(signal.SIGTERM, signal.SIG_IGN)
        signal.signal(signal.SIGINT, signal.SIG_IGN)

        logger.info("Signal %d caught, terminating...", signum)
        global stop
        stop = True

    signal.signal(signal.SIGTERM, handler)
    signal.signal(signal.SIGINT, handler)

    from ..backend import schedule

    timing = int(arguments['--period']) \
        if arguments['--period'] else settings.SCHEDULING_INTERVAL
    logger.info("Scheduling every %d seconds", timing)

    global stop
    while not stop:

        start = time.time()
        logger.debug("Starting scheduler cycle...")
        schedule.schedule()
        duration = time.time() - start
        if duration < timing:
            time.sleep(timing - duration)

    logger.info("Gracefully exiting the scheduler")
