#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from rest_framework import serializers

from ..common.serializers import VersionableSerializer
from ..ui.templatetags.markup import restructuredtext

from .models import Search, Leaderboard
from .fields import DictListField

import simplejson as json

import six


class SearchResultSerializer(serializers.Serializer):
    name = serializers.SerializerMethodField()
    description = serializers.CharField(source='short_description')

    def __init__(self, *args, **kwargs):
        self.name_field = kwargs.pop('name_field', 'fullname')

        super(SearchResultSerializer, self).__init__(*args, **kwargs)

    def get_name(self, obj):
        name_attr = getattr(obj, self.name_field)
        if isinstance(name_attr, six.string_types):
            return name_attr
        return name_attr()


class SearchSerializer(VersionableSerializer):
    name = serializers.CharField(source="fullname")
    author = serializers.CharField(source="author.username")
    user_name = serializers.SerializerMethodField()
    is_owner = serializers.SerializerMethodField()
    accessibility = serializers.SerializerMethodField()
    html_description = serializers.SerializerMethodField()
    update_url = serializers.SerializerMethodField()
    has_leaderboard = serializers.SerializerMethodField()

    class Meta(VersionableSerializer.Meta):
        model = Search
        default_fields = ['name', 'author', 'user_name', 'is_owner', 'accessibility', 'short_description', 'update_url']

    def get_user_name(self, obj):
        return self.context['user'].username

    def get_is_owner(self, obj):
        return self.context['user'] == obj.author

    def get_accessibility(self, obj):
        (has_access, accessibility) = obj.accessibility_for(self.context['user'])
        return accessibility

    def get_html_description(self, obj):
        d = obj.description
        if len(d) > 0:
            return restructuredtext(d)
        return ''

    def get_update_url(self, obj):
        return obj.get_api_update_url()

    def get_has_leaderboard(self, obj):
        return obj.has_leaderboard()


class SearchWriteSerializer(serializers.ModelSerializer):
    description = serializers.CharField(required=False)
    leaderboard = serializers.BooleanField(required=False)
    filters = DictListField()
    settings = DictListField()

    class Meta:
        model = Search
        fields = ['name', 'short_description', 'description',
                  'filters', 'settings', 'leaderboard']
        default_fields = []

    def validate_name(self, name):
        sanitized_name = Search.sanitize_name(name)
        user = self.context['user']
        if Search.objects.filter(author=user, name=sanitized_name).exists():
            raise serializers.ValidationError('Name already used')
        return sanitized_name

    def validate(self, validated_data):
        if 'description' in validated_data and\
            not 'short_description' in validated_data:
            raise serializers.ValidationError('Missing short description')

        return validated_data

    def create(self, validated_data):
        user = self.context['user']
        search = Search(author=user, version=1)
        search.name = validated_data['name']
        search.filters = json.dumps(validated_data.get('filters'))
        search.settings = json.dumps(validated_data.get('settings'))
        search.short_description = validated_data.get('short_description', '')
        search.description = validated_data.get('description', '')
        search.save()

        #if leaderboard activation was asked, activate the leaderboard
        if validated_data.get('leaderboard'):
            leaderboard = Leaderboard(search=search)
            leaderboard.notify.add(user)

        return search

    def update(self, instance, validated_data):
        if 'filters' in validated_data:
            instance.filters = json.dumps(validated_data.get('filters'))
        if 'settings' in validated_data:
            instance.settings = json.dumps(validated_data.get('settings'))
        instance.short_description = validated_data.get('short_description', instance.short_description)
        instance.description = validated_data.get('description', instance.description)
        instance.save()

        #if leaderboard activation was asked, activate the leaderboard
        leaderboard = validated_data.get('leaderboard')
        if leaderboard is True: #create
            # for some reason, get_or_create always raises an IntegrityError
            try:
                obj = Leaderboard.objects.get(search=instance)
            except Leaderboard.DoesNotExist:
                obj = Leaderboard(search=instance)
                obj.save()
            obj.notify.add(instance.author)
        elif leaderboard is False and instance.has_leaderboard(): #delete
            instance.leaderboard.delete()

        return instance
