#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################


import os
import shutil
import simplejson as json

from django.test import TestCase
from django.contrib.auth.models import User
from django.conf import settings
from django.core.urlresolvers import reverse

import beat.core.toolchain

from .models import Toolchain

from ..common.models import Shareable
from ..common.testutils import BaseTestCase, tearDownModule



class ToolchainsCreationFunction(TestCase):

    DECLARATION = {
        "blocks": [ {
                "name": "block1",
                "inputs": [
                    "in"
                ],
                "outputs": [
                    "out"
                ],
                "synchronized_channel": "dataset1"
            }
        ],
        "analyzers": [ {
                "name": "analysis",
                "inputs": [
                    "in"
                ],
                "synchronized_channel": "dataset1"
            }
        ],
        "datasets": [ {
                "name": "dataset1",
                "outputs": [
                    "output1"
                ],
            }
        ],
        "connections": [ {
                "from": "dataset1.output1",
                "to": "block1.in",
                "channel": "dataset1"
            },
            {
                "from": "block1.out",
                "to": "analysis.in",
                "channel": "dataset1"
            }
        ],
        "representation": {
            "connections": {},
            "blocks": {},
            "channel_colors": {}
            }
    }


    def setUp(self):
        if os.path.exists(settings.TOOLCHAINS_ROOT):
            shutil.rmtree(settings.TOOLCHAINS_ROOT)


    def tearDown(self):
        if os.path.exists(settings.TOOLCHAINS_ROOT):
            shutil.rmtree(settings.TOOLCHAINS_ROOT)


    def test_no_declaration(self):
        user = User.objects.create_user('johndoe', 'johndoe@test.org', '1234')

        (toolchain, errors) = Toolchain.objects.create_toolchain(
            author=user,
            name='toolchain1',
            short_description='some description',
            description='some longer description',
            )

        assert toolchain, errors

        self.assertEqual(toolchain.author, user)
        self.assertEqual(toolchain.name, 'toolchain1')
        self.assertEqual(toolchain.short_description, 'some description')
        self.assertEqual(toolchain.description, 'some longer description')
        self.assertEqual(toolchain.sharing, Shareable.PRIVATE)
        self.assertTrue(toolchain.previous_version is None)
        self.assertTrue(toolchain.fork_of is None)

        storage = beat.core.toolchain.Storage(settings.PREFIX,
                toolchain.fullname())
        assert storage.exists

        serializer = beat.core.toolchain.Toolchain(settings.PREFIX,
                storage.fullname)
        assert serializer.valid

    def test_with_valid_declaration(self):
        user = User.objects.create_user('johndoe', 'johndoe@test.org', '1234')

        (toolchain, errors) = Toolchain.objects.create_toolchain(
            author=user,
            name='toolchain1',
            short_description='some description',
            description='some longer description',
            declaration=ToolchainsCreationFunction.DECLARATION,
            )

        assert toolchain, errors

        self.assertEqual(toolchain.author, user)
        self.assertEqual(toolchain.name, 'toolchain1')
        self.assertEqual(toolchain.short_description, 'some description')
        self.assertEqual(toolchain.description, 'some longer description')
        self.assertEqual(toolchain.sharing, Shareable.PRIVATE)
        self.assertTrue(toolchain.previous_version is None)
        self.assertTrue(toolchain.fork_of is None)

        storage = beat.core.toolchain.Storage(settings.PREFIX, toolchain.fullname())
        assert storage.exists

        serializer = beat.core.toolchain.Toolchain(settings.PREFIX, storage.fullname)
        assert serializer.valid, '\n  * %s' % '\n  * '.join(serializer.errors)

        self.assertEqual(len(serializer.data['blocks']), 1)
        self.assertEqual(len(serializer.data['analyzers']), 1)
        self.assertEqual(len(serializer.data['datasets']), 1)
        self.assertEqual(len(serializer.data['connections']), 2)


    def test_with_invalid_declaration(self):
        user = User.objects.create_user('johndoe', 'johndoe@test.org', '1234')

        (toolchain, errors) = Toolchain.objects.create_toolchain(
            author=user,
            name='toolchain1',
            short_description='some description',
            description='some longer description',
            declaration={
                "connections": [
                    {
                        "from": "unknown.output",
                        "to": "missing.input",
                        "channel": "dataset1"
                    }
                ]
            },
            )

        assert errors


class ToolchainAccessibilityFunctionsBase(BaseTestCase):

    def setUp(self):
        if os.path.exists(settings.TOOLCHAINS_ROOT):
            shutil.rmtree(settings.TOOLCHAINS_ROOT)

        # Users
        user1 = User.objects.create_user('jackdoe', 'jackdoe@test.org', '1234')
        user2 = User.objects.create_user('johndoe', 'johndoe@test.org', '1234')
        User.objects.create_user('janedoe', 'janedoe@test.org', '1234')
        User.objects.create_user('jacquelinedoe', 'jacquelinedoe@test.org', '1234')

        # Personal
        (toolchain1, errors) = Toolchain.objects.create_toolchain(
            author=user1,
            name='personal',
            )

        assert toolchain1, errors

        # Accessible to one user
        (toolchain2, errors) = Toolchain.objects.create_toolchain(
            author=user1,
            name='public_for_one_user',
            )

        assert toolchain2, errors

        toolchain2.sharing = Shareable.SHARED
        toolchain2.save()

        toolchain2.shared_with.add(user2)

        # Accessible to all
        (toolchain3, errors) = Toolchain.objects.create_toolchain(
            author=user1,
            name='public_for_all',
            )

        assert toolchain3, errors

        toolchain3.sharing = Shareable.PUBLIC
        toolchain3.save()


    def tearDown(self):
        if os.path.exists(settings.TOOLCHAINS_ROOT):
            shutil.rmtree(settings.TOOLCHAINS_ROOT)



class NotSharedToolchain_CheckAccessibilityFunction(ToolchainAccessibilityFunctionsBase):

    def test_author_has_access(self):
        toolchain = Toolchain.objects.get(name='personal')

        errors = toolchain.is_accessible(users=['jackdoe'])
        self.assertEqual(len(errors), 0)


    def test_other_user_has_no_access(self):
        toolchain = Toolchain.objects.get(name='personal')

        errors = toolchain.is_accessible(users=['johndoe'])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The toolchain 'jackdoe/personal/1' isn't accessible to the following users: johndoe")



class PublicForOneUserToolchain_CheckAccessibilityFunction(ToolchainAccessibilityFunctionsBase):

    def test_author_has_access(self):
        toolchain = Toolchain.objects.get(name='public_for_one_user')

        errors = toolchain.is_accessible(users=['jackdoe'])
        self.assertEqual(len(errors), 0)


    def test_authorized_user_has_full_access(self):
        toolchain = Toolchain.objects.get(name='public_for_one_user')

        errors = toolchain.is_accessible(users=['johndoe'])
        self.assertEqual(len(errors), 0)


    def test_unauthorized_user_has_no_access(self):
        toolchain = Toolchain.objects.get(name='public_for_one_user')

        errors = toolchain.is_accessible(users=['janedoe'])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The toolchain 'jackdoe/public_for_one_user/1' isn't accessible to the following users: janedoe")


    def test_mixed_case(self):
        toolchain = Toolchain.objects.get(name='public_for_one_user')

        errors = toolchain.is_accessible(users=['johndoe', 'janedoe'])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The toolchain 'jackdoe/public_for_one_user/1' isn't accessible to the following users: janedoe")



class PublicForAllUsersToolchain_CheckAccessibilityFunction(ToolchainAccessibilityFunctionsBase):

    def test_author_has_access(self):
        toolchain = Toolchain.objects.get(name='public_for_all')

        errors = toolchain.is_accessible(users=['jackdoe'])
        self.assertEqual(len(errors), 0)


    def test_other_user_has_full_access(self):
        toolchain = Toolchain.objects.get(name='public_for_all')

        errors = toolchain.is_accessible(users=['johndoe'])
        self.assertEqual(len(errors), 0)



class NotSharedToolchain_ShareFunction(ToolchainAccessibilityFunctionsBase):

    def test_sharing_as_public_for_all(self):
        toolchain = Toolchain.objects.get(name='personal')

        toolchain.share()

        reference = {
            'status': 'public',
        }

        self.checkObjectDBSharingPreferences(toolchain, reference)


    def test_sharing_as_public_for_one_user(self):
        toolchain = Toolchain.objects.get(name='personal')

        toolchain.share(users=['johndoe'])

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe'],
        }

        self.checkObjectDBSharingPreferences(toolchain, reference)


    def test_sharing_as_public_for_multiple_users(self):
        toolchain = Toolchain.objects.get(name='personal')

        toolchain.share(users=['johndoe', 'janedoe'])

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe', 'janedoe'],
        }

        self.checkObjectDBSharingPreferences(toolchain, reference)



class PublicForOneUserToolchain_ShareFunction(ToolchainAccessibilityFunctionsBase):

    def test_sharing_as_public_for_all(self):
        toolchain = Toolchain.objects.get(name='public_for_one_user')

        toolchain.share()

        reference = {
            'status': 'public',
        }

        self.checkObjectDBSharingPreferences(toolchain, reference)


    def test_sharing_as_public_for_the_same_user(self):
        toolchain = Toolchain.objects.get(name='public_for_one_user')

        toolchain.share(users=['johndoe'])

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe'],
        }

        self.checkObjectDBSharingPreferences(toolchain, reference)


    def test_sharing_as_public_for_another_user(self):
        toolchain = Toolchain.objects.get(name='public_for_one_user')

        toolchain.share(users=['janedoe'])

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe', 'janedoe'],
        }

        self.checkObjectDBSharingPreferences(toolchain, reference)


    def test_sharing_as_public_for_multiple_users(self):
        toolchain = Toolchain.objects.get(name='public_for_one_user')

        toolchain.share(users=['janedoe', 'jacquelinedoe'])

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe', 'janedoe', 'jacquelinedoe'],
        }

        self.checkObjectDBSharingPreferences(toolchain, reference)



class PublicForAllToolchain_ShareFunction(ToolchainAccessibilityFunctionsBase):

    def test_sharing_as_public_for_all(self):
        toolchain = Toolchain.objects.get(name='public_for_all')

        toolchain.share()

        reference = {
            'status': 'public',
        }

        self.checkObjectDBSharingPreferences(toolchain, reference)


    def test_sharing_as_public_for_one_user(self):
        toolchain = Toolchain.objects.get(name='public_for_all')

        toolchain.share(users=['johndoe'])

        reference = {
            'status': 'public',
        }

        self.checkObjectDBSharingPreferences(toolchain, reference)


    def test_sharing_as_public_for_multiple_users(self):
        toolchain = Toolchain.objects.get(name='public_for_all')

        toolchain.share(users=['johndoe', 'janedoe'])

        reference = {
            'status': 'public',
        }

        self.checkObjectDBSharingPreferences(toolchain, reference)



class ToolchainsAPIBase(BaseTestCase):
    class Meta:
        model = Toolchain

    DECLARATION = {
        "blocks": [ {
                "name": "block1",
                "inputs": [
                    "in"
                ],
                "outputs": [
                    "out"
                ],
                "synchronized_channel": "dataset1"
            }
        ],
        "analyzers": [ {
                "name": "analysis",
                "inputs": [
                    "in"
                ],
                "synchronized_channel": "dataset1"
            }
        ],
        "datasets": [ {
                "name": "dataset1",
                "outputs": [
                    "output1"
                ]
            }
        ],
        "connections": [ {
                "from": "dataset1.output1",
                "to": "block1.in",
                "channel": "dataset1"
            },
            {
                "from": "block1.out",
                "to": "analysis.in",
                "channel": "dataset1"
            }
        ],
        "representation": {
            "connections": {},
            "blocks": {},
            "channel_colors": {}
            }
    }


    def setUp(self):
        if os.path.exists(settings.TOOLCHAINS_ROOT):
            shutil.rmtree(settings.TOOLCHAINS_ROOT)

        # Users
        user1 = User.objects.create_user('jackdoe', 'jackdoe@test.org', '1234')
        user2 = User.objects.create_user('johndoe', 'johndoe@test.org', '1234')
        User.objects.create_user('janedoe', 'janedoe@test.org', '1234')
        User.objects.create_user('jacquelinedoe', 'jacquelinedoe@test.org', '1234')

        # Personal
        (toolchain, errors) = Toolchain.objects.create_toolchain(
            author=user1,
            name='personal',
            declaration=ToolchainsAPIBase.DECLARATION,
            )

        assert toolchain, errors

        # Accessible to one user
        (toolchain, errors) = Toolchain.objects.create_toolchain(
            author=user1,
            name='public_for_one_user',
            declaration=ToolchainsAPIBase.DECLARATION,
            )
        toolchain.share(users=[user2])

        assert toolchain, errors

        # Accessible to all
        (toolchain, errors) = Toolchain.objects.create_toolchain(
            author=user1,
            name='public_for_all',
            declaration=ToolchainsAPIBase.DECLARATION,
            )
        toolchain.share()

        assert toolchain, errors


    def tearDown(self):
        if os.path.exists(settings.TOOLCHAINS_ROOT):
            shutil.rmtree(settings.TOOLCHAINS_ROOT)


class ToolchainsListRetrieval(ToolchainsAPIBase):
    def setUp(self):
        super(ToolchainsListRetrieval, self).setUp()

        self.url = reverse('api_toolchains:all')
        self.owner_url = reverse('api_toolchains:list_create', args=['jackdoe'])

    def test_only_public_and_visible_toolchains_for_anonymous_user(self):
        response = self.client.get(self.url)
        content = self.checkResponse(response, 200, content_type='application/json')

        self.assertTrue(isinstance(content, list))
        self.assertEqual(len(content), 1)

        entry = filter(lambda x: x['name'] == 'jackdoe/public_for_all/1', content)[0]
        self.assertTrue(entry is not None)
        self.assertEqual(entry['accessibility'], 'public')
        self.assertFalse(entry.has_key('is_owner'))


    def test_all_accessible_toolchains(self):
        self.client.login(username='johndoe', password='1234')

        response = self.client.get(self.url)

        content = self.checkResponse(response, 200, content_type='application/json')

        self.assertTrue(isinstance(content, list))
        self.assertEqual(len(content), 2)

        entry = filter(lambda x: x['name'] == 'jackdoe/public_for_all/1', content)[0]
        self.assertTrue(entry is not None)
        self.assertEqual(entry['accessibility'], 'public')
        self.assertFalse(entry['is_owner'])

        entry = filter(lambda x: x['name'] == 'jackdoe/public_for_one_user/1', content)[0]
        self.assertTrue(entry is not None)
        self.assertEqual(entry['accessibility'], 'confidential')
        self.assertFalse(entry['is_owner'])


    def test_own_toolchains(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.get(self.owner_url)

        content = self.checkResponse(response, 200, content_type='application/json')

        self.assertTrue(isinstance(content, list))
        self.assertEqual(len(content), 3)

        entry = filter(lambda x: x['name'] == 'jackdoe/personal/1', content)[0]
        self.assertTrue(entry is not None)
        self.assertEqual(entry['accessibility'], 'private')
        self.assertTrue(entry['is_owner'])

        entry = filter(lambda x: x['name'] == 'jackdoe/public_for_all/1', content)[0]
        self.assertTrue(entry is not None)
        self.assertEqual(entry['accessibility'], 'public')
        self.assertTrue(entry['is_owner'])

        entry = filter(lambda x: x['name'] == 'jackdoe/public_for_one_user/1', content)[0]
        self.assertTrue(entry is not None)
        self.assertEqual(entry['accessibility'], 'confidential')
        self.assertTrue(entry['is_owner'])


    def test_other_user_toolchains(self):
        self.client.login(username='johndoe', password='1234')

        response = self.client.get(self.url)

        content = self.checkResponse(response, 200, content_type='application/json')

        self.assertTrue(isinstance(content, list))
        self.assertEqual(len(content), 2)

        entry = filter(lambda x: x['name'] == 'jackdoe/public_for_all/1', content)[0]
        self.assertTrue(entry is not None)
        self.assertEqual(entry['accessibility'], 'public')
        self.assertFalse(entry['is_owner'])

        entry = filter(lambda x: x['name'] == 'jackdoe/public_for_one_user/1', content)[0]
        self.assertTrue(entry is not None)
        self.assertEqual(entry['accessibility'], 'confidential')
        self.assertFalse(entry['is_owner'])


    def test_user_toolchains_for_anonymous_user(self):
        response = self.client.get(self.owner_url)

        content = self.checkResponse(response, 200, content_type='application/json')

        self.assertTrue(isinstance(content, list))
        self.assertEqual(len(content), 1)

        entry = filter(lambda x: x['name'] == 'jackdoe/public_for_all/1', content)[0]
        self.assertTrue(entry is not None)
        self.assertEqual(entry['accessibility'], 'public')
        self.assertFalse(entry.has_key('is_owner'))



class ToolchainsNameCheck(ToolchainsAPIBase):

    def setUp(self):
        super(ToolchainsNameCheck, self).setUp()
        user = User.objects.get(username='jackdoe')
        (toolchain4, errors) = Toolchain.objects.create_toolchain(
            author=user,
            name='toolchain-4',
            short_description='short description 4',
            )

        assert toolchain4, errors

        self.url = reverse('api_toolchains:check_name')


    def test_no_check_for_anonymous_user(self):
        response = self.client.post(self.url,
            json.dumps({
                'name': 'valid-name1',
                'private': True,
            }), content_type='application/json')
        self.checkResponse(response, 403)

    def test_bad_request_for_post_method_without_needed_data(self):
        self.client.login(username='jackdoe', password='1234')
        response = self.client.post(self.url,
                                    json.dumps({}),
                                    content_type='application/json')
        self.checkResponse(response, 400)


    def test_not_used_valid_name(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'valid-name1',
                'private': True,
            }), content_type='application/json')

        content = self.checkResponse(response, 200, content_type='application/json')

        self.assertTrue(isinstance(content, dict))
        self.assertEqual(content['name'], 'valid-name1')
        self.assertFalse(content['used'])


    def test_not_used_invalid_name(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'invalid name_1 #due-to{}%&é',
                'private': True,
            }), content_type='application/json')

        content = self.checkResponse(response, 200, content_type='application/json')

        self.assertTrue(isinstance(content, dict))
        self.assertEqual(content['name'], 'invalid-name_1--due-to-----')
        self.assertFalse(content['used'])


    def test_used_valid_name(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'personal',
                'private': True,
            }), content_type='application/json')

        content = self.checkResponse(response, 200, content_type='application/json')

        self.assertTrue(isinstance(content, dict))
        self.assertEqual(content['name'], 'personal')
        self.assertTrue(content['used'])


    def test_used_invalid_name(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'toolchain 4',
                'private': True,
            }), content_type='application/json')

        content = self.checkResponse(response, 200, content_type='application/json')

        self.assertTrue(isinstance(content, dict))
        self.assertEqual(content['name'], 'toolchain-4')
        self.assertTrue(content['used'])



class ToolchainCreation(ToolchainsAPIBase):

    DECLARATION_INVALID = {
        "blocks": [ {
                "name": "block1",
                "inputs": [
                    "in"
                ],
                "outputs": [
                    "out"
                ],
                "synchronized_channel": "dataset1"
            }
        ],
        "analyzers": [ {
                "name": "analysis",
                "inputs": [
                    "in"
                ],
                "synchronized_channel": "dataset1"
            }
        ],
        "datasets": [ {
                "name": "dataset1",
                "outputs": [
                    "output1"
                ]
            }
        ],
        "connections": [ {
                "from": "dataset1.output1",
                "to": "block1.in",
                "channel": "dataset1"
            },
            {
                "from": "block2.out",
                "to": "analysis.in",
                "channel": "dataset1"
            }
        ],
        "representation": {
            "connections": {},
            "blocks": {},
            "channel_colors": {}
            }
    }


    def setUp(self):

        super(ToolchainCreation, self).setUp()

        user = User.objects.get(username='jackdoe')
        (toolchain4, errors) = Toolchain.objects.create_toolchain(
            author=user,
            name='toolchain-4',
            short_description='short description 4',
            )

        assert toolchain4, errors

        self.url = reverse('api_toolchains:list_create', args=['jackdoe'])

    def tearDown(self):
        if os.path.exists(settings.TOOLCHAINS_ROOT):
            shutil.rmtree(settings.TOOLCHAINS_ROOT)


    def checkToolchainFile(self, filename, jsondata=None):

        storage = beat.core.toolchain.Storage(settings.PREFIX, filename)
        assert storage.exists

        if jsondata is None: #checks only toolchain
            tc = beat.core.toolchain.Toolchain(settings.PREFIX,
                    storage.fullname)
        else: #checks experiment (and toolchain indirectly)
            tc = beat.core.toolchain.Toolchain(settings.PREFIX, jsondata)

        assert tc.valid, \
                '\n  * %s' % '\n  * '.join(tc.errors)

    def test_no_creation_for_anonymous_user(self):
        response = self.client.post(self.url)
        self.checkResponse(response, 403)


    def test_no_creation_for_other_user(self):
        self.client.login(username='jackdoe', password='1234')
        url = reverse('api_toolchains:list_create', args=['johndoe'])
        response = self.client.post(url)
        self.checkResponse(response, 403)


    def test_bad_request_for_post_method_without_needed_data(self):
        self.client.login(username='jackdoe', password='1234')
        response = self.client.post(self.url,
                                    json.dumps({}),
                                    content_type='application/json')
        self.checkResponse(response, 400)


    def test_not_used_valid_name(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'valid-name1'}), content_type='application/json')

        url = reverse('api_toolchains:object', args=['jackdoe', 'valid-name1', 1])
        content = self.checkResponse(response, 201,
                                     content_type='application/json',
                                     location=url)

        self.assertTrue(isinstance(content, dict))
        self.assertEqual(content['name'], 'valid-name1')
        self.assertTrue(content['url'].endswith(url))

        try:
            toolchain = Toolchain.objects.get(author__username='jackdoe',
                                              name='valid-name1', version=1)
        except:
            self.assertTrue(False)

        self.assertEqual(toolchain.short_description, '')

        self.checkToolchainFile(toolchain.fullname())


    def test_not_used_invalid_name(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'invalid name1'
            }), content_type='application/json')

        url = reverse('api_toolchains:object', args=['jackdoe', 'invalid-name1', 1])
        content = self.checkResponse(response, 201,
                                     content_type='application/json',
                                     location=url)

        self.assertTrue(isinstance(content, dict))
        self.assertEqual(content['name'], 'invalid-name1')
        self.assertTrue(content['url'].endswith(url))

        try:
            toolchain = Toolchain.objects.get(author__username='jackdoe',
                                              name='invalid-name1', version=1)
        except:
            self.assertTrue(False)

        self.assertEqual(toolchain.short_description, '')

        self.checkToolchainFile(toolchain.fullname())


    def test_used_valid_name(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'personal'
            }), content_type='application/json')

        self.checkResponse(response, 400)


    def test_used_invalid_name(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'toolchain 4'
            }), content_type='application/json')

        self.checkResponse(response, 400)


    def test_with_description(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'new_toolchain',
                'description': 'blah'
                }), content_type='application/json')

        url = reverse('api_toolchains:object', args=['jackdoe', 'new_toolchain', 1])
        content = self.checkResponse(response, 201,
                content_type='application/json',
                location=url)

        url = reverse('api_toolchains:object', args=['jackdoe', 'new_toolchain', 1])
        self.assertTrue(isinstance(content, dict))
        self.assertEqual(content['name'], 'new_toolchain')
        self.assertTrue(content['url'].endswith(url))

        try:
            toolchain = Toolchain.objects.get(author__username='jackdoe', name='new_toolchain')
        except:
            self.assertTrue(False)

        self.assertEqual(toolchain.short_description, '')
        self.assertEqual(toolchain.description, 'blah')

        self.checkToolchainFile(toolchain.fullname())


    def test_with_valid_declaration(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'new_toolchain',
                'declaration': ToolchainsAPIBase.DECLARATION,
                }), content_type='application/json')

        url = reverse('api_toolchains:object', args=['jackdoe', 'new_toolchain', 1])
        content = self.checkResponse(response, 201,
                                     content_type='application/json',
                                     location=url)

        self.assertTrue(isinstance(content, dict))
        self.assertEqual(content['name'], 'new_toolchain')
        self.assertTrue(content['url'].endswith(url))

        try:
            toolchain = Toolchain.objects.get(author__username='jackdoe', name='new_toolchain')
        except:
            self.assertTrue(False)

        self.checkToolchainFile(toolchain.fullname(), ToolchainsAPIBase.DECLARATION)


    def test_with_invalid_declaration(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'new_toolchain',
                'declaration': ToolchainCreation.DECLARATION_INVALID,
                }), content_type='application/json')

        content = self.checkResponse(response, 400, content_type='application/json')
        assert content.find("The toolchain declaration is **invalid**") != -1


    def test_no_forking_for_anonymous_user(self):
        response = self.client.post(self.url,
            json.dumps({
                'name': 'name1',
                'description': 'blah',
                'fork_of': 'johndoe/toolchain1/1'
                }), content_type='application/json')

        self.checkResponse(response, 403)


    def test_no_forking_for_other_user(self):
        self.client.login(username='jackdoe', password='1234')

        url = reverse('api_toolchains:list_create', args=['johndoe'])
        response = self.client.post(url,
            json.dumps({
                'name': 'name1',
                'description': 'blah',
                'fork_of': 'johndoe/toolchain1/1'
                }), content_type='application/json')

        self.checkResponse(response, 403)


    def test_unsuccessful_forking_with_no_name(self):
        self.client.login(username='johndoe', password='1234')

        url = reverse('api_toolchains:list_create', args=['johndoe'])
        response = self.client.post(url,
            json.dumps({
                'description': 'blah',
                'fork_of': 'jackdoe/public_for_one_user/1'
                }), content_type='application/json')

        content = self.checkResponse(response, 400, content_type='application/json')
        self.assertEqual(content, {'name': ['This field is required.']})


    def test_unsuccessful_forking_of_non_accredited_person_public_for_one_user_toolchain(self):
        self.client.login(username='janedoe', password='1234')

        url = reverse('api_toolchains:list_create', args=['janedoe'])
        response = self.client.post(url,
            json.dumps({
                'name': 'name1',
                'description': 'blah',
                'fork_of': 'jackdoe/public_for_one_user/1'
                }), content_type='application/json')

        content = self.checkResponse(response, 400, content_type='application/json')
        self.assertEqual(content, {'non_field_errors': ['No access allowed']})


    def test_unsuccessful_forking_of_personal_toolchain(self):
        self.client.login(username='janedoe', password='1234')

        url = reverse('api_toolchains:list_create', args=['janedoe'])
        response = self.client.post(url,
            json.dumps({
                'name': 'name1',
                'description': 'blah',
                'fork_of': 'jackdoe/personal/1'
                }), content_type='application/json')

        content = self.checkResponse(response, 400, content_type='application/json')
        self.assertEqual(content, {'non_field_errors': ['No access allowed']})


    def test_successful_forking_from_public_for_all_toolchain(self):
        self.client.login(username='janedoe', password='1234')

        url = reverse('api_toolchains:list_create', args=['janedoe'])
        response = self.client.post(url,
            json.dumps({
                'name': 'name1',
                'description': 'blah',
                'fork_of': 'jackdoe/public_for_all/1'
                }), content_type='application/json')

        url = reverse('api_toolchains:object', args=['janedoe', 'name1', 1])
        content = self.checkResponse(response, 201,
                                     content_type='application/json',
                                     location=url)

        self.assertTrue(isinstance(content, dict))
        self.assertEqual(content['name'], 'name1')
        self.assertTrue(content['url'].endswith(url))

        try:
            toolchain = Toolchain.objects.get(author__username='jackdoe',
                                              name='public_for_all', version=1)

            toolchain_forked = Toolchain.objects.get(author__username='janedoe',
                                                     name='name1', version=1)
        except:
            self.assertTrue(False)

        self.assertEqual(toolchain_forked.description, 'blah')
        self.assertEqual(toolchain.description, '')
        self.assertEqual(toolchain_forked.fork_of, toolchain)

        orginal_storage = beat.core.toolchain.Storage(settings.PREFIX,
                toolchain.fullname())
        assert orginal_storage.exists
        destination_storage = beat.core.toolchain.Storage(settings.PREFIX,
                toolchain_forked.fullname())
        assert destination_storage.exists

        self.checkToolchainFile(toolchain.fullname(),
                ToolchainsAPIBase.DECLARATION)
        self.checkToolchainFile(toolchain_forked.fullname(),
                ToolchainsAPIBase.DECLARATION)


    def test_successful_forking_public_for_one_user(self):
        self.client.login(username='johndoe', password='1234')

        url = reverse('api_toolchains:list_create', args=['johndoe'])
        response = self.client.post(url,
            json.dumps({
                'name': 'name1',
                'description': 'blah',
                'fork_of': 'jackdoe/public_for_one_user/1'
                }), content_type='application/json')

        url = reverse('api_toolchains:object', args=['johndoe', 'name1', 1])
        content = self.checkResponse(response, 201,
                                     content_type='application/json',
                                     location=url)

        self.assertTrue(isinstance(content, dict))
        self.assertEqual(content['name'], 'name1')
        self.assertTrue(content['url'].endswith(url))

        try:
            toolchain        = Toolchain.objects.get(author__username='jackdoe', name='public_for_one_user', version=1)
            toolchain_forked = Toolchain.objects.get(author__username='johndoe', name='name1', version=1)
        except:
            self.assertTrue(False)

        self.assertEqual(toolchain.short_description, '')
        self.assertEqual(toolchain_forked.description, 'blah')
        self.assertEqual(toolchain_forked.fork_of, toolchain)

        orginal_storage = beat.core.toolchain.Storage(settings.PREFIX,
                toolchain.fullname())
        assert orginal_storage.exists
        destination_storage = beat.core.toolchain.Storage(settings.PREFIX,
                toolchain_forked.fullname())
        assert destination_storage.exists

        self.checkToolchainFile(toolchain.fullname(),
                ToolchainsAPIBase.DECLARATION)
        self.checkToolchainFile(toolchain_forked.fullname(),
                ToolchainsAPIBase.DECLARATION)


    def test_successful_forking_with_declaration(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'name1',
                'description': 'blah',
                'declaration': ToolchainsAPIBase.DECLARATION,
                'fork_of': 'jackdoe/public_for_one_user/1'
                }), content_type='application/json')

        url = reverse('api_toolchains:object', args=['jackdoe', 'name1', 1])
        self.checkResponse(response, 201,
                           content_type='application/json',
                           location=url)

        toolchain_forked = Toolchain.objects.get(author__username='jackdoe', name='name1')

        self.checkToolchainFile(toolchain_forked.fullname(), ToolchainsAPIBase.DECLARATION)



class ToolchainUpdate(ToolchainsAPIBase):

    DECLARATION_UPDATE = {
        "blocks": [ {
                "name": "block2",
                "inputs": [
                    "in"
                ],
                "outputs": [
                    "out"
                ],
                "synchronized_channel": "dataset1"
            }
        ],
        "analyzers": [ {
                "name": "analysis",
                "inputs": [
                    "in"
                ],
                "synchronized_channel": "dataset1"
            }
        ],
        "datasets": [ {
                "name": "dataset1",
                "outputs": [
                    "output1"
                ]
            }
        ],
        "connections": [ {
                "from": "dataset1.output1",
                "to": "block2.in",
                "channel": "dataset1"
            },
            {
                "from": "block2.out",
                "to": "analysis.in",
                "channel": "dataset1"
            }
        ],
        "representation": {
            "connections": {},
            "blocks": {},
            "channel_colors": {}
            }
    }


    DECLARATION_INVALID = {
        "blocks": [ {
                "name": "block1",
                "inputs": [
                    "in"
                ],
                "outputs": [
                    "out"
                ],
                "synchronized_channel": "dataset1"
            }
        ],
        "analyzers": [ {
                "name": "analysis",
                "inputs": [
                    "in"
                ],
                "synchronized_channel": "dataset1"
            }
        ],
        "datasets": [ {
                "name": "dataset1",
                "outputs": [
                    "output1"
                ],
            }
        ],
        "connections": [ {
                "from": "dataset1.output1",
                "to": "block1.in",
                "channel": "dataset1"
            },
            {
                "from": "block2.out",
                "to": "analysis.in",
                "channel": "dataset1"
            }
        ],
        "representation": {
            "connections": {},
            "blocks": {},
            "channel_colors": {}
            }
    }

    def setUp(self):
        super(ToolchainUpdate, self).setUp()
        self.url = reverse('api_toolchains:object', args=['jackdoe', 'personal', 1])

    def checkToolchainFileAndData(self, filename, data, jsondata=None):

        storage = beat.core.toolchain.Storage(settings.PREFIX, filename)
        assert storage.exists

        if jsondata is None: #checks only toolchain
            tc = beat.core.toolchain.Toolchain(settings.PREFIX,
                    storage.fullname)
        else: #checks experiment (and toolchain indirectly)
            tc = beat.core.toolchain.Toolchain(settings.PREFIX, jsondata)

        assert tc.valid, \
                '\n  * %s' % '\n  * '.join(tc.errors)

        self.assertTrue(isinstance(data, dict))
        self.assertTrue(data.has_key('declaration'))


    def test_no_update_for_anonymous_user(self):
        response = self.client.put(self.url,
             json.dumps({
                'description': 'blah',
                }), content_type='application/json')

        self.checkResponse(response, 403)


    def test_no_update_for_other_user(self):
        self.client.login(username='johndoe', password='1234')
        response = self.client.put(self.url,
             json.dumps({
                'description': 'blah',
                }), content_type='application/json')

        self.checkResponse(response, 403)


    def test_fail_to_update_with_invalid_username(self):
        self.client.login(username='jackdoe', password='1234')

        url = reverse('api_toolchains:object', args=['unknown', 'personal', 1])
        response = self.client.put(url,
            json.dumps({
                'description': 'blah',
            }), content_type='application/json')

        self.checkResponse(response, 403)


    def test_fail_to_update_with_invalid_toolchain_name(self):
        self.client.login(username='jackdoe', password='1234')

        url = reverse('api_toolchains:object', args=['jackdoe', 'unknown', 1])
        response = self.client.put(url,
            json.dumps({
                'description': 'blah',
            }), content_type='application/json')

        self.checkResponse(response, 404)


    def test_fail_to_update_with_invalid_version_number(self):
        self.client.login(username='jackdoe', password='1234')

        url = reverse('api_toolchains:object', args=['jackdoe', 'personal', 2])
        response = self.client.put(url,
            json.dumps({
                'description': 'blah',
            }), content_type='application/json')

        self.checkResponse(response, 404)


    def test_fail_to_update_without_version_number(self):
        self.client.login(username='jackdoe', password='1234')

        url = reverse('api_toolchains:object', args=['jackdoe', 'personal'])
        response = self.client.put(url,
            json.dumps({
                'description': 'blah',
            }), content_type='application/json')

        self.checkResponse(response, 400)


    def test_fail_to_update_without_content_not_json_request(self):
        self.client.login(username='jackdoe', password='1234')
        response = self.client.put(self.url)
        self.checkResponse(response, 400)


    def test_fail_to_update_without_content(self):
        self.client.login(username='jackdoe', password='1234')
        response = self.client.put(self.url, '{}',
                                    content_type='application/json')
        self.checkResponse(response, 400, content_type='application/json')


    def test_successful_update_description_only(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.put(self.url,
            json.dumps({
                'description': 'blah',
            }), content_type='application/json')

        self.checkResponse(response, 200, content_type='application/json')

        toolchain = Toolchain.objects.get(author__username='jackdoe', name='personal')
        self.assertEqual(toolchain.description, 'blah')


    def test_successful_update_declaration_only(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.put(self.url,
            json.dumps({
                'declaration': ToolchainUpdate.DECLARATION_UPDATE,
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        toolchain = Toolchain.objects.get(author__username='jackdoe', name='personal')
        response = self.client.get(self.url)
        data = self.checkResponse(response, 200, content_type='application/json')

        self.checkToolchainFileAndData(toolchain.fullname(), data, ToolchainUpdate.DECLARATION_UPDATE)


    def test_fail_update_invalid_declaration_only(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.put(self.url,
            json.dumps({
                'declaration': ToolchainUpdate.DECLARATION_INVALID,
            }), content_type='application/json')

        content = self.checkResponse(response, 400,
                                     content_type='application/json'
                                    )
        assert content['declaration'][0].find("invalid output endpoint 'block2.out'") != -1


    def test_successful_update_invalid_declaration_description(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.put(self.url,
            json.dumps({
                'description': 'blah',
                'declaration': self.DECLARATION_INVALID,
            }), content_type='application/json')

        content = self.checkResponse(response, 400,
                                     content_type='application/json'
                                    )
        assert content['declaration'][0].find("invalid output endpoint 'block2.out'") != -1


    def test_fail_update_declaration_and_description(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.put(self.url,
            json.dumps({
                'description': 'blah',
                'declaration': ToolchainUpdate.DECLARATION_UPDATE,
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        toolchain = Toolchain.objects.get(author__username='jackdoe', name='personal')
        self.assertEqual(toolchain.description, 'blah')

        response = self.client.get(self.url)
        data = self.checkResponse(response, 200, content_type='application/json')

        self.checkToolchainFileAndData(toolchain.fullname(), data, ToolchainUpdate.DECLARATION_UPDATE)



class ToolchainRetrieval(ToolchainsAPIBase):

    def checkToolchainData(self, data):
        self.assertTrue(isinstance(data, dict))
        self.assertTrue(data.has_key('declaration'))

        tc = beat.core.toolchain.Toolchain(settings.PREFIX, data['declaration'])

        self.assertEqual(len(tc.data['blocks']), 1)
        self.assertEqual(len(tc.data['datasets']), 1)
        self.assertEqual(len(tc.data['analyzers']), 1)
        self.assertEqual(len(tc.data['connections']), 2)


    def test_no_retrieval_of_confidential_toolchain_for_anonymous_user(self):
        url = reverse('api_toolchains:object', args=['jackdoe', 'personal', 1])
        response = self.client.get(url)
        self.checkResponse(response, 404)


    def test_fail_to_retrieve_with_invalid_username(self):
        self.client.login(username='jackdoe', password='1234')

        url = reverse('api_toolchains:object', args=['unknown', 'personal'])
        response = self.client.get(url)
        self.checkResponse(response, 404)


    def test_fail_to_retrieve_with_invalid_toolchain_name(self):
        self.client.login(username='jackdoe', password='1234')

        url = reverse('api_toolchains:object', args=['jackdoe', 'unknown'])
        response = self.client.get(url)
        self.checkResponse(response, 404)


    def test_no_retrieval_of_confidential_toolchain(self):
        self.client.login(username='johndoe', password='1234')
        url = reverse('api_toolchains:object', args=['jackdoe', 'personal', 1])
        response = self.client.get(url)
        self.checkResponse(response, 404)


    def test_successful_retrieval_of_public_toolchain_for_anonymous_user(self):
        url = reverse('api_toolchains:object', args=['jackdoe', 'public_for_all', 1])
        response = self.client.get(url)
        data = self.checkResponse(response, 200, content_type='application/json')

        self.assertEqual(data['accessibility'], 'public')
        self.assertFalse(data.has_key('is_owner'))

        self.checkToolchainData(data)


    def test_successful_retrieval_of_public_toolchain(self):
        self.client.login(username='johndoe', password='1234')

        url = reverse('api_toolchains:object', args=['jackdoe', 'public_for_all', 1])
        response = self.client.get(url)
        data = self.checkResponse(response, 200, content_type='application/json')

        self.assertEqual(data['is_owner'], False)
        self.assertEqual(data['accessibility'], 'public')

        self.checkToolchainData(data)


    def test_successful_retrieval_of_public_for_one_user_toolchain(self):
        self.client.login(username='johndoe', password='1234')

        url = reverse('api_toolchains:object', args=['jackdoe', 'public_for_one_user', 1])
        response = self.client.get(url)
        data = self.checkResponse(response, 200, content_type='application/json')

        self.assertEqual(data['is_owner'], False)
        self.assertEqual(data['accessibility'], 'confidential')

        self.checkToolchainData(data)


    def test_successful_retrieval_of_own_public_toolchain(self):
        self.client.login(username='jackdoe', password='1234')

        url = reverse('api_toolchains:object', args=['jackdoe', 'public_for_all', 1])
        response = self.client.get(url)
        data = self.checkResponse(response, 200, content_type='application/json')

        self.assertEqual(data['is_owner'], True)
        self.assertEqual(data['accessibility'], 'public')

        self.checkToolchainData(data)


    def test_successful_retrieval_of_own_private_toolchain(self):
        self.client.login(username='jackdoe', password='1234')

        url = reverse('api_toolchains:object', args=['jackdoe', 'personal', 1])
        response = self.client.get(url)
        data = self.checkResponse(response, 200, content_type='application/json')

        self.assertEqual(data['is_owner'], True)
        self.assertEqual(data['accessibility'], 'private')

        self.checkToolchainData(data)



class ToolchainDeletion(ToolchainsAPIBase):
    def setUp(self):
        super(ToolchainDeletion, self).setUp()
        self.url = reverse('api_toolchains:object', args=['jackdoe', 'personal', 1])

    def test_no_deletion_for_anonymous_user(self):
        response = self.client.delete(self.url)
        self.checkResponse(response, 403)


    def test_fail_to_delete_with_invalid_username(self):
        self.client.login(username='jackdoe', password='1234')
        url = reverse('api_toolchains:object', args=['unknown', 'personal', 1])
        response = self.client.delete(url)
        self.checkResponse(response, 403)


    def test_fail_to_delete_with_invalid_toolchain_name(self):
        self.client.login(username='jackdoe', password='1234')
        url = reverse('api_toolchains:object', args=['jackdoe', 'unknown', 1])
        response = self.client.delete(url)
        self.checkResponse(response, 404)


    def test_fail_to_delete_with_invalid_version(self):
        self.client.login(username='jackdoe', password='1234')
        url = reverse('api_toolchains:object', args=['jackdoe', 'personal', 2])

        response = self.client.delete(url)
        self.checkResponse(response, 404)


    def test_fail_to_delete_without_version_number(self):
        self.client.login(username='jackdoe', password='1234')

        url = reverse('api_toolchains:object', args=['jackdoe', 'personal'])
        response = self.client.delete(url)
        self.checkResponse(response, 400)


    def test_no_deletion_of_not_owned_toolchain(self):
        self.client.login(username='johndoe', password='1234')
        response = self.client.delete(self.url)
        self.checkResponse(response, 403)


    def test_successful_deletion_of_private_toolchain(self):
        self.client.login(username='jackdoe', password='1234')
        response = self.client.delete(self.url)
        self.checkResponse(response, 204)



class ToolchainSharingAPIBase(ToolchainsAPIBase):

    def checkJSONSharingPreferences(self, sharing_preferences, reference):
        self.assertEqual(len(sharing_preferences.keys()), len(reference.keys()))

        for (key, value) in reference.items():
            self.assertTrue(sharing_preferences.has_key(key))

            if isinstance(value, list):
                self.assertEqual(len(sharing_preferences[key]), len(value))
                for entry in value:
                    self.assertTrue(entry in sharing_preferences[key])
            else:
                self.assertEqual(sharing_preferences[key], value)


class ToolchainSharingAPI_Failures(ToolchainSharingAPIBase):
    def setUp(self):
        super(ToolchainSharingAPI_Failures, self).setUp()

        self.url = reverse('api_toolchains:share', args=['jackdoe', 'personal', 1])


    def test_fail_sharing_for_anonymous_user(self):
        response = self.client.post(self.url,
            json.dumps({
            }), content_type='application/json')

        self.checkResponse(response, 403)


    def test_fail_sharing_for_unauthorized_user(self):
        self.client.login(username='johndoe', password='1234')
        response = self.client.post(self.url,
            json.dumps({
            }), content_type='application/json')

        self.checkResponse(response, 403)


    def test_fail_sharing_to_public_for_unknown_user(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['doejack'],
            }), content_type='application/json')

        data = self.checkResponse(response, 400, content_type='application/json')
        self.assertTrue(data is not None)
        self.assertTrue(isinstance(data['users'], list))



class NotSharedToolchain_SharingAPI(ToolchainSharingAPIBase):
    def setUp(self):
        super(NotSharedToolchain_SharingAPI, self).setUp()

        self.url = reverse('api_toolchains:share', args=['jackdoe', 'personal', 1])

    def test_sharing_as_public_for_all(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'public',
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'personal', reference)


    def test_sharing_as_public_for_one_user(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['johndoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'personal', reference)


    def test_sharing_as_public_for_multiple_users(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['johndoe', 'janedoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe', 'janedoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'personal', reference)



class PublicForOneUserToolchain_SharingAPI(ToolchainSharingAPIBase):
    def setUp(self):
        super(PublicForOneUserToolchain_SharingAPI, self).setUp()

        self.url = reverse('api_toolchains:share', args=['jackdoe', 'public_for_one_user', 1])

    def test_sharing_as_public_for_all(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'public',
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_one_user', reference)


    def test_sharing_as_public_for_the_same_user(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['johndoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_one_user', reference)


    def test_sharing_as_public_for_another_user(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['janedoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe', 'janedoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_one_user', reference)


    def test_sharing_as_public_for_multiple_users(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['janedoe', 'jacquelinedoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe', 'janedoe', 'jacquelinedoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_one_user', reference)



class PublicForAllToolchain_SharingAPI(ToolchainSharingAPIBase):
    def setUp(self):
        super(PublicForAllToolchain_SharingAPI, self).setUp()

        self.url = reverse('api_toolchains:share', args=['jackdoe', 'public_for_all', 1])


    def test_sharing_as_public_for_all(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'public',
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_all', reference)


    def test_sharing_as_public_for_one_user(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users': ['johndoe']
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'public',
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_all', reference)


    def test_sharing_as_public_for_multiple_users(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users': ['johndoe', 'janedoe']
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'public',
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_all', reference)
